(* ----------------------------------------------------------------------
   $Id: Control.m3,v 1.12 1995/06/23 07:51:23 sgp13 Exp $

   Control module for testing the CPU and KeyboardUI.

   ----------------------------------------------------------------------

   $Log: Control.m3,v $
# Revision 1.12  1995/06/23  07:51:23  sgp13
# Registers are now accessed directly
#
# Revision 1.11  1995/06/21  13:59:28  sgp13
# Superfluous Init.Reset() calls removed.
#
# Revision 1.10  1995/06/19  14:04:02  sgp13
# Graphical user interface now includes teletype and implements character by
# character input.
#
# Revision 1.9  1995/05/03  21:38:43  sgp13
# Focal assumes that the system starts up as if a KCC has been issued, so one
# is issued on startup.
#
# Revision 1.8  1995/03/23  12:01:19  sgp13
# Fixed to supply filename information to the GUI.
# Attaches stdin and stdout to the keyboard and teletype respectively.
#
# Revision 1.7  1995/03/22  13:09:15  sgp13
# Graphical user interface added: choice between the two UI types
# given in initialisation file.
#
# Revision 1.6  1995/03/15  21:01:29  sgp13
# New sections of initialisation file added: [IOFiles] and [AutoExec]
#
# Revision 1.5  1995/03/15  14:51:54  sgp13
# Initialisation of character readers patched to give knowledge of cpu object,
# so that interrupt indices can be allocated to them
#
# Revision 1.4  1995/03/14  12:21:56  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.3  1995/03/07  23:19:10  sgp13
# Altered to implement Tape and Teletype I/O
#
# Revision 1.2  1995/02/24  22:19:50  sgp13
# Altered to cope with new system.
#
# Revision 1.1  1995/01/19  09:29:18  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Control EXPORTS Main;

IMPORT ASCII, FileRd, FileWr, FormsVBT, OSError, Rd, Scan, Stdio, Text, Thread;
IMPORT TypescriptVBT, Wr;

IMPORT Buffer, BufferRd, CharRd, Init, Diag, Memory, Registers;
IMPORT CharReader, CharWriter;

IMPORT GeneralUI, GraphicalUI, KeyboardUI, Processor;

EXCEPTION AbnormalExit;

<* FATAL AbnormalExit, Registers.OutOfRange *>

VAR
  (* the closures for the threads to be run *)
  processor := NEW(Processor.T);
  keyboard := NEW(KeyboardUI.T);
  gui := NEW(GraphicalUI.T);
  teletypeKeyboard := NEW(CharReader.T);
  tapeReader := NEW(CharReader.T);

  (* the thread IDs *)
  prcThread : Thread.T;
  kbdThread : Thread.T;
  guiThread : Thread.T;
  teletypeKeyboardThread : Thread.T;
  tapeReaderThread : Thread.T;

  (* filenames for the reader/punch stream connections for passing to the *)
  (* GUI, to use to set up the display *)
  readerFileName : TEXT := NIL;
  punchFileName : TEXT := NIL;

  (* filenames for the memory image files *)
  memSourceFileName : TEXT := NIL;
  memSinkFileName : TEXT := NIL;

  file : Rd.T := NIL;

  (* buffer for use by the GUI *)
  ttbuffer := NEW(Buffer.T).init();

  (* variables to use to pass information to initialisers *)
  ttreader : CharRd.T;
  ttwriter : Wr.T;

  wrFilter : CharWriter.Filter;

  (* array of actions for the [AutoExec] part of the initialisation file *)
  autoExecInitArray := ARRAY [0..4] OF Init.Action {
      Init.Action { "InstructionLimit", InitInstructionLimit },
      Init.Action { "StartAddress", InitStartAddress },
      Init.Action { "SwitchRegister", InitSwitchRegister },
      Init.Action { "AutoExec", InitAutoExec },
      Init.Action { "UserInterface", InitUserInterface } };

  (* array of actions for the [IOFiles] part of the initialisation file *)
  IOInitArray := ARRAY [0..3] OF Init.Action {
      Init.Action { "Keyboard", InitKeyboard },
      Init.Action { "Teletype", InitTeletype },
      Init.Action { "Reader", InitReader },
      Init.Action { "Punch", InitPunch } };

  (* array of actions for the [Memory] part of the initialisation file *)
  memoryInitArray := ARRAY [0..1] OF Init.Action {
      Init.Action { "InFile", SetMemSourceFileName },
      Init.Action { "OutFile", SetMemSinkFileName } };


(* --------------- PROCEDURES FOR [AutoExec] ------------------------------ *)

(* ----------------------------------------------------------------------
   Procedure : InitAutoExec

   Initialise the automatic execution flag
   ---------------------------------------------------------------------- *)
PROCEDURE InitAutoExec(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    processor.autoExec := Scan.Bool(value);
  END InitAutoExec;


(* ----------------------------------------------------------------------
   Procedure : InitInstructionLimit

   Initialise the maximum number of instructions allowed per run
   ---------------------------------------------------------------------- *)
PROCEDURE InitInstructionLimit(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    processor.instructionLimit := Scan.Int(value);
  END InitInstructionLimit;


(* ----------------------------------------------------------------------
   Procedure : InitStartAddress
   
   Initialise the value of the PC
   ---------------------------------------------------------------------- *)
PROCEDURE InitStartAddress(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    processor.cpu.PC := Scan.Int(value, 8);
  END InitStartAddress;


(* ----------------------------------------------------------------------
   Procedure : InitSwitchRegister
   
   Initialise the value of the SR
   ---------------------------------------------------------------------- *)
PROCEDURE InitSwitchRegister(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    processor.cpu.SR := Scan.Int(value, 8);
  END InitSwitchRegister;


(* ----------------------------------------------------------------------
   Procedure : InitUserInterface

   Initialise the flag whether to start a user interface
   ---------------------------------------------------------------------- *)
PROCEDURE InitUserInterface(<* UNUSED *> name : TEXT; value : TEXT) =
  PROCEDURE ToLowerCase(source : TEXT) : TEXT =
    VAR
      result : TEXT := "";
    BEGIN
      FOR i := 0 TO Text.Length(source) - 1 DO
        result:= result & Text.FromChar(ASCII.Lower[Text.GetChar(source,i)]);
      END;
      RETURN result;
    END ToLowerCase;
  BEGIN
    value := ToLowerCase(value);
    IF Text.Equal(value, "none") THEN
      processor.userInterface := Processor.UIType.None;
    ELSIF Text.Equal(value, "graphical") THEN
      processor.userInterface := Processor.UIType.Graphical;
    ELSIF Text.Equal(value, "keyboard") THEN
      processor.userInterface := Processor.UIType.Keyboard;
    ELSE
      (*Diag.Message(Diag.T.Error, "invalid [AutoExec]UserInterface type:" &
                                 " using keyboard");*)
      processor.userInterface := Processor.UIType.Keyboard;
    END;      
  END InitUserInterface;



(* --------------- PROCEDURES FOR [IOFiles] ------------------------------ *)

(* ----------------------------------------------------------------------
   Procedure : InitKeyboard

   Initialise the keyboard input by attaching to the given file
   ---------------------------------------------------------------------- *)
PROCEDURE InitKeyboard(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    IF processor.userInterface = Processor.UIType.Graphical THEN
      (*Diag.Message(Diag.T.Control, "On screen input overrides specified file");*)
    ELSE
      EVAL GeneralUI.ConnectInput(GeneralUI.CharStreamDevice.Teletype, value);
    END;
  END InitKeyboard;

(* ----------------------------------------------------------------------
   Procedure : InitPunch

   Initialise the punch output by attaching to the given file.
   If successful, then store filename to pass on to the GUI
   ---------------------------------------------------------------------- *)
PROCEDURE InitPunch(<* UNUSED *> name : TEXT; value : TEXT) =
  VAR
    result : GeneralUI.StreamResult;
  BEGIN
    result := GeneralUI.ConnectOutput(GeneralUI.CharStreamDevice.Tape, value);

    IF result = GeneralUI.StreamResult.Success THEN
      punchFileName := value;
    END;
  END InitPunch;

(* ----------------------------------------------------------------------
   Procedure : InitReader

   Initialise the reader input by attaching to the given file
   If successful, then store filename to pass on to the GUI
   ---------------------------------------------------------------------- *)
PROCEDURE InitReader(<* UNUSED *> name : TEXT; value : TEXT) =
  VAR
    result : GeneralUI.StreamResult;
  BEGIN
    result := GeneralUI.ConnectInput(GeneralUI.CharStreamDevice.Tape, value);

    IF result = GeneralUI.StreamResult.Success THEN
      readerFileName := value;
    END;
  END InitReader;

(* ----------------------------------------------------------------------
   Procedure : InitTeletype

   Initialise the teletype output by attaching to the given file
   ---------------------------------------------------------------------- *)
PROCEDURE InitTeletype(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    IF processor.userInterface = Processor.UIType.Graphical THEN
      (*Diag.Message(Diag.T.Control,"On screen output overrides specified file");*)
    ELSE
      EVAL GeneralUI.ConnectOutput(GeneralUI.CharStreamDevice.Teletype, value);
    END;
  END InitTeletype;



(* --------------- PROCEDURES FOR [Memory] ------------------------------- *)

(* ----------------------------------------------------------------------
   Procedure : SetMemSourceFileName

   Invoked if InFile occurs in the [Memory] section of the initialisation
   file. Uses parameter to specify file from which to load the core memory
   ---------------------------------------------------------------------- *)
PROCEDURE SetMemSourceFileName(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    memSourceFileName := value;
  END SetMemSourceFileName;


(* ----------------------------------------------------------------------
   Procedure : SetMemSinkFileName

   Invoked if OutFile occurs in the [Memory] section of the initialisation
   file. Uses parameter to specify file to which to save the core memory
   ---------------------------------------------------------------------- *)
PROCEDURE SetMemSinkFileName(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    memSinkFileName := value;
  END SetMemSinkFileName;



BEGIN
  (* ---------------------------------------------------------------- *)
  (*              INITIALISE THE EMULATOR                             *)
  (* ---------------------------------------------------------------- *)
  (* we use so many interesting variables in this thread that we need a *)
  (* bigger stack size *)
  Thread.MinDefaultStackSize(8192);

  (* system initialisation *)
  (* read the names of the memory files to use if any have been specified *)
  Init.Scan("Memory", memoryInitArray);

  (* all encased in TRY...FINALLY to ensure that diagnostics file properly *)
  (* closed in event of an error *)
  TRY    
    (* if there is a memory file to load then do so *)
    TRY
      IF memSourceFileName # NIL THEN
        file := FileRd.Open(memSourceFileName)
      END;

      EVAL processor.init(file);

    EXCEPT
    | Memory.Failure =>
        (*Diag.Message(Diag.T.Control, "Memory initialisation failed\n");*)
        RAISE AbnormalExit;
    | OSError.E =>
        (*Diag.Message(Diag.T.Control,"Memory initialisation file not found\n");*)
        RAISE AbnormalExit;
    END;

    (* initialise the processor according to the [AutoExec] information *)
    Init.Scan("AutoExec", autoExecInitArray);

    (* initialise the thread closures for the teletype keyboard and the *)
    (* tape reader *)
    EVAL teletypeKeyboard.init(processor.cpu, teletype := TRUE);
    EVAL tapeReader.init(processor.cpu, teletype := FALSE);

    (* initialise the tape and teletype objects - supplying the refs to *)
    (* the appropriate threads - the assignment of readeers and writers *)
    (* to the teletype comes later *)
    (* initialise the tape reader with no streams to read or write *)
    (* output to file, so use default filter *)
    EVAL processor.bus.tape.csInit(tapeReader, processor.cpu,teletype:= FALSE,
                                   filter := NIL);
    processor.bus.tape.reader.reader := NIL;

    (* if output to GUI then use appropriate output filter *)
    IF processor.userInterface = Processor.UIType.Graphical THEN
      wrFilter := GraphicalUI.WrFilter
    ELSE
      wrFilter := NIL
    END;

    EVAL processor.bus.teletype.csInit(teletypeKeyboard, processor.cpu,
                   teletype := TRUE, filter := wrFilter);

    (* whether or not a user interface is required, it must be set *)
    (* up with pointers to cpu and bus for use by the GeneralUI module, *)
    (* which handles the connection and disconnection of file streams *)
    IF processor.userInterface = Processor.UIType.None OR
       processor.userInterface = Processor.UIType.Keyboard THEN
      keyboard.cpu := processor.cpu;       (* the cpu data object *)
      keyboard.bus := processor.bus;       (* the bus data object *)
    ELSE
      gui.cpu := processor.cpu;       (* the cpu data object *)
      gui.bus := processor.bus;       (* the bus data object *)
    END;

    (* initialise the GeneralUI module *)
    (* NB calling this before initialising the keyboard/gui object is
       potentially dangerous, but is done so that the [IOFiles] processing
       can occur before the setting up of the user interface, so that
       the GUI in particular receives the tape stream filenames
    *)
    IF processor.userInterface = Processor.UIType.None OR
       processor.userInterface = Processor.UIType.Keyboard THEN
      GeneralUI.ModuleInit(keyboard);
    ELSE
      GeneralUI.ModuleInit(gui);
    END;

    (* initialise startup file streams *)
    Init.Scan("IOFiles", IOInitArray);

    (* if a user interface is required then initialise it *)
    IF processor.userInterface = Processor.UIType.Keyboard THEN
      (* call the initialisation routine to set up keyboard specifics *)
      EVAL keyboard.init(Stdio.stdin);
    END;

    IF processor.userInterface = Processor.UIType.Graphical THEN
      (* initialise GUI thread *)
      (* 'buffer' is the buffer into which to put the characters typed on *)
      (* the typescript acting as the teletype *)
      EVAL gui.init(readerFileName, punchFileName, ttbuffer);
    END;

    (* if using the graphical user interface then assign the readers *)
    (* and writers associated with the FormsVBT form *)
    (* otherwise assign nothing at this stage *)
    IF processor.userInterface = Processor.UIType.Graphical THEN
      ttreader := NEW(BufferRd.T);
      NARROW(ttreader, BufferRd.T).connect(ttbuffer);
      WITH ts = FormsVBT.GetVBT(gui.form, "teletype") DO
        ttwriter := TypescriptVBT.GetWr(NARROW(ts, TypescriptVBT.T));
      END;
      processor.bus.teletype.reader.reader := ttreader;
      processor.bus.teletype.writer.connect(ttwriter);
    END;

    (* issue CAF command *)
    processor.bus.command(processor.cpu, 8_007);

    (* issue KCC command *)
    processor.bus.command(processor.cpu, 8_032);

    (* ----------------------------------------------------------------- *)
    (* ----------------- START THE EMULATOR ---------------------------- *)

    (*             ( by starting the threads running )                   *)

    (* ----------------------------------------------------------------- *)
    teletypeKeyboardThread := Thread.Fork(teletypeKeyboard);
    tapeReaderThread := Thread.Fork(tapeReader);
    prcThread := Thread.Fork(processor);

    (* if user interface required then start it up *)
    IF processor.userInterface = Processor.UIType.Graphical THEN
      guiThread := Thread.Fork(gui);
    END;

    IF processor.userInterface = Processor.UIType.Keyboard THEN
      kbdThread := Thread.Fork(keyboard);
    END;

    (* ----------------------------------------------------------------- *)
    (*                 WAIT UNTIL FINISHED                               *)
    (* ----------------------------------------------------------------- *)
    (* wait until quit is signalled *)
    processor.cpu.quit.wait();


    (* ----------------------------------------------------------------- *)
    (*              CLOSE DOWN GRACEFULLY                                *)
    (* ----------------------------------------------------------------- *)

    (* alert the threads to stop *)
    Thread.Alert(prcThread);
    (* if user interface present then alert it *)
    IF processor.userInterface = Processor.UIType.Graphical THEN
      Thread.Alert(guiThread);
    END;

    IF processor.userInterface = Processor.UIType.Keyboard THEN
      Thread.Alert(kbdThread);
    END;

    Thread.Alert(teletypeKeyboardThread);
    Thread.Alert(tapeReaderThread);


    (* wait for the threads to close gracefully *)
    EVAL Thread.Join(prcThread);
    (* if user interface present then close it *)
    IF processor.userInterface = Processor.UIType.Graphical THEN
      EVAL Thread.Join(guiThread);
    END;

    IF processor.userInterface = Processor.UIType.Keyboard THEN
      EVAL Thread.Join(kbdThread);
    END;

    EVAL Thread.Join(teletypeKeyboardThread);
    EVAL Thread.Join(tapeReaderThread);

    (* disconnect the files from these threads - side effect of flushing *)
    (* output files and closing all the files involved *)
    EVAL GeneralUI.DisconnectInput(GeneralUI.CharStreamDevice.Tape);
    EVAL GeneralUI.DisconnectOutput(GeneralUI.CharStreamDevice.Tape);

    IF NOT processor.userInterface = Processor.UIType.Graphical THEN
      EVAL GeneralUI.DisconnectInput(GeneralUI.CharStreamDevice.Teletype);
      EVAL GeneralUI.DisconnectOutput(GeneralUI.CharStreamDevice.Teletype);
    END;

    (* dump memory to output file if specified *)
    TRY
      IF memSinkFileName # NIL THEN
        WITH file = FileWr.Open(memSinkFileName) DO
          processor.cpu.memory.dump(file);
        END;
      END;
    EXCEPT
    | Memory.Failure =>
        (*Diag.Message(Diag.T.Control, "Memory dump failed\n");*)
        RAISE AbnormalExit;
    | OSError.E =>
        (*Diag.Message(Diag.T.Control, "Could not open memory dump file\n");*)
        RAISE AbnormalExit;
    END;

  FINALLY
    Diag.Close();
  END;

END Control.











