(* ----------------------------------------------------------------------
   $Id: GeneralUI.i3,v 1.12 1995/06/24 21:56:05 sgp13 Exp $

   This interface provides all the procedures which are used by both the
   keyboard and graphical user interfaces. The initialisation procedure is
   required for the setting up of a reference to the object which contains
   all the information about the system. This is an undesirable way of doing
   things but appears to be the only way because, when the procedures are
   invoked from the graphical UI there is no provision for passing a parameter
   which points to the information object.


   ----------------------------------------------------------------------

   $Log: GeneralUI.i3,v $
# Revision 1.12  1995/06/24  21:56:05  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.11  1995/06/23  07:54:05  sgp13
# Update forced
#
# Revision 1.10  1995/06/22  14:02:05  sgp13
# Update forced
#
# Revision 1.9  1995/06/19  14:11:47  sgp13
# Update forced
#
# Revision 1.8  1995/03/23  12:08:48  sgp13
# Update forced
#
# Revision 1.7  1995/03/22  13:14:27  sgp13
# Altered for graphical user interface:
#
# 1) see Continue procedure
# 2) Connect / Disconnect altered to return result
#
# Revision 1.6  1995/03/14  12:23:49  sgp13
# Update forced.
#
# Revision 1.5  1995/03/07  23:20:00  sgp13
# Connect and Disconnet procedures added.
#
# Revision 1.4  1995/02/24  22:23:53  sgp13
# The data object is now a Processor object containing CPU and Bus
#
# Revision 1.3  1995/01/23  12:53:53  sgp13
# SingleStep now clears flag before checking it to force a cycle to occur.
#
# Revision 1.2  1995/01/20  13:13:20  sgp13
# Function of Halt command corrected by clearing the endOfSingleCycle before
# checking it.
#
# Revision 1.1  1995/01/19  09:25:58  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
INTERFACE GeneralUI;

IMPORT Thread, Wr;
IMPORT FormsVBT, VBT;
IMPORT Bus, CPU;

TYPE
  T = Thread.Closure OBJECT
    cpu : CPU.T;             (* the CPU object *)
    bus : Bus.T;             (* the bus and objects attached to it *)
    statusUpdater : StatusUpdater;
    displayType := DisplayType.MD;
  END;

  StatusUpdater = PROCEDURE () RAISES { Wr.Failure, Thread.Alerted };
  DisplayType = { State, Status, AC, MD, MQ, Bus };
  CharStreamDevice = { Teletype, Tape };

  (* type to convey information about the result of a Connect/Disconnect *)
  (* operation *)
  StreamResult = { Success, FileFailure, AlreadyConnected, NotConnected };

(* ----------------------------------------------------------------------
   Procedure : AddressLoad

   Load address into PC from switch register.
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE AddressLoad(form : FormsVBT.T := NIL; event : TEXT := NIL;
                      data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : Clear

   Clear AC, L, interrupts, I/O flags
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Clear(form : FormsVBT.T := NIL; event : TEXT := NIL;
                data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : ConnectInput

   Connect the specified file to the specified input device
   ---------------------------------------------------------------------- *)
PROCEDURE ConnectInput(device : CharStreamDevice; filename : TEXT)
  : StreamResult;


(* ----------------------------------------------------------------------
   Procedure : ConnectOutput

   Connect the specified file to the specified output device
   ---------------------------------------------------------------------- *)
PROCEDURE ConnectOutput(device : CharStreamDevice; filename : TEXT)
  : StreamResult;


(* ----------------------------------------------------------------------
   Procedure : Continue

   Start free-running execution from current address
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Continue(form : FormsVBT.T := NIL; event : TEXT := NIL;
                   data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : Deposit

   Deposit value from switch register, advance address
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Deposit(form : FormsVBT.T := NIL; event : TEXT := NIL;
                  data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : DisconnectInput

   Disconnect the specified input device
   ---------------------------------------------------------------------- *)
PROCEDURE DisconnectInput(device : CharStreamDevice) : StreamResult;


(* ----------------------------------------------------------------------
   Procedure : DisconnectOutput

   Disconnect the specified output device
   ---------------------------------------------------------------------- *)
PROCEDURE DisconnectOutput(device : CharStreamDevice) : StreamResult;


(* ----------------------------------------------------------------------
   Procedure : Examine

   Examine value at current address, advance address by one
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Examine(form : FormsVBT.T := NIL; event : TEXT := NIL;
                  data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : ExtdAddressLoad

   Load the instruction field register from the 6-8 bits of the switch
   register. Load the data field register from bits 9-11 of SR.
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE ExtdAddressLoad(form : FormsVBT.T := NIL; event : TEXT := NIL;
           data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : Halt

   Stop free-running execution at start of next cycle
   Valid when machine running.
   ---------------------------------------------------------------------- *)
PROCEDURE Halt(form : FormsVBT.T := NIL; event : TEXT := NIL;
           data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : ModuleInit

   Initialise module by passing a pointer to the machine information object
   ---------------------------------------------------------------------- *)
PROCEDURE ModuleInit(init : T);


(* ----------------------------------------------------------------------
   Procedure : Quit

   Quit program.
   Valid at any time.
   ---------------------------------------------------------------------- *)
PROCEDURE Quit(form : FormsVBT.T := NIL; event : TEXT := NIL;
               data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : SingleStep

   If running, stop at start of next cycle,
   otherwise execute single processor cycle
   Valid at any time.
   ---------------------------------------------------------------------- *)
PROCEDURE SingleStep(form : FormsVBT.T := NIL; event : TEXT := NIL;
           data : REFANY := NIL; time : VBT.TimeStamp := 0);


(* ----------------------------------------------------------------------
   Procedure : SWLine

   Toggle the state of the SW line
   Valid at any time.
   ---------------------------------------------------------------------- *)
PROCEDURE SWLine(form : FormsVBT.T := NIL; event : TEXT := NIL;
           data : REFANY := NIL; time : VBT.TimeStamp := 0);


END GeneralUI.


