(* ----------------------------------------------------------------------
   $Id: GeneralUI.m3,v 1.12 1995/06/24 21:55:44 sgp13 Exp $

   This interface provides all the procedures which are used by both the
   keyboard and graphical user interfaces. The initialisation procedure is
   required for the setting up of a reference to the object which contains
   all the information about the system. This is an undesirable way of doing
   things but appears to be the only way because, when the procedures are
   invoked from the graphical UI there is no provision for passing a parameter
   which points to the information object.

   ----------------------------------------------------------------------

   $Log: GeneralUI.m3,v $
# Revision 1.12  1995/06/24  21:55:44  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.11  1995/06/23  07:53:44  sgp13
# Registers are now accessed directly
#
# Revision 1.10  1995/06/22  14:01:32  sgp13
# Memory made direct access
#
# Revision 1.9  1995/06/19  14:11:23  sgp13
# Changes to allow characters to be read from buffer or file stream.
#
# Revision 1.8  1995/03/23  12:08:27  sgp13
# Patched so that it does not attempt to close stdin or stdout
#
# Revision 1.7  1995/03/22  13:13:40  sgp13
# Altered for graphical user interface:
#
# 1) see Continue procedure
# 2) Connect / Disconnect altered to return result
#
# Revision 1.6  1995/03/14  12:23:21  sgp13
# Deadlock on Halt if already halted removed.
#
# Revision 1.5  1995/03/07  23:20:23  sgp13
# Connect and Disconnect procedures added.
#
# Revision 1.4  1995/02/24  22:24:20  sgp13
# The data object is now a Processor object containing CPU and Bus
#
# Revision 1.3  1995/01/23  12:53:01  sgp13
# SingleStep now clears the endOfCycle flag before checking it to ensure that
# a cycle is forced.
#
# Revision 1.2  1995/01/20  13:12:38  sgp13
# Function of Halt command corrected by clearing the endOfSingleCycle before
# checking it.
#
# Revision 1.1  1995/01/19  09:26:09  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE GeneralUI;

IMPORT FileRd, FileWr, FormsVBT, OSError, Rd, Stdio, Thread, VBT, Wr;
IMPORT CPU, CharRd, Diag, Registers, StreamRd;
IMPORT CharWriter;

VAR
  info : T;       (* the machine information object *)

<* FATAL Registers.OutOfRange *>

(* ----------------------------------------------------------------------
   Procedure : AddressLoad

   Load address into PC from switch register.
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE AddressLoad(<* UNUSED *> form : FormsVBT.T := NIL;
                      <* UNUSED *> event : TEXT := NIL;
                      <* UNUSED *> data : REFANY := NIL;
                      <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF NOT runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=AddressLoad");*)
        WITH address = cpu.SR DO
          (* set new address *)
          cpu.PC := address;
          (* read new memory location *)
          cpu.MB := cpu.EM_IFptr[address];
        END;
        (* status display *)
        info.statusUpdater();
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=AddressLoad:" &
                                       "failed-machine running");*)
      END;
    END;
  END AddressLoad;


(* ----------------------------------------------------------------------
   Procedure : Clear

   Clear AC, L, interrupts, I/O flags
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Clear(<* UNUSED *> form : FormsVBT.T := NIL;
                <* UNUSED *> event : TEXT := NIL;
                <* UNUSED *> data : REFANY := NIL;
                <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF NOT runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=Clear");*)
        info.bus.command(cpu, 8_007);
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=Clear:" &
                                       "failed-machine running");*)
      END;
    END;
  END Clear;


(* ----------------------------------------------------------------------
   Procedure : ConnectInput

   Connect the specified file to the specified input device
   ---------------------------------------------------------------------- *)
PROCEDURE ConnectInput(device : CharStreamDevice; filename : TEXT)
  : StreamResult =
  VAR
    file : Rd.T;
    reader : StreamRd.T;
  BEGIN
    TRY
      (* open file *)
      file := FileRd.Open(filename);
      (* create reader object to enclose the file *)
      reader := NEW(StreamRd.T);
      reader.connect(file);

      CASE device OF
      | CharStreamDevice.Teletype =>
            info.bus.teletype.reader.connect(reader);
            (*Diag.Message(Diag.T.GeneralUI, "Connect: " & filename &
                         " connected to teletype input");*)
      | CharStreamDevice.Tape =>
            info.bus.tape.reader.connect(reader);
            (*Diag.Message(Diag.T.GeneralUI, "Connect: " & filename &
                         " connected to tape input");*)
      END;

      RETURN StreamResult.Success;
    EXCEPT
    | CharRd.AlreadyConnected =>
          (*Diag.Message(Diag.T.GeneralUI, "Connect:"
              &" there is already a file connected to specified input");*)
          Rd.Close(file);
          RETURN StreamResult.AlreadyConnected;
    | OSError.E => (*Diag.Message(Diag.T.GeneralUI, "Connect: file open failed");*)
          RETURN StreamResult.FileFailure;
    END;
  END ConnectInput;


(* ----------------------------------------------------------------------
   Procedure : ConnectOutput

   Connect the specified file to the specified output device

   Returns TRUE or FALSE
   ---------------------------------------------------------------------- *)
PROCEDURE ConnectOutput(device : CharStreamDevice; filename : TEXT)
  : StreamResult =
  VAR
    file : Wr.T;
  BEGIN
    TRY
      file := FileWr.Open(filename);

      CASE device OF
      | CharStreamDevice.Teletype =>
            info.bus.teletype.writer.connect(file);
            (*Diag.Message(Diag.T.GeneralUI, "Connect: " & filename &
                         " connected to teletype output");*)
      | CharStreamDevice.Tape =>
            info.bus.tape.writer.connect(file);
            (*Diag.Message(Diag.T.GeneralUI, "Connect: " & filename &
                         " connected to tape output");*)
      END;

      RETURN StreamResult.Success;
    EXCEPT
    | CharWriter.AlreadyConnected =>
          (*Diag.Message(Diag.T.GeneralUI, "Connect:"
              &" there is already a file connected to specified output");*)
          Wr.Close(file);
          RETURN StreamResult.AlreadyConnected;
    | OSError.E => (*Diag.Message(Diag.T.GeneralUI, "Connect: file open failed");*)
                   RETURN StreamResult.FileFailure;
    END;
  END ConnectOutput; 


(* ----------------------------------------------------------------------
   Procedure : Continue

   Start free-running execution from current address
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Continue(<* UNUSED *> form : FormsVBT.T := NIL;
                   <* UNUSED *> event : TEXT := NIL;
                   <* UNUSED *> data : REFANY := NIL;
                   <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF NOT runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=Continue");*)
        cpu.run.continuous := TRUE;
        cpu.run.request.signal();

       (* --------------------------------------------------------------------
          this bit is to ensure a response from Keyboard UI to say that
          we've started running, but is incompatible with Graphical UI because
          there is a limit to the amount of processing which should be done
          in response to an event

          cpu.endOfFreeRunCycle.wait();
          info.statusUpdater();

          ------------------------------------------------------------------ *)
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=Continue:" &
                                     "failed-machine running");*)
      END;
    END;
  END Continue;
      


(* ----------------------------------------------------------------------
   Procedure : Deposit

   Deposit value from switch register, advance address. NB value also written
   to MB so that the actual function of the PDP8/e is echoed correctly
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Deposit(<* UNUSED *> form : FormsVBT.T := NIL;
                  <* UNUSED *> event : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF NOT runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=Deposit");*)
        WITH value = cpu.SR, PC = cpu.PC DO
          cpu.MB := value;
          cpu.EM_IFptr[PC] := value;
          IF PC = LAST(CPU.Register_12) THEN
            PC := 0;
          ELSE
            INC(PC);
          END;
        END;
        info.statusUpdater();
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=Deposit:" &
                                       "failed-machine running");*)
      END;
    END;
  END Deposit;


(* ----------------------------------------------------------------------
   Procedure : DisconnectInput

   Disconnect the specified input device
   ---------------------------------------------------------------------- *)
PROCEDURE DisconnectInput(device : CharStreamDevice) : StreamResult =
  VAR
    file : Rd.T := NIL;
    reader : StreamRd.T;
  BEGIN
    CASE device OF
    | CharStreamDevice.Teletype =>
                 reader := info.bus.teletype.reader.disconnect();
    | CharStreamDevice.Tape =>
                 reader := info.bus.tape.reader.disconnect();
    END;

    IF reader # NIL THEN
      file := reader.disconnect();
    END;

    IF file # NIL THEN
      (*Diag.Message(Diag.T.GeneralUI, "Disconnect: successful");*)
      (* do not attempt to close Stdio.stdin *)
      IF file # Stdio.stdin THEN
        Rd.Close(file);
      END;
      RETURN StreamResult.Success;
    ELSE
      (*Diag.Message(Diag.T.GeneralUI, "Disconnect: no file connected");*)
      RETURN StreamResult.NotConnected;
    END;

  END DisconnectInput;


(* ----------------------------------------------------------------------
   Procedure : DisconnectOutput

   Disconnect the specified output device
   ---------------------------------------------------------------------- *)
PROCEDURE DisconnectOutput(device : CharStreamDevice) : StreamResult =
  VAR
    file : Wr.T;
  BEGIN
    CASE device OF
    | CharStreamDevice.Teletype =>
                 file := info.bus.teletype.writer.disconnect();
    | CharStreamDevice.Tape =>
                 file := info.bus.tape.writer.disconnect();
    END;

    IF file # NIL THEN
      (*Diag.Message(Diag.T.GeneralUI, "Disconnect: successful");*)
      Wr.Flush(file);
      (* do not attempt to close Stdio.stdout *)
      IF file # Stdio.stdout THEN
        Wr.Close(file);
      END;
      RETURN StreamResult.Success;
    ELSE
      (*Diag.Message(Diag.T.GeneralUI, "Disconnect: no file connected");*)
      RETURN StreamResult.NotConnected;
    END;

  END DisconnectOutput; 


(* ----------------------------------------------------------------------
   Procedure : Examine

   Examine value at current address, advance address by one.
   Examine works by reading into MB
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE Examine(<* UNUSED *> form : FormsVBT.T := NIL;
                  <* UNUSED *> event : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF NOT runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=Examine");*)
        WITH PC = cpu.PC DO
          cpu.MB := cpu.EM_IFptr[PC];
          IF PC = LAST(CPU.Register_12) THEN
            PC := 0;
          ELSE
            INC(PC);
          END;
        END;
        info.statusUpdater();
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=Examine:" &
                                     "failed-machine running");*)
      END;
    END;
  END Examine;


(* ----------------------------------------------------------------------
   Procedure : ExtdAddressLoad

   Load the instruction field register from the 6-8 bits of the switch
   register. Load the data field register from bits 9-11 of SR.
   Valid when machine NOT running.
   ---------------------------------------------------------------------- *)
PROCEDURE ExtdAddressLoad(<* UNUSED *> form : FormsVBT.T := NIL;
                          <* UNUSED *> event : TEXT := NIL;
                          <* UNUSED *> data : REFANY := NIL;
                          <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF NOT runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=ExtdAddrLoad");*)
        WITH value = cpu.SR DO
          cpu.EM_IFnum := (value DIV 2_1000) MOD 2_1000;
          cpu.EM_IFptr := cpu.memory.mem[cpu.EM_IFnum];
          cpu.EM_DFnum := value MOD 2_1000;
          cpu.EM_DFptr := cpu.memory.mem[cpu.EM_DFnum];
        END;
        info.statusUpdater();
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=ExtdAddrLoad:" &
                                     "failed-machine running");*)
      END;
    END;
  END ExtdAddressLoad;


(* ----------------------------------------------------------------------
   Procedure : Halt

   Stop free-running execution at start of next cycle
   Valid when machine running.
   ---------------------------------------------------------------------- *)
PROCEDURE Halt(<* UNUSED *> form : FormsVBT.T := NIL;
               <* UNUSED *> event : TEXT := NIL;
               <* UNUSED *> data : REFANY := NIL;
               <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;

      IF runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=Halt");*)
        LOCK cpu.run DO
          cpu.endOfSingleCycle.clear();
          cpu.run.continuous := FALSE;
        END;

        (* wait for end of cycle signal and update display *)
        cpu.endOfSingleCycle.wait();
        info.statusUpdater();
      ELSE
        (*Diag.Message(Diag.T.GeneralUI, "command=Halt:" &
                                       "failed-machine not running");*)
      END;
    END;
  END Halt;


(* ----------------------------------------------------------------------
   Procedure : ModuleInit

   Initialise module by passing a pointer to the machine information object
   ---------------------------------------------------------------------- *)
PROCEDURE ModuleInit(init : T) =
  BEGIN
    info := init;
  END ModuleInit;


(* ----------------------------------------------------------------------
   Procedure : Quit

   Quit program.
   Valid at any time.
   ---------------------------------------------------------------------- *)
PROCEDURE Quit(<* UNUSED *> form : FormsVBT.T := NIL;
               <* UNUSED *> event : TEXT := NIL;
               <* UNUSED *> data : REFANY := NIL;
               <* UNUSED *> time : VBT.TimeStamp := 0) =
  BEGIN
    (*Diag.Message(Diag.T.GeneralUI, "command=Quit");*)
    info.cpu.quit.signal();
  END Quit;


(* ----------------------------------------------------------------------
   Procedure : SingleStep

   Execute single processor cycle.
   Valid if processor NOT running
   ---------------------------------------------------------------------- *)
PROCEDURE SingleStep(<* UNUSED *> form : FormsVBT.T := NIL;
                     <* UNUSED *> event : TEXT := NIL;
                     <* UNUSED *> data : REFANY := NIL;
                     <* UNUSED *> time : VBT.TimeStamp := 0) = 
  VAR
    runFlag : BOOLEAN;
  BEGIN
    WITH cpu = info.cpu DO
      LOCK cpu.run DO
        runFlag := cpu.run.running;
      END;
      
      IF runFlag THEN
        (*Diag.Message(Diag.T.GeneralUI, "command=SingleStep:" &
                                     "failed-machine running");*)
      ELSE
        TRY
          LOCK cpu.run DO
            cpu.endOfSingleCycle.clear();
            cpu.run.request.signal();
          END;
          cpu.endOfSingleCycle.wait();
        EXCEPT
        | Thread.Alerted =>
        END;
        info.statusUpdater();
      END;
    END;
  END SingleStep;


(* ----------------------------------------------------------------------
   Procedure : SWLine

   Toggle the state of the SW line
   Valid at any time.
   ---------------------------------------------------------------------- *)
PROCEDURE SWLine(<* UNUSED *> form : FormsVBT.T := NIL;
                 <* UNUSED *> event : TEXT := NIL;
                 <* UNUSED *> data : REFANY := NIL;
                 <* UNUSED *> time : VBT.TimeStamp := 0) =
  BEGIN
    (*Diag.Message(Diag.T.GeneralUI, "command=SWLine");*)
    info.cpu.SW := NOT(info.cpu.SW);
    info.statusUpdater();
  END SWLine;

BEGIN
END GeneralUI.





