(* ----------------------------------------------------------------------
   $Id: GraphicalUI.m3,v 1.7 1995/06/24 21:56:57 sgp13 Exp $

   Graphical user interface implementation.
   ----------------------------------------------------------------------

   $Log: GraphicalUI.m3,v $
# Revision 1.7  1995/06/24  21:56:57  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.6  1995/06/23  07:52:54  sgp13
# Registers are now accessed directly
#
# Revision 1.5  1995/06/20  10:09:50  sgp13
# Changes due to IntBus version 1.2
#
# Revision 1.4  1995/06/19  14:16:17  sgp13
# Typescript added for teletype I/O.
#
# Revision 1.3  1995/03/30  10:19:36  sgp13
# Altered to make more realistic
#
# Revision 1.2  1995/03/23  12:10:35  sgp13
# Code added to initialise the state of the filename display.
#
# Revision 1.1  1995/03/22  13:08:09  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE GraphicalUI;

IMPORT ASCII, FVTypes, FormsVBT, KeyboardKey, Rd, Rsrc, Text, TextPort, Thread;
IMPORT Trestle, TrestleComm, TypescriptVBT, VBT;

IMPORT Bits, Buffer, Diag;
IMPORT GeneralUI;

IMPORT GraphicalUIBundle;
IMPORT GUISwitchRegister;

TYPE
  ConnectProc = PROCEDURE(device : GeneralUI.CharStreamDevice; 
                          filename: TEXT) : GeneralUI.StreamResult;
  DisconnectProc = PROCEDURE(device : GeneralUI.CharStreamDevice)
                      : GeneralUI.StreamResult;

  StreamInfo = RECORD
    name           : TEXT;                       (* the name used in the .fv *)
    device         : GeneralUI.CharStreamDevice; (* device on bus *)
    connectProc    : ConnectProc;
    disconnectProc : DisconnectProc;
  END;

  (* this is to allow me to override the filter function of the Typescript *)
  (* with my own filter so that I can have the characters available as the *)
  (* the keys are pressed *)
  MyForm = FormsVBT.T OBJECT
  OVERRIDES
    realize := Realize;
  END;

CONST
  numDevices = 4;


VAR
  info : T;  (* store information about the system here *)
             (* to be accessible by the GUI *)

  (* the information about how to process each device *)
  device := ARRAY [1..numDevices] OF StreamInfo {
    StreamInfo { "keyboard", GeneralUI.CharStreamDevice.Teletype,
                 GeneralUI.ConnectInput, GeneralUI.DisconnectInput },
    StreamInfo { "teletype", GeneralUI.CharStreamDevice.Teletype,
                 GeneralUI.ConnectOutput, GeneralUI.DisconnectOutput },
    StreamInfo { "reader", GeneralUI.CharStreamDevice.Tape,
                 GeneralUI.ConnectInput, GeneralUI.DisconnectInput },
    StreamInfo { "punch", GeneralUI.CharStreamDevice.Tape,
                 GeneralUI.ConnectOutput, GeneralUI.DisconnectOutput } };


(* ----------------------------------------------------------------------
   Procedure : Connect

   Connect the specified file to the appropriate device.

   'name' is the name of the button pressed
   ---------------------------------------------------------------------- *)
PROCEDURE Connect(             form : FormsVBT.T := NIL;
                               name : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) = 
  VAR
    pos : INTEGER; (* position within string *)

    (* the name of the TypeIn for the filename *)
    typeInName : TEXT;

    (* the filename to connect *)
    filename : TEXT;

    done := FALSE;
    result : GeneralUI.StreamResult;
  BEGIN
    (* strip the button name down to the part before the '_' *)
    pos := Text.FindChar(name, '_');
    name := Text.Sub(name, 0, pos);
    typeInName := name & "_file";

    (* read filename *)
    filename := FormsVBT.GetText(form, typeInName);

    (* find the entry in the 'device type' array *)
    FOR i := 1 TO numDevices DO
      IF Text.Equal(device[i].name, name) THEN
        result := device[i].connectProc(device[i].device, filename);
        done := TRUE;
      END;
    END;

    IF done THEN
      CASE result OF
      | GeneralUI.StreamResult.Success =>
          (* if the connection was successful then: *)
          (* disable the 'connect' button and the TypeIn *)
          (* enable the 'disconnect' button *)
          FormsVBT.MakePassive(form, name & "_file");
          FormsVBT.MakeDormant(form, name & "_connect");
          FormsVBT.MakeActive(form, name & "_disconnect");

      | GeneralUI.StreamResult.FileFailure =>
          FormMessage(form, "could not open file");

      | GeneralUI.StreamResult.AlreadyConnected =>
          FormMessage(form, "file already connected");

      | GeneralUI.StreamResult.NotConnected => 
          FormMessage(form,"internal error: NotConnected returned to Connect");
          (*Diag.Message(Diag.T.Error, "internal: KeyboardUI.Connect " &
                       "received unexpected result NotConnected");*)
      END;

    ELSE
      (*Diag.Message(Diag.T.Error, "internal inconsistency: " &
                                  "device not found in Connect");*)
    END;

  END Connect;


(* ----------------------------------------------------------------------
   Procedure : CreateForm

   Create and initialise the FormsVBT form

   NB The "s0_zero" event should be the action to take when switch 0 is hit
      while it is in state zero: ie the transition function to take it _from_
      zero to _one_, ie BitZeroOn.
   ---------------------------------------------------------------------- *)
PROCEDURE CreateForm() : FormsVBT.T =
  BEGIN
    TRY
      (* this looks for the formsvbt file for the form, first in the *)
      (* current directory and then in the file bundled with the system *)
      (* when it was built. this is to allow for polishing the GUI *)
      (* without having to recompile the system *)
      WITH path = Rsrc.BuildPath(".", GraphicalUIBundle.Get()) DO
        (* the use of 'MyForm' enforces the implementation of my own *)
        (* 'realize' method to override the character filter on the *)
        (* Typescript's TextPort *)
        WITH form = NEW (MyForm).initFromRsrc("GraphicalUI.fv", path) DO
          (* switch register *)
          FormsVBT.AttachProc(form, "s0_zero", GUISwitchRegister.BitZeroOn);
          FormsVBT.AttachProc(form, "s0_one", GUISwitchRegister.BitZeroOff);
          FormsVBT.AttachProc(form, "s1_zero", GUISwitchRegister.BitOneOn);
          FormsVBT.AttachProc(form, "s1_one", GUISwitchRegister.BitOneOff);
          FormsVBT.AttachProc(form, "s2_zero", GUISwitchRegister.BitTwoOn);
          FormsVBT.AttachProc(form, "s2_one", GUISwitchRegister.BitTwoOff);
          FormsVBT.AttachProc(form, "s3_zero", GUISwitchRegister.BitThreeOn);
          FormsVBT.AttachProc(form, "s3_one", GUISwitchRegister.BitThreeOff);
          FormsVBT.AttachProc(form, "s4_zero", GUISwitchRegister.BitFourOn);
          FormsVBT.AttachProc(form, "s4_one", GUISwitchRegister.BitFourOff);
          FormsVBT.AttachProc(form, "s5_zero", GUISwitchRegister.BitFiveOn);
          FormsVBT.AttachProc(form, "s5_one", GUISwitchRegister.BitFiveOff);
          FormsVBT.AttachProc(form, "s6_zero", GUISwitchRegister.BitSixOn);
          FormsVBT.AttachProc(form, "s6_one", GUISwitchRegister.BitSixOff);
          FormsVBT.AttachProc(form, "s7_zero", GUISwitchRegister.BitSevenOn);
          FormsVBT.AttachProc(form, "s7_one", GUISwitchRegister.BitSevenOff);
          FormsVBT.AttachProc(form, "s8_zero", GUISwitchRegister.BitEightOn);
          FormsVBT.AttachProc(form, "s8_one", GUISwitchRegister.BitEightOff);
          FormsVBT.AttachProc(form, "s9_zero", GUISwitchRegister.BitNineOn);
          FormsVBT.AttachProc(form, "s9_one", GUISwitchRegister.BitNineOff);
          FormsVBT.AttachProc(form, "s10_zero", GUISwitchRegister.BitTenOn);
          FormsVBT.AttachProc(form, "s10_one", GUISwitchRegister.BitTenOff);
          FormsVBT.AttachProc(form, "s11_zero",GUISwitchRegister.BitElevenOn);
          FormsVBT.AttachProc(form, "s11_one", GUISwitchRegister.BitElevenOff);
          
          (* connect and disconnect procedures *)
          FormsVBT.AttachProc(form, "reader_connect", Connect);
          FormsVBT.AttachProc(form, "punch_connect", Connect);
          FormsVBT.AttachProc(form, "reader_disconnect", Disconnect);
          FormsVBT.AttachProc(form, "punch_disconnect", Disconnect);

          (* address load switches *)
          FormsVBT.AttachProc(form, "AddrLoad", GeneralUI.AddressLoad);
          FormsVBT.AttachProc(form, "ExtdAddrLoad", GeneralUI.ExtdAddressLoad);

          (* other function switches *)
          FormsVBT.AttachProc(form, "Clear", GeneralUI.Clear);
          FormsVBT.AttachProc(form, "Continue", GeneralUI.Continue);
          FormsVBT.AttachProc(form, "Examine", GeneralUI.Examine);
          FormsVBT.AttachProc(form, "Halt", GeneralUI.Halt);
          FormsVBT.AttachProc(form, "SingleStep", GeneralUI.SingleStep);
          FormsVBT.AttachProc(form, "Deposit", GeneralUI.Deposit);

          (* SW signal *)
          FormsVBT.AttachProc(form, "sw_zero", SWLineOn);
          FormsVBT.AttachProc(form, "sw_one", SWLineOff);

          (* Quit signal *)
          FormsVBT.AttachProc(form, "Quit", GeneralUI.Quit);

          (* display type choice radio button *)
          FormsVBT.AttachProc(form, "displaytype", DisplayType);

          (* message box *)
          FormsVBT.AttachProc(form, "messageok", CancelMessage);

          RETURN form;
        END;
      END;
    EXCEPT
    | Rsrc.NotFound, FormsVBT.Error, Rd.Failure => RAISE InitFailed;
    END;
  END CreateForm;


(* ----------------------------------------------------------------------
   Procedure : Disconnect

   Disconnect the specified device

   'name' is the name of the button pressed
   ---------------------------------------------------------------------- *)
PROCEDURE Disconnect(             form : FormsVBT.T := NIL;
                                  name : TEXT := NIL;
                     <* UNUSED *> data : REFANY := NIL;
                     <* UNUSED *> time : VBT.TimeStamp := 0) = 
  VAR
    pos : INTEGER; (* position within string *)
    done := FALSE;
    result : GeneralUI.StreamResult;
  BEGIN
    (* strip the button name down to the part before the '_' *)
    pos := Text.FindChar(name, '_');
    name := Text.Sub(name, 0, pos);

    (* find the entry in the 'device type' array *)
    FOR i := 1 TO numDevices DO
      IF Text.Equal(device[i].name, name) THEN
        result := device[i].disconnectProc(device[i].device);
        done := TRUE;
      END;
    END;

    IF done THEN
      CASE result OF
      | GeneralUI.StreamResult.Success =>
          (* disable the 'disconnect' button, enable the TypeIn and clear *)
          (* contents and enable the 'connect' button *)
          FormsVBT.MakeDormant(form, name & "_disconnect");
          FormsVBT.MakeActive(form, name & "_file");
          FormsVBT.PutText(form, name & "_file", "");
          FormsVBT.MakeActive(form, name & "_connect");

      | GeneralUI.StreamResult.FileFailure =>
           FormMessage(form,
                       "internal error: FileFailure returned to Disconnect\n");
           (*Diag.Message(Diag.T.Error, "internal: KeyboardUI.Disconnect " &
                        "received unexpected result FileFailure");*)

      | GeneralUI.StreamResult.AlreadyConnected =>
           FormMessage(form, 
                  "internal error: AlreadyConnected returned to Disconnect\n");
           (*Diag.Message(Diag.T.Error, "internal: KeyboardUI.Disconnect " &
                        "received unexpected result AlreadyConnected");*)

      | GeneralUI.StreamResult.NotConnected =>
           FormMessage(form, "no file connected to this stream");
      END;
    ELSE
      (*Diag.Message(Diag.T.Error, "internal inconsistency: " &
                                  "device not found in Disconnect");*)
    END;
  END Disconnect;


(* ----------------------------------------------------------------------
   Procedure : DisplayType

   Set the display type according to selection from the radio button
   group on the front panel
   ---------------------------------------------------------------------- *)
PROCEDURE DisplayType(             form : FormsVBT.T := NIL;
                      <* UNUSED *> event : TEXT := NIL;
                      <* UNUSED *> data : REFANY := NIL;
                      <* UNUSED *> time : VBT.TimeStamp := 0) = 
  CONST
    displayTypeName = ARRAY GeneralUI.DisplayType OF TEXT {
           "state", "status", "ac", "md", "mq", "null" };
  VAR
    choice : TEXT;
  BEGIN
    choice := FormsVBT.GetChoice(form, "displaytype");

    (* loop through each possible type and set according to that chosen *)
    FOR i := FIRST(GeneralUI.DisplayType) TO LAST(GeneralUI.DisplayType) DO
      IF Text.Equal(choice, displayTypeName[i]) THEN
        info.displayType := i;
        (*Diag.Message(Diag.T.GraphicalUI, "display type changed to: " & choice);*)
      END;
    END;

    info.statusUpdater();
  END DisplayType;

(* ----------------------------------------------------------------------
   Procedure : DoGraphicalUI

   Run the graphical user interface
   ---------------------------------------------------------------------- *)
PROCEDURE DoGraphicalUI(self : T) : REFANY RAISES {} =
  BEGIN
    TRY
      TRY
        Trestle.Install(self.form);
        WHILE NOT Thread.TestAlert() DO
          self.cpu.endOfFreeRunCycle.wait();
          self.statusUpdater();
        END;
      FINALLY
        Trestle.Delete(self.form);
      END;
    EXCEPT
    | TrestleComm.Failure => (*Diag.Message(Diag.T.Error, "Trestle failed");*)
                             self.cpu.quit.signal();
    | Thread.Alerted =>
    END;

    RETURN self;

  END DoGraphicalUI;


(* ----------------------------------------------------------------------
   Procedure : ObjectInit

   Initialise the graphical user interface:

   assign a status update procedure
   initialise a FormsVBT form
   initialise the switch register module
   store the pointer to 'self' for access by FormsVBT attached procedures

   set up the filename fields depending on whether files are connected to
   the reader and/or punch: ie:
       if there is a file connected then display its name, disable the
       TypeIn field and connect button and enable the disconnect button

       if there is not a file connected, ensable the connect button, and
       the TypeIn field and disable the disconnect button

   ---------------------------------------------------------------------- *)
PROCEDURE ObjectInit(self: T; readerFileName: TEXT; punchFileName: TEXT;
                     buffer: Buffer.T) : T RAISES { InitFailed } =
  BEGIN
    self.statusUpdater := StatusUpdater;
    self.form := CreateForm();
    self.buffer := buffer;               (* buffer for use by MyFilter *)
    GUISwitchRegister.ModuleInit(self.cpu);

    (* store the object for reference by FormsVBT attached procedures *)
    info := self;

    SetUpStreamDisplay("reader", readerFileName);
    SetUpStreamDisplay("punch", punchFileName);

    RETURN self;
  END ObjectInit;


(* ----------------------------------------------------------------------
   Procedure : SetUpStreamDisplay


   ---------------------------------------------------------------------- *)
PROCEDURE SetUpStreamDisplay(stream : TEXT; filename : TEXT) =
  BEGIN
    IF filename = NIL OR Text.Equal(filename, "") THEN
      (* no file connected *)
      FormsVBT.MakeActive(info.form, stream & "_file");
      FormsVBT.MakeActive(info.form, stream & "_connect");
      FormsVBT.MakeDormant(info.form, stream & "_disconnect");
    ELSE
      (* file connected *)
      FormsVBT.PutText(info.form, stream & "_file", filename);
      FormsVBT.MakePassive(info.form, stream & "_file");
      FormsVBT.MakeDormant(info.form, stream & "_connect");
      FormsVBT.MakeActive(info.form, stream & "_disconnect");
    END;
  END SetUpStreamDisplay;

(* ----------------------------------------------------------------------
   Procedure : StatusUpdater

   Update the status on the display
   ---------------------------------------------------------------------- *)
PROCEDURE StatusUpdater() =
  CONST
    (* the names of the 'lights' on the display panel *)
    extdMemLightName = ARRAY [9..11] OF TEXT { "extdmem0", "extdmem1",
                                               "extdmem2" };
    memLightName = ARRAY [0..11] OF TEXT { "mem0", "mem1", "mem2", "mem3",
                                           "mem4", "mem5", "mem6", "mem7",
                                           "mem8", "mem9", "mem10", "mem11" };
    displayLightName = ARRAY [0..11] OF TEXT {
           "display0", "display1", "display2", "display3",
           "display4", "display5", "display6", "display7",
           "display8", "display9", "display10", "display11" };

  VAR
    memBits : Bits.IntArray;
    displayBits : Bits.IntArray;
  BEGIN
    (* extended memory display *)
    (* explode the bits to numbers *)
    memBits := Bits.ExplodeToInts(info.cpu.EM_IFnum);
    FOR i := 9 TO 11 DO
      FormsVBT.PutInteger(info.form, extdMemLightName[i], memBits[i]);
    END;

    WITH cpu = info.cpu, interrupt = cpu.interrupt DO
      (* explode memory address bits to numbers *)
      memBits := Bits.ExplodeToInts(cpu.PC);

      (* set up the array of bits to be read on the display line *)
      (* AC, MD, MQ are self-explanatory *)
      CASE info.displayType OF
      | GeneralUI.DisplayType.State =>
           (* very little of the state information makes any sense within *)
           (* the emulator - here we display 'SW' and the instruction code *)
           displayBits := Bits.ExplodeToInts(cpu.IR * 8_100);
           displayBits[8] := Bits.BoolToInt(cpu.SW);

      | GeneralUI.DisplayType.Status =>
           (* load the IF and DF values into the display array *)
           displayBits := Bits.ExplodeToInts(cpu.EM_IFnum * 8 +
                                             cpu.EM_DFnum);

           (* add the individual status bits *)
           displayBits[0] := Bits.BoolToInt(cpu.L); (* link register *)
           displayBits[1] := 0;              (* GT flag - unimplemented *)
           displayBits[2] := Bits.BoolToInt(interrupt.arePending());
                                                       (* interrupt bus *)
           displayBits[3] := Bits.BoolToInt(NOT(interrupt.uninhibited));
                                             (* interrupt inhibit flag *)
           displayBits[4] := Bits.BoolToInt(interrupt.areEnabled());
                                             (* ION flag *)
           displayBits[5] := 0;              (* user mode flag *)

      | GeneralUI.DisplayType.AC =>
             displayBits := Bits.ExplodeToInts(cpu.AC);
      | GeneralUI.DisplayType.MD =>
             displayBits := Bits.ExplodeToInts(cpu.MB);
      | GeneralUI.DisplayType.MQ => 
             displayBits := Bits.ExplodeToInts(cpu.MQ);
      | GeneralUI.DisplayType.Bus =>
             (* bus display not implemented *)
             displayBits := Bits.ExplodeToInts(0);
      END;

      FOR i := 0 TO 11 DO
        FormsVBT.PutInteger(info.form, memLightName[i], memBits[i]);
        FormsVBT.PutInteger(info.form, displayLightName[i], displayBits[i]);
      END;

      (* run flag *)
      FormsVBT.PutInteger(info.form, "runlight",
                          Bits.BoolToInt(info.cpu.run.running));

    END; (* WITH *)


  END StatusUpdater;


(* ----------------------------------------------------------------------
   Procedure : SWLineOff

   Set the SW line to 'off'
   ---------------------------------------------------------------------- *)
PROCEDURE SWLineOff(<* UNUSED *> form : FormsVBT.T := NIL;
                     <* UNUSED *> event : TEXT := NIL;
                     <* UNUSED *> data : REFANY := NIL;
                     <* UNUSED *> time : VBT.TimeStamp := 0) = 
  BEGIN
    info.cpu.SW := FALSE;
    info.statusUpdater();
  END SWLineOff;


(* ----------------------------------------------------------------------
   Procedure : SWLineOn

   Set the SW line to 'on'
   ---------------------------------------------------------------------- *)
PROCEDURE SWLineOn(<* UNUSED *> form : FormsVBT.T := NIL;
                     <* UNUSED *> event : TEXT := NIL;
                     <* UNUSED *> data : REFANY := NIL;
                     <* UNUSED *> time : VBT.TimeStamp := 0) = 
  BEGIN
    info.cpu.SW := TRUE;
    info.statusUpdater();
  END SWLineOn;

(* ----------------------------------------------------------------------
   Procedure : CancelMessage

   Cancels the message displayed
   ---------------------------------------------------------------------- *)
PROCEDURE CancelMessage(             form : FormsVBT.T := NIL;
                        <* UNUSED *> name : TEXT := NIL;
                        <* UNUSED *> data : REFANY := NIL;
                        <* UNUSED *> time : VBT.TimeStamp := 0) = 
  BEGIN
    FormsVBT.PutInteger(form, "messagebar", 0);
  END CancelMessage;


(* ----------------------------------------------------------------------
   Procedure : FormMessage

   Display an error message in the box provided on the form
   ---------------------------------------------------------------------- *)
PROCEDURE FormMessage(form : FormsVBT.T; message : TEXT) =
  BEGIN
    FormsVBT.PutInteger(form, "messagebar", 1);
    FormsVBT.PutText(form, "messages", message);
  END FormMessage;


(* ----------------------------------------------------------------------
   Procedure : MyFilter

   This procedure is invoked every time a character is entered into the
   Typescript called "teletype". It performs a few translations of characters
   and then stuffs them into the buffer from which the CharReader thread
   will read, as well as invoking the default filter method to write them
   into the Typescript
   ---------------------------------------------------------------------- *)
PROCEDURE MyFilter(p : TextPort.T; cd : VBT.KeyRec) =
  VAR
    normKey : CHAR;
  BEGIN
    IF cd.wentDown THEN
      (* if a control character then map as appropriate *)
      (* ignore most characters *)
      IF cd.whatChanged > 16_FF00 THEN
        CASE cd.whatChanged OF
        | KeyboardKey.Tab => normKey := ASCII.HT;
        | KeyboardKey.Delete => normKey := ASCII.DEL;
        | KeyboardKey.Return => normKey := ASCII.CR;
        ELSE
          normKey := ASCII.NUL;
        END;
      ELSE
        (* map lower case to upper case *)
        normKey := ASCII.Upper[VAL(cd.whatChanged, CHAR)];

        (* reflect this mapping back into the character seen by the *)
        (* default filter *)
        cd.whatChanged := ORD(normKey);
      END;

      (*TextPort.T.filter(p, cd);*)
      info.buffer.put(normKey);

    END;
  END MyFilter;


(* ----------------------------------------------------------------------
   Procedure : Realize

   Each FormsVBT object comes here. When we see the Typescript "teletype"
   then create it with a reference to our own filter function. Otherwise
   invoke the default realisation method.
   ---------------------------------------------------------------------- *)
PROCEDURE Realize(fv: FormsVBT.T; type, name: TEXT) : VBT.T
  RAISES { FormsVBT.Error } =
  BEGIN
    IF Text.Equal(type, "Typescript") AND Text.Equal(name, "teletype") THEN
      RETURN NEW(FVTypes.FVTypescript,
                 tp := NEW(TypescriptVBT.Port, filter:=MyFilter));
    ELSE
      RETURN FormsVBT.T.realize(fv, type, name)
    END;
  END Realize;


(* ----------------------------------------------------------------------
   Procedure : WrFilter

   Procedure to filter out the characters which a typescript cannot correctly
   display. This is done by returning the null character if one of these
   characters is input.
   ---------------------------------------------------------------------- *)
PROCEDURE WrFilter(c : CHAR) : CHAR =
  BEGIN
    CASE c OF
    | ASCII.CR, ASCII.DEL => RETURN ASCII.NUL
    ELSE
      RETURN c;
    END
  END WrFilter;

BEGIN
END GraphicalUI.





