(* ----------------------------------------------------------------------
   $Id: InterpretedOPR.m3,v 1.7 1995/06/23 07:57:54 sgp13 Exp $

   This is the interpreter for OPR instructions.
   This version is for interpreting instructions at run-time.
   ----------------------------------------------------------------------

   $Log: InterpretedOPR.m3,v $
# Revision 1.7  1995/06/23  07:57:54  sgp13
# Registers are now accessed directly
#
# Revision 1.6  1995/06/14  09:56:15  sgp13
# Array of pointers to procedures made global to fix compiling error on SPARC
# architecture.
#
# Revision 1.5  1995/03/14  12:29:02  sgp13
# Added group 2 NOP
# Changed IAC to carry into L
# Changed CMA to be more clearly coded.
#
# Revision 1.4  1995/02/24  22:26:43  sgp13
# Changes reflect changes to Bits interface
#
# Revision 1.3  1995/02/10  12:16:39  sgp13
# DoOPR renamed Implement.
# This is now one of two possibilities to use to export the OPR interface.
#
# Revision 1.2  1995/02/04  13:02:29  sgp13
# Bugs fixed - direction of rotation especially
#
# Revision 1.1  1995/02/01  13:38:00  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE InterpretedOPR EXPORTS OPR;

IMPORT Word, Fmt;
IMPORT Diag, Memory, Registers;
IMPORT Bits, CPU;

<* FATAL Registers.OutOfRange *>

  TYPE
    SkipGroup = { And, Or };
    ConditionTester = PROCEDURE(self:CPU.T) : BOOLEAN;
    SkipGroupArray = ARRAY SkipGroup OF ConditionTester;

  VAR
    testArray := ARRAY [5..7] OF SkipGroupArray {
                   SkipGroupArray { DoSPA, DoSMA },
                   SkipGroupArray { DoSNA, DoSZA },
                   SkipGroupArray { DoSZL, DoSNL } };

(* ----------------------------------------------------------------------
   Procedure : Implement

   Top level of interpreting an OPR instruction. Explodes its bit pattern
   into an array, and calls the procedure for the appropriate instruction
   group.
   ---------------------------------------------------------------------- *)
PROCEDURE Implement(operand : CPU.Operand; self : CPU.T) =
  VAR
    bits : Bits.Array;
  BEGIN
    (* expand the bit pattern into an array so that the status of each bit *)
    (* can be easily tested *)
    bits := Bits.Explode(operand);
    
    IF NOT bits[3] THEN
      DoGroup1(self (* or output *), operand, bits);
    ELSE
      IF NOT bits[11] THEN
        DoGroup2(self (* or output *), operand, bits);
      ELSE
        DoGroup3(self (* or output *), operand, bits);
      END;
    END;

  END Implement;


(* ----------------------------------------------------------------------
   Procedure : DoGroup1

   Procedure to interpret a group 1 instruction. The order of testing of the
   bits ensures that the actions are performed in the order specified in
   the PDP8 handbook.

   NB This should be easily translateable to produce code for the
      precompiled version.
   ---------------------------------------------------------------------- *)
PROCEDURE DoGroup1(self : CPU.T (* or output : Wr.T *);
                   operand : CPU.Operand; bits : Bits.Array) =
  VAR
    ACtemp : INTEGER; (* temporary store for AC - care needed in translation *)
    ACbits : Bits.Array;
    Ltemp : BOOLEAN;  (* temporary store for L - care needed in translation *)
    actionTaken : BOOLEAN := FALSE;  (* flag whether the instruction was *)
                                     (* recognised at all *)
                                
  BEGIN
    (* care is needed when translating the 'WITH' for the precompiled *)
    (* version. May not be necessary at all *)
    WITH AC = self.AC, PC = self.PC, L = self.L,
         SR = self.SR, MQ = self.MQ DO

      IF bits[8] AND bits[9] THEN
        (* Illegal instruction *)
        (* NB TAKE CARE TRANSLATING THIS FOR PRECOMPILED VERSION *)
        (*Diag.Message(Diag.T.Error, "Illegal OPR instruction (7" &
                     Fmt.Pad(Fmt.Int(operand,8),3,'0')&"): PC=" &
                     Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
        actionTaken := TRUE;
        RETURN;
      END;

      IF operand = 0 THEN
        (* group 1 NOP *)
        (*Diag.Message(Diag.T.InterpOPR, "NOP(group 1)");*)
        actionTaken := TRUE;
      END;

      IF bits[4] THEN
        (* CLA *)
        (*Diag.Message(Diag.T.InterpOPR, "CLA(group 1)");*)
        AC := 0;
        actionTaken := TRUE;
      END;

      IF bits[5] THEN
        (* CLL *)
        (*Diag.Message(Diag.T.InterpOPR, "CLL");*)
        L := FALSE;
        actionTaken := TRUE;
      END;

      IF bits[6] THEN
        (* CMA *)
        (*Diag.Message(Diag.T.InterpOPR, "CMA");*)
        ACbits := Bits.Explode(AC);
        FOR i := 0 TO 11 DO
          ACbits[i] := NOT(ACbits[i]);
        END;
        AC := Bits.Implode(ACbits);
        actionTaken := TRUE;
      END;

      IF bits[7] THEN
        (* CML *)
        (*Diag.Message(Diag.T.InterpOPR, "CML");*)
        L := NOT(L);
        actionTaken := TRUE;
      END;

      IF bits[11] THEN
        (* IAC *)
        (*Diag.Message(Diag.T.InterpOPR, "IAC");*)
        IF AC = LAST(CPU.Register_12) THEN
          AC := 0;
        ELSE
          INC(AC);
        END;

        (* according to Doug Jones' online manual: treat L/AC as 13 bit reg*)
        IF AC = 0 THEN
          L := NOT(L);
        END;
        actionTaken := TRUE;
      END;

      IF bits[9] THEN
        IF bits[10] THEN
          (* RTL *)
          (*Diag.Message(Diag.T.InterpOPR, "RTL");*)

          (* do preliminary shift on AC *)
          ACtemp := (AC MOD Bits.nextToTopBitOfByte) * 4;

          (* shift link value into new position *)
          IF L THEN
            INC(ACtemp, 2);
          END;
          
          (* get correct value for LSB of new AC *)
          IF Word.And(AC, Bits.topBitOfByte) # 0 THEN
            INC(ACtemp);
          END;
          
          (* get correct value for new link *)
          IF Word.And(AC, Bits.nextToTopBitOfByte) # 0 THEN
            L := TRUE;
          ELSE
            L := FALSE;
          END;
          
          (* write new accumulator value *)
          AC := ACtemp;

          actionTaken := TRUE;
        ELSE
          (* RAL *)
          (*Diag.Message(Diag.T.InterpOPR, "RAL");*)
          ACtemp := AC * 2;
          Ltemp := L;

          IF ACtemp >= Memory.byteSize THEN
            L := TRUE;
            AC := ACtemp MOD Memory.byteSize;
          ELSE
            L := FALSE;
            AC := ACtemp;
          END;
          IF Ltemp THEN
            INC(AC);
          END;
        END;
        actionTaken := TRUE;
      END;

      IF bits[8] THEN
        IF bits[10] THEN
          (* RTR *)
          (*Diag.Message(Diag.T.InterpOPR, "RTR");*)

          (* set up working value of AC *)
          ACtemp := AC DIV 4;
            
          (* set up bit 1 as appropriate *)
          IF L THEN
            INC(ACtemp, Bits.nextToTopBitOfByte);
          END;
          
          (* SET up bit 0 as appropriate *)
          IF Word.And(AC, Bits.bottomBitOfByte) # 0 THEN
            INC(ACtemp, Bits.topBitOfByte);
          END;

          (* SET up link as appropriate *)
          IF Word.And(AC, Bits.nextToBottomBitOfByte) # 0 THEN
            L := TRUE;
          ELSE
            L := FALSE;
          END;
            
          (* write new AC *)
          AC := ACtemp;

          actionTaken := TRUE;

        ELSE
          (* RAR *)
          (*Diag.Message(Diag.T.InterpOPR, "RAR");*)
          Ltemp := L;
          ACtemp := AC;

          (* check required value of link *)
          IF Word.And(ACtemp, 1) # 0 THEN
            L := TRUE;
          ELSE
            L := FALSE;
          END;
          
          (* set up and write new value of AC *)
          ACtemp := ACtemp DIV 2;
          IF Ltemp THEN
            INC(ACtemp, Bits.topBitOfByte);
          END;
          AC := ACtemp;

          actionTaken := TRUE;
        END;
      END;
      
      IF bits[10] AND (NOT bits[8] AND NOT bits[9]) THEN
        (* BSW *)
        (*Diag.Message(Diag.T.InterpOPR, "BSW");*)
        WITH high = Word.And(AC, Bits.highHalfOfByte)
                    DIV Bits.twoToTheSixth,
             low = Word.And(AC, Bits.lowHalfOfByte) DO
          AC := low * Bits.twoToTheSixth + high;
        END;
        actionTaken := TRUE;
      END;

      (* if no part of the code found the right bits set than flag illegal *)
      (* instruction *)
      IF NOT actionTaken THEN
        (*Diag.Message(Diag.T.Error, "Illegal OPR instruction (7" &
                     Fmt.Pad(Fmt.Int(operand,8),3,'0')&"): PC=" &
                     Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
      END;

    END; (* WITH *)
  END DoGroup1;


(* ----------------------------------------------------------------------
   Procedure : DoGroup2

   Procedure to interpret a group 2 instruction. The order of testing of the
   bits ensures that the actions are performed in the order specified in
   the PDP8 handbook.

   NB This will need a bit of rewriting for the precompiled version
   ---------------------------------------------------------------------- *)
PROCEDURE DoGroup2(self: CPU.T (* or output : Wr.T *);
                   operand : CPU.Operand; bits: Bits.Array) =
  VAR
    count : INTEGER := 0;         (* number of clauses in the skip test *)
    result : BOOLEAN;             (* result of the skip test *)
    skipgroup : SkipGroup;
    actionTaken : BOOLEAN := FALSE;  (* flag whether the instruction was *)
                                     (* recognised at all *)

  BEGIN
    (* NOP instruction *)
    IF operand = 8_400 THEN
      (*Diag.Message(Diag.T.InterpOPR, "NOP(group 2)");*)
      RETURN;
    END;

    (* care is needed when translating the 'WITH' for the precompiled *)
    (* version. May not be necessary at all *)
    WITH AC = self.AC, PC = self.PC, L = self.L,
         SR = self.SR, MQ = self.MQ DO

      (* check which group of skip instructions we're using, and set the *)
      (* initial result value accordingly *)
      IF bits[8] THEN
        skipgroup := SkipGroup.And;
        result := TRUE;
      ELSE
        skipgroup := SkipGroup.Or;
        result := FALSE;
      END;

      (* loop through performing each of the required tests *)
      FOR bit := 5 TO 7 DO
        IF bits[bit] THEN
          IF skipgroup = SkipGroup.And THEN
            result := result AND testArray[bit][skipgroup](self);
          ELSE
            result := result OR testArray[bit][skipgroup](self);
          END;
          INC(count);
        END;
      END;
      
      (* if an unconditional skip *)
      IF count = 0 AND bits[8] THEN
        (* SKIP *)
        IF PC = LAST(CPU.Register_12) THEN
          PC := 0;
        ELSE
          INC(PC);
        END;

        (*Diag.Message(Diag.T.InterpOPR, "SKIP(unconditional) to PC: " &
                     Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
        actionTaken := TRUE;
      END;

      (* if conditional skip *)
      IF count > 0 THEN
        actionTaken := TRUE;
        IF result THEN
          (* conditional SKIP - taken *)
          IF PC = LAST(CPU.Register_12) THEN
            PC := 0;
          ELSE
            INC(PC);
          END;
          (*Diag.Message(Diag.T.InterpOPR, "SKIP(conditional) taken to PC: " &
                     Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
        ELSE
          (* conditional SKIP - not taken *)
          (*Diag.Message(Diag.T.InterpOPR, "SKIP(conditional) not taken:" &
                       " next PC: " & Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
        END;
      END;
        
      IF bits[4] THEN
        (* CLA *)
        (*Diag.Message(Diag.T.InterpOPR, "CLA(group2)");*)
        AC := 0;
        actionTaken := TRUE;
      END;

      IF bits[9] THEN
        (* OSR *)
        (*Diag.Message(Diag.T.InterpOPR, "OSR");*)
        AC := Word.Or(AC, SR);
        actionTaken := TRUE;
      END;
      
      IF bits[10] THEN
        (* HLT *)
        (*Diag.Message(Diag.T.InterpOPR, "HLT");*)
        LOCK self.run DO
          self.run.continuous := FALSE;
        END;
        actionTaken := TRUE;
      END;
      (* if no part of the code found the right bits set than flag illegal *)
      (* instruction *)
      IF NOT actionTaken THEN
        (*Diag.Message(Diag.T.Error, "Illegal OPR instruction (7" &
                     Fmt.Pad(Fmt.Int(operand,8),3,'0')&"): PC=" &
                     Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
      END;
    END; (* WITH *)
  END DoGroup2;


(* ----------------------------------------------------------------------
   Procedure : DoGroup3

   Procedure to interpret a group 3 instruction. The order of testing of the
   bits ensures that the actions are performed in the order specified in
   the PDP8 handbook.

   NB this is the non-extended arithmetic version

   NB This should be easily translateable to produce code for the
      precompiled version.
   ---------------------------------------------------------------------- *)
PROCEDURE DoGroup3(self : CPU.T (* or output : Wr.T *);
                   operand : CPU.Operand; bits : Bits.Array) =
  VAR
    temp : INTEGER;        (* temporary variable for swap - consider whether *)
                           (* necessary in precompiled version ? *)
    actionTaken : BOOLEAN := FALSE;  (* flag whether the instruction was *)
                                     (* recognised at all *)
  BEGIN
    (* care is needed when translating the 'WITH' for the precompiled *)
    (* version. May not be necessary at all *)
    WITH AC = self.AC, PC = self.PC, L = self.L,
         SR = self.SR, MQ = self.MQ DO

      (* if EAE instruction bits set then treat as an illegal instruction *)
      IF bits[6] OR bits[8] OR bits[9] OR bits[10] THEN
        (* Illegal instruction - EAE specific code specified *)
        (*Diag.Message(Diag.T.Error,"Illegal OPR instruction: EAE not "
                                  & "implemented");*)
        actionTaken := TRUE;
      END;

      (* if no variable bits set then NOP - i.e. operand = 401 *)
      IF operand = 8_401 THEN
        (* group 3 NOP *)
        (*Diag.Message(Diag.T.InterpOPR, "NOP(group 3)");*)
        actionTaken := TRUE;
      END;

      IF bits[4] THEN
        (* CLA *)
        (*Diag.Message(Diag.T.InterpOPR, "CLA(group 3)");*)
        AC := 0;
        actionTaken := TRUE;
      END;

      IF bits[5] AND bits[7] THEN
        (* swap between MQ and AC *)
        (*Diag.Message(Diag.T.InterpOPR, "SWP");*)
        temp := MQ;
        MQ := AC;
        AC := temp;
        actionTaken := TRUE;
      ELSE
        IF bits[5] THEN
          (* MQA *)
          (*Diag.Message(Diag.T.InterpOPR, "MQA");*)
          AC := Word.Or(AC, MQ);
          actionTaken := TRUE;
        END;
        
        IF bits[7] THEN
          (* MQL *)
          (*Diag.Message(Diag.T.InterpOPR, "MQL");*)
          MQ := AC;
          AC := 0;
          actionTaken := TRUE;
        END;

      END;

      (* if no part of the code found the right bits set than flag illegal *)
      (* instruction *)
      IF NOT actionTaken THEN
        (*Diag.Message(Diag.T.Error, "Illegal OPR instruction (7" &
                     Fmt.Pad(Fmt.Int(operand,8),3,'0')&"): PC=" &
                     Fmt.Pad(Fmt.Int(PC,8),4,'0'));*)
      END;
    END; (* WITH *)
  END DoGroup3;


(* ----------------------------------------------------------------------
   Procedure : Do___

   The tests for the skip conditions.
   ---------------------------------------------------------------------- *)

PROCEDURE DoSPA(self: CPU.T) : BOOLEAN =
  BEGIN
    (*Diag.Message(Diag.T.InterpOPR, "SPA");*)
    RETURN self.AC < Memory.halfByteSize;
  END DoSPA;

PROCEDURE DoSMA(self: CPU.T) : BOOLEAN =
  BEGIN
    (*Diag.Message(Diag.T.InterpOPR, "SMA");*)
    RETURN self.AC >= Memory.halfByteSize;
  END DoSMA;

PROCEDURE DoSNA(self: CPU.T) : BOOLEAN =
  BEGIN
    (*Diag.Message(Diag.T.InterpOPR, "SNA");*)
    RETURN self.AC # 0;
  END DoSNA;

PROCEDURE DoSZA(self: CPU.T) : BOOLEAN =
  BEGIN
    (*Diag.Message(Diag.T.InterpOPR, "SZA");*)
    RETURN self.AC = 0;
  END DoSZA;

PROCEDURE DoSZL(self: CPU.T) : BOOLEAN =
  BEGIN
    (*Diag.Message(Diag.T.InterpOPR, "SZL");*)
    RETURN NOT(self.L);
  END DoSZL;

PROCEDURE DoSNL(self: CPU.T) : BOOLEAN =
  BEGIN
    (*Diag.Message(Diag.T.InterpOPR, "SNL");*)
    RETURN self.L;
  END DoSNL;



BEGIN
END InterpretedOPR.
    
