(* ----------------------------------------------------------------------
   $Id: KeyboardUI.m3,v 1.13 1995/06/24 21:56:19 sgp13 Exp $

   Module supplying the keyboard user interface services. The procedure is
   to read characters from the keyboard and match up against the allowed
   commands list. If a command is found then call the appropriate procedure.
   The procedures specific to the keyboard user interface are also defined
   in this module.
   ----------------------------------------------------------------------

   $Log: KeyboardUI.m3,v $
# Revision 1.13  1995/06/24  21:56:19  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.12  1995/06/23  07:54:37  sgp13
# Registers are now accessed directly
#
# Revision 1.11  1995/06/20  10:15:13  sgp13
# Changes due to IntBus version 1.2
#
# Revision 1.10  1995/03/22  13:18:45  sgp13
# Messages on Connect/Disconnect tightened up, assuming that the emulator is
# run without diagnostics echoing to stdout
#
# Revision 1.9  1995/03/15  21:02:43  sgp13
# Initialisation of GeneralUI module moved to Control.m3
#
# Revision 1.8  1995/03/15  14:54:03  sgp13
# Interrupts now implemented by IntBus.T
# 'IN' command removed (incompatible with IntBus.T method of interrupts)
#
# Revision 1.7  1995/03/14  12:24:24  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.6  1995/03/07  23:24:22  sgp13
# Commands to connect and disconnect I/O streams added.
#
# Revision 1.5  1995/02/24  22:24:58  sgp13
# The data object is now a Processor object containing CPU and Bus
# OutputStatus also changed.
#
# Revision 1.4  1995/02/01  13:29:55  sgp13
# Updated to display values in octal and binary
#
# Revision 1.3  1995/01/23  12:56:37  sgp13
# Enabling of interrupts at start of run is now optional and determined by input
# file, and implemented by CPU module.
#
# Revision 1.2  1995/01/20  13:10:24  sgp13
# Patched to display save field registers in display mode 2
#
# Revision 1.1  1995/01/19  09:26:17  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE KeyboardUI;

IMPORT Thread, Text, Fmt, IO, Wr, Rd, Stdio, ASCII, Lex, Scan;
IMPORT FormsVBT, VBT;
IMPORT Bits, Diag, Registers, WhiteSpaceFilter;
IMPORT GeneralUI;

EXCEPTION Syntax;

<* FATAL Wr.Failure, Registers.OutOfRange *>

TYPE
  Command = RECORD
    name : TEXT;
    proc : Applicator;
  END;

  Applicator = PROCEDURE (form : FormsVBT.T := NIL; event : TEXT := NIL;
                          data : REFANY := NIL; time : VBT.TimeStamp := 0);


  (* types to record the device specified in the CO and DC commands *)
  CharStreamDirection = { Input, Output };
  StreamInfo = RECORD
    device : GeneralUI.CharStreamDevice;
    direction : CharStreamDirection;
  END;

VAR
  info : T;
  commands := ARRAY [0..14] OF Command {
       Command { "QU", GeneralUI.Quit },
       Command { "HA", GeneralUI.Halt },
       Command { "SR", SwitchRegister },
       Command { "SW", GeneralUI.SWLine },
       Command { "AL", GeneralUI.AddressLoad },
       Command { "XL", GeneralUI.ExtdAddressLoad },
       Command { "CL", GeneralUI.Clear },
       Command { "GO", GeneralUI.Continue },
       Command { "SS", GeneralUI.SingleStep },
       Command { "EX", GeneralUI.Examine },
       Command { "DE", GeneralUI.Deposit },
       Command { "DI", Display },
       Command { "HE", Help },
       Command { "CO", Connect },
       Command { "DC", Disconnect }};


(* ----------------------------------------------------------------------
   Procedure : DoKeyboardUI

   Procedure invoked to start the user interface thread.
   ---------------------------------------------------------------------- *)
PROCEDURE DoKeyboardUI(self : T) : REFANY RAISES {} =
  VAR
    command : TEXT;
    commandIndex : INTEGER;
  BEGIN
    (* store reference to the information object so that it can be
       seen by the 'Applicator' procedures.
    *)
    info := self;

    IO.Put("PDP8/e keyboard interface : waiting for input\n");

    TRY
      (* loop until alerted to finish *)
      WHILE NOT Thread.TestAlert() DO
        (* read two character command, filtering out white space *)
        (* the command is also converted to upper case *)
        command := Text.FromChar(ASCII.Upper[self.input.get()])
                    & Text.FromChar(ASCII.Upper[self.input.get()]);

        (* search for the command in the command array *)
        commandIndex := -1;
        FOR i := FIRST(commands) TO LAST(commands) DO
          IF Text.Equal(command, commands[i].name) THEN
            commandIndex := i;
            EXIT;
          END;
        END;

        (* if the command is found then print its name else flag an error *)
        IF commandIndex # -1 THEN
          commands[commandIndex].proc();
        ELSE
          (*Diag.Message(Diag.T.KeyboardUI, "Unrecognised command: " & command );*)
          IO.Put("Unrecognised command: " & command );
          IO.Put("\nType 'HE' for help\n");
        END;

        IF Text.Equal(command, "QU") THEN
          EXIT;
        END;

      END;
    EXCEPT
    | Thread.Alerted =>
    | Rd.Failure => IO.Put("Command reading failed\n");
                    self.cpu.quit.signal();
    | Rd.EndOfFile => IO.Put("End of command input file\n");
                      self.cpu.quit.signal();
    | Wr.Failure => IO.Put("Display failure\n");
                    self.cpu.quit.signal();
    END;

    RETURN self;
  END DoKeyboardUI;


(* ----------------------------------------------------------------------
   Procedure : Connect

   Connect the specified file to the appropriate device (see help info for
   syntax). Read the first two characters to specify the device to connect to,
   then read to the next '*' for the filename
   ---------------------------------------------------------------------- *)
PROCEDURE Connect(<* UNUSED *> form : FormsVBT.T := NIL;
                  <* UNUSED *> event : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    stream : StreamInfo;
    ch : CHAR;
    filename : TEXT := "";
    result : GeneralUI.StreamResult;
  BEGIN
    TRY
      (* read the device specifiers from kbd *)
      stream := SelectStream();

      (* read filename *)
      ch := info.input.get();
      WHILE ch # '*' DO
        filename := filename & Text.FromChar(ch);
        ch := info.input.get();
      END;

      (* attempt to connect - according to whether input or output *)
      CASE stream.direction OF
      | CharStreamDirection.Input =>
                      result := GeneralUI.ConnectInput(stream.device,filename);
      | CharStreamDirection.Output =>
                      result:= GeneralUI.ConnectOutput(stream.device,filename);
      END;

      (* report on the results of the operation *)
      CASE result OF
      | GeneralUI.StreamResult.Success => (* no report on success *)
      | GeneralUI.StreamResult.FileFailure =>
           IO.Put("failed to open " & filename & "\n");
      | GeneralUI.StreamResult.AlreadyConnected =>
           IO.Put("file already connected to this stream: disconnect first\n");
      | GeneralUI.StreamResult.NotConnected => 
           IO.Put("internal error: NotConnected returned to Connect\n");
           (*Diag.Message(Diag.T.Error, "internal: KeyboardUI.Connect " &
                        "received unexpected result NotConnected");*)
      END;

    EXCEPT
    | Syntax => (*Diag.Message(Diag.T.KeyboardUI, "Connect: bad syntax");*)
    | Rd.Failure, Rd.EndOfFile =>
                (*Diag.Message(Diag.T.KeyboardUI, "Connect: input failure");*)
    | Thread.Alerted =>
    END;

  END Connect;


(* ----------------------------------------------------------------------
   Procedure : Disconnect

   Disconnect the specified device (see help info for syntax). Read the first
   two characters to specify the device.

   ---------------------------------------------------------------------- *)
PROCEDURE Disconnect(<* UNUSED *> form : FormsVBT.T := NIL;
                     <* UNUSED *> event : TEXT := NIL;
                     <* UNUSED *> data : REFANY := NIL;
                     <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    stream : StreamInfo;
    result : GeneralUI.StreamResult;
  BEGIN
    TRY
      (* read the device specifiers from kbd *)
      stream := SelectStream();

      (* attempt to disconnect - according to whether input or output *)
      CASE stream.direction OF
      | CharStreamDirection.Input =>
                      result := GeneralUI.DisconnectInput(stream.device);
      | CharStreamDirection.Output =>
                      result := GeneralUI.DisconnectOutput(stream.device);
      END;

      CASE result OF
      | GeneralUI.StreamResult.Success => (* no report on success *)
      | GeneralUI.StreamResult.FileFailure =>
           IO.Put("internal error: FileFailure returned to Disconnect\n");
           (*Diag.Message(Diag.T.Error, "internal: KeyboardUI.Disconnect " &
                        "received unexpected result FileFailure");*)
      | GeneralUI.StreamResult.AlreadyConnected =>
           IO.Put("internal error: AlreadyConnected returned to Disconnect\n");
           (*Diag.Message(Diag.T.Error, "internal: KeyboardUI.Disconnect " &
                        "received unexpected result AlreadyConnected");*)
      | GeneralUI.StreamResult.NotConnected => 
           IO.Put("no file connected to this stream\n");
      END;

    EXCEPT
    | Syntax => (*Diag.Message(Diag.T.KeyboardUI, "Disconnect: bad syntax");*)
    | Rd.Failure, Rd.EndOfFile =>
                (*Diag.Message(Diag.T.KeyboardUI, "Disconnect: input failure");*)
    | Thread.Alerted =>
    END;

  END Disconnect;


(* ----------------------------------------------------------------------
   Procedure : Display

   Change the information display. Accepts a numerical argument which
   specifies the new setting for the simulation of the rotary switch on
   the front panel which defines what information is displayed.

   This command may be issued at any time
   ---------------------------------------------------------------------- *)
PROCEDURE Display(<* UNUSED *> form : FormsVBT.T := NIL;
                  <* UNUSED *> event : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    ch : CHAR;
    choice : INTEGER;
  BEGIN
    TRY
      (* get choice *)
      ch := info.input.get();
      choice := ORD(ch) - ORD('1');

      (* if in range then change display type, else flag error *)
      IF choice >= 0 AND choice < NUMBER(GeneralUI.DisplayType) THEN
        (*Diag.Message(Diag.T.KeyboardUI, "command=Display " &Fmt.Int(choice+1));*)
        info.displayType := VAL(choice, GeneralUI.DisplayType);
        info.statusUpdater();
      ELSE
        (*Diag.Message(Diag.T.KeyboardUI, "Display: invalid parameter");*)
        IO.Put("DI: invalid parameter\n");
        Wr.Flush(Stdio.stdout);
      END;
    EXCEPT
    | Rd.Failure, Rd.EndOfFile, Thread.Alerted =>
    END;

  END Display;


(* ----------------------------------------------------------------------
   Procedure : Help

   Display brief resume of the commands. This command may be issued at any
   time.
   ---------------------------------------------------------------------- *)
PROCEDURE Help(<* UNUSED *> form : FormsVBT.T := NIL;
                  <* UNUSED *> event : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) =
  BEGIN
    (*Diag.Message(Diag.T.KeyboardUI, "command=Help");*)
    IO.Put("AL: take address from switch register\n");
    IO.Put("CL: clear AC, L, interrupts, I/O flags\n");
    IO.Put("CO {R|W} {K|T} <filename>*: connect file to Reader|Writer\n");
    IO.Put("                            of teletype(K)|Tape reader\n");
    IO.Put("DC {R|W} {K|T}: disconnect file\n");
    IO.Put("DE: deposit value in SR at current address, advance address\n");
    IO.Put("DI <1-6> : equivalent to settings of selector on front panel\n");
    IO.Put("EX: examine value at current address, advance address by one\n");
    IO.Put("GO: start free-running execution\n");
    IO.Put("HA: stop free-running execution\n");
    IO.Put("HE: print this info\n");
    IO.Put("QU: quit\n");
    IO.Put("SR <xxxx> set switch register\n");
    IO.Put("SS: execute single processor cycle\n");
    IO.Put("SW: toggle SW line\n");
    IO.Put("XL: load IF from 6-8 of SR, DF from 9-11 of SR\n");
    IO.Put("\n");
  END Help;



(* ----------------------------------------------------------------------
   Procedure : ObjectInit

   Initialise KeyboardUI object given CPU object information, and the name
   of the input stream.
   Have to call the GeneralUI initialisor to do most of the work.
   ---------------------------------------------------------------------- *)
PROCEDURE ObjectInit(self : T; input : Rd.T) : T =
  BEGIN
    self.statusUpdater := OutputStatus;
    self.input := NEW(WhiteSpaceFilter.T).init(input);
    RETURN self;
  END ObjectInit;



(* ----------------------------------------------------------------------
   Procedure : OutputStatus

   Display the current status of the machine.
   The information displayed is the value of the instruction field register,
   the value of the program counter, and then information which depends on the
   value of the display flag.
   
   The types of display correspond roughly to the settings of the switch on the
   front panel of the machine.
   ---------------------------------------------------------------------- *)
PROCEDURE OutputStatus() RAISES { Wr.Failure, Thread.Alerted } =
  VAR
    status : TEXT;
  BEGIN
    WITH cpu = info.cpu DO
      (* display memory field and address *)
      status := "EMA: " & Fmt.Int(cpu.EM_IFnum) &
                " Addr: " & Fmt.Pad(Fmt.Int(cpu.PC, 8), 4, '0') & " ";

      (* run flag *)
      LOCK cpu.run DO
        IF cpu.run.running THEN
          status := status & "running ";
        ELSE
          status := status & "halted ";
        END;
      END;

      CASE info.displayType OF
      | GeneralUI.DisplayType.State =>
            status:=status & "IR: " & Fmt.Int(cpu.IR)
                           & " SW: " & Bits.BoolToText(cpu.SW);
      | GeneralUI.DisplayType.Status =>
            status := status & "Status: " &
              (* link *)              Bits.BoolToText(cpu.L) &
              (* GT *)                "-";
            (* interrupt bus *)
            IF cpu.interrupt.arePending() THEN
              status := status & "1";
            ELSE
              status := status & "0";
            END;

            (* interrupt inhibit *)
            IF cpu.interrupt.uninhibited THEN
              status := status & "0";
            ELSE
              status := status & "1";
            END;

            (* ION enabled *)
            IF cpu.interrupt.areEnabled() THEN
              status := status & "1";
            ELSE
              status := status & "0";
            END;

            status := status &
            (* user mode *)         "-" &
            (* separator *)         ":IF=" &
            (* IF *)                Fmt.Int(cpu.EM_IFnum) &
            (* separator *)         ":DF=" &
            (* DF *)                Fmt.Int(cpu.EM_DFnum) &
            (* separator *)         ":SIF=" &
            (* SIF *)               Fmt.Int(cpu.EM_SIF) &
            (* separator *)         ":SDF=" &
            (* SDF *)               Fmt.Int(cpu.EM_SDF);

      | GeneralUI.DisplayType.AC =>
            status := status &"AC: "& Fmt.Pad(Fmt.Int(cpu.AC,8),4,'0')
                           & " (" & Fmt.Pad(Fmt.Int(cpu.AC, 2),12,'0')
                           & ")";
      | GeneralUI.DisplayType.MD =>
            status := status &"MD: "& Fmt.Pad(Fmt.Int(cpu.MB,8),4,'0')
                           & " (" & Fmt.Pad(Fmt.Int(cpu.MB, 2),12,'0')
                           & ")";
      | GeneralUI.DisplayType.MQ =>
            status := status &"MQ: "& Fmt.Pad(Fmt.Int(cpu.MQ,8),4,'0')
                           & " (" & Fmt.Pad(Fmt.Int(cpu.MQ, 2),12,'0')
                           & ")";
      | GeneralUI.DisplayType.Bus =>
            status := status & "Bus display not implemented";
      END;
    END;

    (*Diag.Message(Diag.T.KeyboardUI, status);*)
    IO.Put(status & "\n");
    Wr.Flush(Stdio.stdout);
  END OutputStatus;


(* ----------------------------------------------------------------------
   Procedure : SelectStream

   Reads characters for CO and DC to establish the device required
   ---------------------------------------------------------------------- *)
PROCEDURE SelectStream() : StreamInfo RAISES { Syntax, Rd.Failure,
                                               Rd.EndOfFile, Thread.Alerted } =
  VAR
    stream : StreamInfo;
    ch : CHAR;
  BEGIN
    ch := ASCII.Upper[info.input.get()];
    CASE ch OF
    | 'R' => stream.direction := CharStreamDirection.Input;
    | 'W' => stream.direction := CharStreamDirection.Output;
    ELSE
      RAISE Syntax;
    END;

    ch := ASCII.Upper[info.input.get()];
    CASE ch OF
    | 'K' => stream.device := GeneralUI.CharStreamDevice.Teletype;
    | 'T' => stream.device := GeneralUI.CharStreamDevice.Tape;
    ELSE
      RAISE Syntax;
    END;
    
    RETURN stream;

  END SelectStream;


(* ----------------------------------------------------------------------
   Procedure : SwitchRegister

   Set the switch register. This command may be issued at any time.
   ---------------------------------------------------------------------- *)
PROCEDURE SwitchRegister(<* UNUSED *> form : FormsVBT.T := NIL;
                  <* UNUSED *> event : TEXT := NIL;
                  <* UNUSED *> data : REFANY := NIL;
                  <* UNUSED *> time : VBT.TimeStamp := 0) =
  VAR
    string : TEXT := "";
    value : INTEGER;

  BEGIN
    TRY
      (* read new value from keyboard *)
      FOR i := 1 TO 4 DO
        string := string & Text.FromChar(info.input.get());
      END;

      (* convert to a number *)
      TRY
        value := Scan.Int(string, 8);
        (*Diag.Message(Diag.T.KeyboardUI, "command=SwitchRegister " & string);*)
        info.cpu.SR := value;
        info.statusUpdater();
      EXCEPT
      | Lex.Error  =>
          IO.Put("invalid input\n");
          (*Diag.Message(Diag.T.KeyboardUI, "command=SwitchRegister:" &
                       " invalid input");*)
      END;
    EXCEPT
    | Rd.Failure, Rd.EndOfFile, Thread.Alerted =>
    END;

  END SwitchRegister;

BEGIN
END KeyboardUI.




