(* ----------------------------------------------------------------------
   $Id: Processor.m3,v 1.11 1995/06/24 21:57:53 sgp13 Exp $

   The module for the central processor thread.
   The information objects are the CPU and Bus types.

   This module used to be called the CPU module, but that got confusing
   with a CPUdata object as well. To refer to previous versions not
   not recorded in the RCS file, see the CPUCycle directory.

   ----------------------------------------------------------------------

   $Log: Processor.m3,v $
# Revision 1.11  1995/06/24  21:57:53  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.10  1995/06/23  14:12:23  sgp13
# interrupt information used directly
#
# Revision 1.9  1995/06/23  07:55:24  sgp13
# Registers are now accessed directly
#
# Revision 1.8  1995/06/22  14:02:35  sgp13
# Memory made direct access
#
# Revision 1.7  1995/06/21  14:00:29  sgp13
# Unnecessary Init.Reset() removed.
#
# Revision 1.6  1995/06/20  10:16:15  sgp13
# Changes due to IntBus version 1.2
#
# Revision 1.5  1995/03/22  13:20:27  sgp13
# T.userInterface: type changed to UIType to allow user to specify
#
# Revision 1.4  1995/03/15  21:02:01  sgp13
# Automatic execution implemented
#
# Revision 1.3  1995/03/15  13:08:46  sgp13
# Interrupt method now uses the IntBus.T object.
#
# Revision 1.2  1995/03/14  12:26:41  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/02/24  22:32:18  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Processor;

IMPORT Thread, Fmt, Word, Rd, Scan, Lex;
IMPORT Diag, Init, Memory, Registers;
IMPORT Bus;
IMPORT CPU, SimpleInstruction, OPR;

<* FATAL Registers.OutOfRange *>

CONST
  (* constants for screening parts of the instruction while decoding *)
  IndirectionBit =       2_000100000000;
  CurrentOrZeroPageBit = 2_000010000000;
  PageBits =             2_111110000000;
  PageOffset =           2_000001111111;

VAR
  (* this flag is set from the initialisation file *)
  interruptsStartEnabled : BOOLEAN := FALSE;
  initActions := ARRAY [0..0] OF Init.Action {
                 Init.Action { "InterruptsStartEnabled", SetInterruptFlag }};


(* ----------------------------------------------------------------------
   Procedure : DoProcessor

   The procedure to implement the CPU thread. This consists of an enormous
   loop which goes round executing instructions. Before each instruction
   it checks whether it should be running and stops if it should not be. It
   also checks for interrupts and DMA requests and handles them as
   appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE DoProcessor(self : T) : REFANY RAISES {} =
  VAR
    DMAflag : INTEGER := 0;
    stopFlag : BOOLEAN := FALSE;

    opaddress : Memory.Address; (* address specified by operand *)
    opfield : REF Memory.Field; (* field to be used with above *)
    directAddress : Memory.Address; (* intermediate address *)

    (* use interrupt cycle counter instead *)
    (* instructionCount : CARDINAL := 0; (* number of instructions executed*)*)
  BEGIN
    WITH cpu = self.cpu, bus = self.bus DO
      (* enable interrupts if they are required on startup *)
      IF interruptsStartEnabled THEN
        (*Diag.Message(Diag.T.Processor, "Interrupts enabled on initialisation");*)
        cpu.interrupt.enable();
      END;

      (* if in autoexec mode then set run flags *)
      IF self.autoExec THEN
        LOCK cpu.run DO
          cpu.run.continuous := TRUE;
          cpu.run.running := TRUE;
        END;
      END;

      TRY
        (* loop until alerted by the control thread - or until we drop *)
        (* out due to raising CPU.Quit *)
        WHILE NOT Thread.TestAlert()
              AND cpu.interrupt.cycleCount < self.instructionLimit DO
          (* test whether we should run on to the next cycle or should stop *)
          (* and wait until told to continue *)
          (* also signal end of cycle on appropriate signal - see notes *)
          LOCK cpu.run DO
            IF NOT(cpu.run.continuous) THEN
              cpu.run.running := FALSE;
              stopFlag := TRUE;
              cpu.endOfSingleCycle.signal();
            ELSE
              stopFlag := FALSE;
              cpu.endOfFreeRunCycle.signal();
            END;
          END;
          
          IF stopFlag THEN
            IF self.userInterface = UIType.None THEN
              (*Diag.Message(Diag.T.Error, "No user interface present: " &
                                         "HLT is terminal");*)
              RAISE CPU.Quit;
            END;
            (*Diag.Message(Diag.T.Processor, "Checking for request to go");*)
            cpu.run.request.wait();
            LOCK cpu.run DO
              cpu.run.running := TRUE;
(*              IF cpu.run.continuous THEN
                (*Diag.Message(Diag.T.Processor, "Requested to go:free run");*)
              ELSE
                (*Diag.Message(Diag.T.Processor, "Requested to go:single step");*)
              END;
*)
            END;
          END;
          
          (* test for DMA request *)
          (* NB do not keep the object locked more than necessary - so that *)
          (* other DMA requests can be registered asap. There is no problem *)
          (* attempting to grant more than one simultaneously as this can *)
          (* only occur from this point *)
(*          LOCK cpu.DMA DO
            DMAflag := cpu.DMA.request;
          END;
*)
(*          IF DMAflag > 0 THEN
            (* identify DMA request, and execute appropriate procedure *)
            (* this latter procedure is responsible for clearing request *)
            (* line. The GrantDMA procedure should check when the DMA *)
            (* executor returns whether any of the DMA lines is still high *)
            (* and if so keep the request flag high, otherwise clear it *)
            GrantDMA(cpu);
          END;
*)
          (* test for interrupt request *)
          (* cpu.interrupt.request is a COUNTER for the number of requests *)
          (* outstanding. alteration is not subject to race conditions *)
          (* because the object is locked during access *)
          (* if interrupts enabled then grant interrupt if there is a request
             pending and if not inhibited *)
          (* making the tests directly is quicker than invoking methods *)
          WITH interrupt = self.cpu.interrupt DO
            IF interrupt.cycleCount >= interrupt.delayCount AND
                interrupt.uninhibited                               (*active*)
                AND interrupt.totalPending > 0 THEN                (*pending*)
              GrantInterrupt(cpu);
            END;
          END;

          (* load and decode instruction - performing address calculation
             if necessary
          *)
          WITH I_Fptr = cpu.EM_IFptr, PC = cpu.PC,
               instruction = I_Fptr[PC],
               opcode = instruction DIV 8_1000 DO
            
            (*Diag.Message(Diag.T.Processor, "PC="
                                     & Fmt.Pad(Fmt.Int(PC, 8), 4, '0') &
                                     " instruction=" &
                                     Fmt.Pad(Fmt.Int(instruction, 8), 4, '0'));*)

            (* write instruction code to IR register *)
            cpu.IR := opcode;

            (* if the instruction is AND TAD ISZ DCA JMP JMS then contains
               an address to decode, so do so 
            *)
            IF opcode < 6 THEN
              (* field setting (defaults to I_F) *)
              opfield := I_Fptr;
              
              (* calculate direct address from instruction *)
              (* first, use appropriate page bits *)
              IF Word.And(instruction, CurrentOrZeroPageBit) # 0 THEN
                directAddress := Word.And(PC, PageBits);
              ELSE
                directAddress := 0;
              END;
              
              (* next add bits from instruction *)
              INC(directAddress, Word.And(instruction, PageOffset));
              
              (* check whether indirect address required *)
              IF Word.And(instruction, IndirectionBit) # 0 THEN
                (* check if autoindexing is required and read from memory *)
                IF directAddress > 7 AND directAddress < 16 THEN
                  WITH value = I_Fptr[directAddress] DO
                    IF value = LAST(Memory.Value) THEN
                      value := 0;
                    ELSE
                      INC(value);
                    END;
                  END;
                END;
                opaddress := I_Fptr[directAddress];
                
                (* if this address refers to the data field, i.e. if the 
                   instruction is AND TAD ISZ or DCA then set the field
                   accordingly
                *)
                IF opcode < 4 THEN
                  opfield := cpu.EM_DFptr;
                END;
              ELSE
                (* direct address required *)
                opaddress := directAddress;
              END;
            END;
            
            (* increment PC *)
            WITH PC = cpu.PC DO
              IF PC = LAST(CPU.Register_12) THEN
                PC := 0;
              ELSE
                INC(PC);
              END;
            END;
            
            (* execute instruction *)
            IF opcode < 6 THEN
            
              (*Diag.Message(Diag.T.Instruction, "Address from operand="
                           & Fmt.Int(opfield) & ":"
                           & Fmt.Pad(Fmt.Int(opaddress, 8),4,'0'));*)
              SimpleInstruction.table[opcode](cpu, opaddress, opfield);
            ELSE
              WITH command = instruction MOD CPU.OperandSize DO
                IF opcode = 7 THEN
                  OPR.Implement(command, cpu);
                ELSE
                  bus.command(cpu, command);
                END;
              END; (* WITH *)
            END;

          END; (* WITH *)

          (* increment number of instructions executed *)
          (* use interrupt cycle counter *)
          (* INC(instructionCount); *)

          (* at the end of each cycle increment the interrupt cycle count *)
          (* to keep track of when interrupts should be active *)
          INC(cpu.interrupt.cycleCount);

        END; (* WHILE *)
        
      EXCEPT
      | CPU.Quit => cpu.quit.signal();
      | Thread.Alerted =>
      END;

      (* if instruction limit reached then output a suitable message *)
      IF cpu.interrupt.cycleCount >= self.instructionLimit THEN
        (*Diag.Message(Diag.T.Error, "Instruction limit exceeded");*)
        cpu.quit.signal();
      END;

    END; (* WITH *)
    (*Diag.Message(Diag.T.Processor, "Quitting");*)

    RETURN self;
  END DoProcessor;


(* ----------------------------------------------------------------------
   Procedure : GrantInterrupt

   Grant the interrupt request. This includes disabling interrupts, setting the
   field registers, placing the current value of PC in address zero and jumping
   to address 1. It is the responsibility of the _calling_device_ to clear
   the interrupt by decrementing the counter.
   ---------------------------------------------------------------------- *)
PROCEDURE GrantInterrupt(cpu : CPU.T) RAISES { CPU.Quit, Thread.Alerted } =
  BEGIN
    WITH PC = cpu.PC DO
      cpu.memory.mem[0]^[0] := PC;
      cpu.EM_SIF := cpu.EM_IFnum;
      cpu.EM_SDF := cpu.EM_DFnum;
      cpu.EM_DFnum := 0;
      cpu.EM_DFptr := cpu.memory.mem[0];
      cpu.EM_IFnum := 0;
      cpu.EM_IFptr := cpu.memory.mem[0];
      PC := 1;
    END;

    cpu.interrupt.disable();

    (*Diag.Message(Diag.T.Interrupt, "Interrupt granted");*)

  END GrantInterrupt;



(* ----------------------------------------------------------------------
   Procedure : GrantDMA

   Determine who requested DMA. Call the appropriate implementation
   procedure. On return test whether any DMA line is still high and if so
   keep the request flag high otherwise reset it. However, do NOT loop to
   grant that DMA - if we do that the processor could potentially be
   locked out forever. NB DMArequest is NOT locked on entry to this
   procedure, and should only be locked for the minimum time necessary.
   ---------------------------------------------------------------------- *)
PROCEDURE GrantDMA(cpu : CPU.T) RAISES { CPU.Quit } =
  BEGIN
    (*Diag.Message(Diag.T.DMA, "Not implemented");*)
    (* this stub rountine must clear the request for testing purposes *)
    LOCK cpu.DMA DO
      DEC(cpu.DMA.request);
    END;
  END GrantDMA;


(* ----------------------------------------------------------------------
   Procedure : ObjectInit

   Initialise the processor thread closure. Initialise the memory from the
   memory dump file if one is specified.
   ---------------------------------------------------------------------- *)
PROCEDURE ObjectInit(self : T; memoryfile : Rd.T := NIL) : T
    RAISES { Memory.Failure, Thread.Alerted } =
  BEGIN
    self.cpu := NEW(CPU.T).init(memoryfile);
    self.bus := NEW(Bus.T).init();
    RETURN self;
  END ObjectInit;


(* ----------------------------------------------------------------------
   Procedure : SetInterruptFlag

   Set the interrupt flag according to the value of the parameter, to
   decide whether we start with interrupts enabled or not
   ---------------------------------------------------------------------- *)
PROCEDURE SetInterruptFlag(name : TEXT; value : TEXT) =
  BEGIN
    TRY
      IF Scan.Bool(value) THEN
        interruptsStartEnabled := TRUE;
      ELSE
        interruptsStartEnabled := FALSE;
      END;
    EXCEPT
    | Lex.Error => (*Diag.Message(Diag.T.Error, "Init file: " & name & "=" &
                                value & ": unrecognised value");*)
    END;
  END SetInterruptFlag;


BEGIN
  (* test for the initialisation flags *)
  Init.Scan("Processor", initActions);
END Processor.
