(* ----------------------------------------------------------------------
   $Id: Bus.m3,v 1.6 1995/06/23 07:44:23 sgp13 Exp $

   This module implements the Bus interface - for specifications see that
   file.
   ----------------------------------------------------------------------

   $Log: Bus.m3,v $
# Revision 1.6  1995/06/23  07:44:23  sgp13
# Registers are now accessed directly
#
# Revision 1.5  1995/03/15  14:43:13  sgp13
# Interrupts now use IntBus.T
#
# Revision 1.4  1995/03/14  12:13:10  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.3  1995/03/07  23:06:14  sgp13
# Tape and Teletype implemented
#
# Revision 1.2  1995/02/24  22:14:35  sgp13
# Interrupts and Extended memory implemented
#
# Revision 1.1  1995/02/21  09:38:34  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Bus;

IMPORT Fmt, Thread;
IMPORT CPU, Diag, Registers;
IMPORT Device, ExtdMem, Interrupts, Tape, Teletype;

<* FATAL Registers.OutOfRange *>

(* ----------------------------------------------------------------------
   Procedure : Command

   Route a bus command to the appropriate device
   ---------------------------------------------------------------------- *)
PROCEDURE Command(self : T; cpu : CPU.T; instruction : INTEGER)
    RAISES { Thread.Alerted } =
  BEGIN

    (* break the instruction down into three-bit sections *)
    WITH devselA = (instruction DIV 8_100) MOD 8_10,
         devselB = (instruction DIV 8_010) MOD 8_10,
         command = instruction MOD 8_10 DO
      TRY
        (* switch depending on the first digit of the device code *)
        CASE devselA OF
        | 0      => (* switch depending on second digit of the device code *)
                    CASE devselB OF
                    | 0   =>
                        (*Diag.Message(Diag.T.Bus, "interrupt command: 6"
                            & Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
                        (* if this is the CAF command (6007) then intercept *)
                        (* the command HERE, so that references to the *)
                        (* objects' reset methods are available *)
                        IF command = 7 THEN
                          DoCAF(self, cpu);
                        ELSE
                          self.interrupts.command(cpu, devselB, command);
                        END;

                    | 1,2 => (*Diag.Message(Diag.T.Bus, "tape command: 6"
                                  & Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
                             self.tape.command(cpu, devselB, command);

                    | 3,4 => (*Diag.Message(Diag.T.Bus, "teletype command: 6"
                                  & Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
                             self.teletype.command(cpu, devselB, command);

                    | 5..7 => (*Diag.Message(Diag.T.Bus,"oscilloscope command: 6"
                                  & Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
                             self.oscilloscope.command(cpu, devselB, command);
                    END;

        | 1, 3..6 => (* devices in these ranges are unimplemented *)
                    RAISE Device.DeviceUnimplemented;

        | 2      => (* all devices in this range are extended memory *)
                    (*Diag.Message(Diag.T.Bus, "extended memory command: 6"
                                  & Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
                    self.extdmem.command(cpu, devselB, command);

        | 7      => CASE devselB OF
                    | 3..5 => (* devices 73-75 are the disc drive *)
                             (*Diag.Message(Diag.T.Bus, "disc drive command: 6"
                                  & Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
                             self.disc.command(cpu, devselB, command);
                    ELSE
                      (* all others are unimplemented *)
                      RAISE Device.DeviceUnimplemented;
                    END;
        END;
      EXCEPT
      | Device.DeviceUnimplemented =>
          (*Diag.Message(Diag.T.Error, "device unimplemented: command 6" &
                                  Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
      | Device.CommandUnimplemented =>
          (*Diag.Message(Diag.T.Error, "command unimplemented: command 6" &
                                  Fmt.Pad(Fmt.Int(instruction, 8),3,'0'));*)
      END;
    END;
  END Command;


(* ----------------------------------------------------------------------
   Procedure : Init

   Create objects on the bus. INITIALIZE the bus.

   The Bus module is NOT responsible for initialising teletype or tape
   because of the associated threads which must be spawned by the control
   module.
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T) : T =
  BEGIN
    self.extdmem := NEW(ExtdMem.T).init();
    self.interrupts := NEW(Interrupts.T).init();

    self.disc := NEW(Device.T).init();
    self.oscilloscope := NEW(Device.T).init();

    self.teletype := NEW(Teletype.T);
    self.tape := NEW(Tape.T);

    RETURN self;
  END Init;


(* ----------------------------------------------------------------------
   Procedure : Reset

   Implement the PDP8 INITIALIZE signal, by calling the reset methods of
   all attached devices.

   NB NOT TO BE CONFUSED WITH THE INITIALIZATION METHOD WHICH SETS UP THE
      OBJECTS AT THE START OF THE PROGRAM

   NB when extra objects are added to the bus, an entry must be added to
      this procedure to run their reset method.
   ---------------------------------------------------------------------- *)
PROCEDURE Reset(self : T) =
  BEGIN

    (* NB definition uncertain at this point - these two may or may not *)
    (* be required to implement the INITIALIZE signal *)
    self.extdmem.reset();
    self.interrupts.reset();

    self.disc.reset();
    self.oscilloscope.reset();
    self.teletype.reset();
    self.tape.reset();

  END Reset;


(* ----------------------------------------------------------------------
   Procedure : DoCAF

   Execute the CAF command:

   AC := 0
   L := 0
   Disable interrupts and clear all interrupt pending flags.
   then call the reset methods for all objects on the bus.
   ---------------------------------------------------------------------- *)
PROCEDURE DoCAF(bus : T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Interrupt, "CAF");*) 

    cpu.AC := 0;
    cpu.L := FALSE;

    cpu.interrupt.reset();

    bus.reset();
  END DoCAF;


BEGIN
END Bus.



