(* ----------------------------------------------------------------------
   $Id: CharDevice.i3,v 1.6 1995/06/23 07:45:36 sgp13 Exp $

   This provides the object on which Teletype.T and Tape.T are built and
   some of the common code. The type consists of a reader object and a writer
   object and the code to initialise and reset the object.
   It should ideally supply more common code too.
   ----------------------------------------------------------------------

   $Log: CharDevice.i3,v $
# Revision 1.6  1995/06/23  07:45:36  sgp13
# Update forced
#
# Revision 1.5  1995/06/19  13:57:09  sgp13
# Altered to allow use of Typescript on GUI
#
# Revision 1.4  1995/03/23  12:02:26  sgp13
# Patch to allow initialisation to attach to stdin and stdout if required
#
# Revision 1.3  1995/03/15  14:44:18  sgp13
# Interrupts now implemented using IntBus.T
#
# Revision 1.2  1995/03/14  12:14:20  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/03/07  23:06:25  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
INTERFACE CharDevice;

IMPORT Thread;
IMPORT CPU;
IMPORT CharReader, CharWriter, Device;


TYPE
  CommandProc = PROCEDURE(device : T; cpu : CPU.T)
                    RAISES { Device.CommandUnimplemented, Thread.Alerted };
  CommandArray = ARRAY [0..7] OF CommandProc;

  T = Device.T OBJECT
    reader : CharReader.T;
    writer : CharWriter.T;
  METHODS
    (* the special initialisation method for a CharStream *)
    (* teletype flags whether this device is the teletype *)
    (* stdio flags whether this device is to be connected to standard IO *)
    csInit(readerThread : CharReader.T; cpu : CPU.T; teletype : BOOLEAN;
           filter : CharWriter.Filter) : T
           := Init;
  OVERRIDES
    reset := Reset;
  END;

PROCEDURE Init(self : T; readerThread : CharReader.T; cpu : CPU.T;
               teletype : BOOLEAN; filter : CharWriter.Filter) : T;
PROCEDURE Reset(self : T);


(* ----------------------------------------------------------------------
   Procedure : ReaderSkipFlag

   Procedure to skip the next instruction if the ready flag of the reading
   device is set.

   This procedure implements KSF (6031) and RSF (6011)
   ---------------------------------------------------------------------- *)
PROCEDURE ReaderSkipFlag(device : T; cpu : CPU.T);


(* ----------------------------------------------------------------------
   Procedure : Unimplemented

   Unimplemented command
   ---------------------------------------------------------------------- *)
PROCEDURE Unimplemented(<* UNUSED *> device : T; <* UNUSED *> cpu : CPU.T)
    RAISES { Device.CommandUnimplemented };


(* ----------------------------------------------------------------------
   Procedure : WriterClearFlag

   Procedure to clear the flag of the writing device

   This procedure implements TCF (6042) and PCF (6022)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterClearFlag(device : T; cpu : CPU.T);


(* ----------------------------------------------------------------------
   Procedure : WriterLoadSequence

   Perform ClearFlag and then PunchCharacter

   This procedure implements TLS (6046) and PLS (6026)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterLoadSequence(device : T; cpu : CPU.T) RAISES {Thread.Alerted};


(* ----------------------------------------------------------------------
   Procedure : WriterPunchCharacter

   Procedure to output a character to the writer. The character output is
   extracted from bits 4-11 (the low 8 bits) of AC

   This procedure implements TPC (6044) and PPC (6024)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterPunchCharacter(device: T; cpu: CPU.T) RAISES {Thread.Alerted};


(* ----------------------------------------------------------------------
   Procedure : WriterSkipFlag

   Procedure to skip the next instruction if the ready flag of the writing
   device is set.

   This procedure implements TSF (6041) and PSF (6021)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterSkipFlag(device : T; cpu : CPU.T);


(* ----------------------------------------------------------------------
   Procedure : ClearReaderReadyFlag

   Clear a reader's ready flag, and cancel the associated interrupt
   if appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE ClearReaderReadyFlag(device : T; cpu : CPU.T);


(* ----------------------------------------------------------------------
   Procedure : ClearWriterReadyFlag

   Clear a writer's ready flag, and cancel the associated interrupt
   ---------------------------------------------------------------------- *)
PROCEDURE ClearWriterReadyFlag(device : T; cpu : CPU.T);


(* ----------------------------------------------------------------------
   Procedure : SetReaderReadyFlag

   Set a reader's ready flag, and raise the associated interrupt if appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE SetReaderReadyFlag(device : T; cpu : CPU.T);


(* ----------------------------------------------------------------------
   Procedure : SetWriterReadyFlag

   Set a writer's ready flag, and raise the associated interrupt if appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE SetWriterReadyFlag(device : T; cpu : CPU.T);





END CharDevice.









