(* ----------------------------------------------------------------------
   $Id: CharDevice.m3,v 1.6 1995/06/23 07:45:10 sgp13 Exp $

   This provides the object on which Teletype.T and Tape.T are built and
   some of the common code. The type consists of a reader object and a writer
   object and the code to initialise and reset the object.
   It should ideally supply more common code too.
   ----------------------------------------------------------------------

   $Log: CharDevice.m3,v $
# Revision 1.6  1995/06/23  07:45:10  sgp13
# Registers are now accessed directly
#
# Revision 1.5  1995/06/19  13:57:28  sgp13
# Altered to allow use of Typescript on GUI
#
# Revision 1.4  1995/03/23  12:02:37  sgp13
# Patch to allow initialisation to attach to stdin and stdout if required
#
# Revision 1.3  1995/03/15  14:44:30  sgp13
# Interrupts now implemented using IntBus.T
#
# Revision 1.2  1995/03/14  12:14:35  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/03/07  23:06:36  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE CharDevice;

IMPORT Thread, Word, Wr;
IMPORT Bits, CPU, Diag;
IMPORT Device, CharReader, CharWriter, IntBus;

<* FATAL IntBus.BadIndex *>

(* ----------------------------------------------------------------------
   Procedure : ClearReaderReadyFlag

   Clear a reader's ready flag, and cancel the associated interrupt
   if appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE ClearReaderReadyFlag(device : T; cpu : CPU.T) =
  BEGIN
    LOCK device.reader.readyLock DO
      device.reader.ready := FALSE;
      IF device.reader.interruptFlag THEN
        cpu.interrupt.cancel(device.reader.interruptID);
      END;
    END;
  END ClearReaderReadyFlag;


(* ----------------------------------------------------------------------
   Procedure : ClearWriterReadyFlag

   Clear a writer's ready flag, and cancel the associated interrupt
   ---------------------------------------------------------------------- *)
PROCEDURE ClearWriterReadyFlag(device : T; cpu : CPU.T) =
  BEGIN
    LOCK device.writer DO
      device.writer.ready := FALSE;
      IF device.reader.interruptFlag THEN
        cpu.interrupt.cancel(device.writer.interruptID);
      END;
    END;
  END ClearWriterReadyFlag;


(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise, by saving the reference to the reader thread, and
   instantiating the writer object.
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T; readerThread : CharReader.T; cpu : CPU.T;
               teletype: BOOLEAN; filter : CharWriter.Filter := NIL) : T =
  BEGIN
    self.reader := readerThread;
    self.writer := NEW(CharWriter.T).init(cpu, teletype, filter);

    self.reader.interruptFlag := TRUE;

    RETURN self;
  END Init;


(* ----------------------------------------------------------------------
   Procedure : Reset

   Reset the object: set the interrupt flag and clear the ready flags
   ---------------------------------------------------------------------- *)
PROCEDURE Reset(self : T) =
  BEGIN
    self.reader.interruptFlag := TRUE;
    LOCK self.reader.readyLock DO
      self.reader.ready := FALSE;
    END;
    LOCK self.writer DO
      self.writer.ready := FALSE;
    END;
  END Reset;


(* ----------------------------------------------------------------------
   Procedure : ReaderSkipFlag

   Procedure to skip the next instruction if the ready flag of the reading
   device is set.

   This procedure implements KSF (6031) and RSF (6011)
   ---------------------------------------------------------------------- *)
PROCEDURE ReaderSkipFlag(device : T; cpu : CPU.T) =
  BEGIN
    LOCK device.reader.readyLock DO
      IF device.reader.ready THEN
        (*Diag.Message(Diag.T.CharDevice, "{K|R}SF: skip taken");*)
        IF cpu.PC = LAST(CPU.Register_12) THEN
          cpu.PC := 0;
        ELSE
          INC(cpu.PC);
        END;
      ELSE
        (*Diag.Message(Diag.T.CharDevice, "{K|R}SF: skip not taken");*)
      END;
    END;
  END ReaderSkipFlag;


(* ----------------------------------------------------------------------
   Procedure : SetReaderReadyFlag

   Set a reader's ready flag, and raise the associated interrupt if appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE SetReaderReadyFlag(device : T; cpu : CPU.T) =
  BEGIN
    LOCK device.reader.readyLock DO
      device.reader.ready := TRUE;
      IF device.reader.interruptFlag THEN
        cpu.interrupt.request(device.reader.interruptID);
      END;
    END;
  END SetReaderReadyFlag;


(* ----------------------------------------------------------------------
   Procedure : SetWriterReadyFlag

   Set a writer's ready flag, and raise the associated interrupt if appropriate
   ---------------------------------------------------------------------- *)
PROCEDURE SetWriterReadyFlag(device : T; cpu : CPU.T) =
  BEGIN
    LOCK device.writer DO
      device.writer.ready := TRUE;
      IF device.reader.interruptFlag THEN
        cpu.interrupt.request(device.writer.interruptID);
      END;
    END;
  END SetWriterReadyFlag;


(* ----------------------------------------------------------------------
   Procedure : Unimplemented

   Unimplemented command
   ---------------------------------------------------------------------- *)
PROCEDURE Unimplemented(<* UNUSED *> device : T; <* UNUSED *> cpu : CPU.T)
    RAISES { Device.CommandUnimplemented } =
  BEGIN
    RAISE Device.CommandUnimplemented;
  END Unimplemented;


(* ----------------------------------------------------------------------
   Procedure : WriterClearFlag

   Procedure to clear the flag of the writing device. If the flag is currently
   set and the device's interrupts are enabled, then cancel the interrupt

   This procedure implements TCF (6042) and PCF (6022)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterClearFlag(device : T; cpu : CPU.T) = 
  BEGIN
    (*Diag.Message(Diag.T.CharDevice, "{T|P}CF");*)
    ClearWriterReadyFlag(device, cpu);
  END WriterClearFlag;


(* ----------------------------------------------------------------------
   Procedure : WriterLoadSequence

   Perform WriterClearFlag and then WriterPunchCharacter

   This procedure implements TLS (6046) and PLS (6026)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterLoadSequence(device : T; cpu : CPU.T) RAISES {Thread.Alerted}=
  BEGIN
    WriterClearFlag(device, cpu);
    WriterPunchCharacter(device, cpu);
  END WriterLoadSequence;


(* ----------------------------------------------------------------------
   Procedure : WriterPunchCharacter

   Procedure to output a character to the writer. The character output is
   extracted from bits 4-11 (the low 8 bits) of AC. If the writer is the
   teletype then clear the top bit (bit 4) as the internal representation
   always has this bit set.

   This procedure implements TPC (6044) and PPC (6024)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterPunchCharacter(device: T; cpu: CPU.T) RAISES {Thread.Alerted}=
  VAR
    ch : INTEGER; (* output character *)
  BEGIN
    (*Diag.Message(Diag.T.CharDevice, "{T|P}PC");*)
    TRY
      (* if there is a stream attached to this writer, then write the *)
      (* character and then set the ready flag and possibly raise an *)
      (* interrupt *)
      (* if the write fails then the device effectively stalls *)
      ch := Word.And(cpu.AC, Bits.lowByte);
      IF device.writer.teletype THEN
        ch := Word.And(ch, Bits.clearBitFour);
      END;
      device.writer.put(VAL(ch, CHAR));
      SetWriterReadyFlag(device, cpu);
    EXCEPT
    | Wr.Failure =>
             (*Diag.Message(Diag.T.CharDevice, "write failed: device stalled");*)
    | CharWriter.NotConnected =>
    END;
  END WriterPunchCharacter;


(* ----------------------------------------------------------------------
   Procedure : WriterSkipFlag

   Procedure to skip the next instruction if the ready flag of the writing
   device is set.

   This procedure implements TSF (6041) and PSF (6021)
   ---------------------------------------------------------------------- *)
PROCEDURE WriterSkipFlag(device : T; cpu : CPU.T) =
  BEGIN
    LOCK device.writer DO
      IF device.writer.ready THEN
        (*Diag.Message(Diag.T.CharDevice, "{T|P}SF: skip taken");*)
        IF cpu.PC = LAST(CPU.Register_12) THEN
          cpu.PC := 0;
        ELSE
          INC(cpu.PC);
        END;
      ELSE
        (*Diag.Message(Diag.T.CharDevice, "{T|P}SF: skip not taken");*)
      END;
    END;
  END WriterSkipFlag;


BEGIN
END CharDevice.



