(* ----------------------------------------------------------------------
   $Id: CharReader.m3,v 1.5 1995/06/19 13:59:17 sgp13 Exp $

   Interface supplying the character reader object and associated thread
   applying procedure. This object implements reading characters from an
   attached input stream for both the Teletype and Tape objects
   ----------------------------------------------------------------------

   $Log: CharReader.m3,v $
# Revision 1.5  1995/06/19  13:59:17  sgp13
# Reader is now a CharRd.T type standard interface to either a BufferRd.T or
# a StreamRd.T
#
# Revision 1.4  1995/05/03  21:29:35  sgp13
# Patched to work for Focal:
#
# 1) delay inserted to allow time for interrupt routine to process characters
# 2) linefeeds mapped to carriage returns
#
# The delay can be altered by specifying in the initialisation file.
#
# Revision 1.3  1995/03/15  14:44:59  sgp13
# Interrupts now implemented using IntBus.T
#
# Revision 1.2  1995/03/14  12:15:05  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/02/24  23:27:15  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE CharReader;

IMPORT ASCII, Lex, Rd, Scan, Text, Thread, Word;
IMPORT Bits, CharRd, Diag, CPU, Flags, Init;

VAR
  (* delay between accepting characters from input *)
  (* needed to ensure that the character reading routine in Focal *)
  (* has sufficient programming time *)
  (* this may be set in the [CharReader] section of the initialisation file *)
  (* using the line Delay=<time>. this line is assumed to be the only line *)
  (* in this section *)
  interCharDelay : LONGREAL := 2.0D0; (* in seconds *)

  initLine : Init.T;  (* line from initialisation file *)

(* ----------------------------------------------------------------------
   Procedure : Connect

   Connect the specified file stream to the reader. Raise an exception if
   there's already a stream connected.
   ---------------------------------------------------------------------- *)
PROCEDURE Connect(self: T; input: CharRd.T) RAISES { CharRd.AlreadyConnected }=
  BEGIN
    (* make sure that no-one's accessing the stream *)
    IF self.reader = NIL THEN
      self.reader := input;
    ELSE
      RAISE CharRd.AlreadyConnected;
    END;
  END Connect;


(* ----------------------------------------------------------------------
   Procedure : Disconnect

   Disconnect the stream from the reader. Return the stream to the caller,
   so that they can use it.
   ---------------------------------------------------------------------- *)
PROCEDURE Disconnect(self : T) : CharRd.T =
  VAR
    temp : CharRd.T;
  BEGIN
    temp := self.reader;
    self.reader := NIL;
    RETURN temp;
  END Disconnect;


(* ----------------------------------------------------------------------
   Procedure : DoCharReader

   Thread application procedure. Loop: wait for signal. if stream attached
   then read from it and set the ready flag (optionally cause interrupt)
   Continue until alerted and then exit.
   ---------------------------------------------------------------------- *)
PROCEDURE DoCharReader(self : T) : REFANY RAISES {} =
  VAR
    ch : CHAR;
  BEGIN
    TRY
      WHILE NOT Thread.TestAlert() DO
        self.run.wait();     (* wait until requested to fetch character *)
        TRY
          (* NB there's a possible race condition here *)
          IF self.reader # NIL THEN
            ch := self.reader.get();

            (* on a teletype: map LF to CR *)
            IF self.teletype AND ch = ASCII.NL THEN
              ch := ASCII.CR;
            END;
                
            (* map characters to integers *)
            self.byte := ORD(ch);

            (* the PDP8/e teletype always sets the top bit of characters *)
            (* it reads *)
            IF self.teletype THEN
              self.byte := Word.Or(self.byte, Bits.setBitFour);
            END;

            LOCK self.readyLock DO
              self.ready := TRUE;
              IF self.interruptFlag THEN
                self.cpu.interrupt.request(self.interruptID);
              END;
            END;
          END;
        EXCEPT
        | Rd.EndOfFile, Rd.Failure, CharRd.NotConnected =>
        END;

        (* it is necessary to wait between characters to ensure that the *)
        (* interrupt based processing routine in Focal has time to process *)
        (* them. this delay may be set in the initialisation file, *)
        (* to be tailored to how fast the system is running. *)
        (* the time is in seconds. *)
        Thread.AlertPause(interCharDelay);
      END;
    EXCEPT
    | Thread.Alerted =>
    END;

    RETURN self;
  END DoCharReader;


(* ----------------------------------------------------------------------
   Procedure : ThreadInit

   Initialise the thread closure object. Instantiate the objects and set
   up the pointer to the CPU object. Allocate and interrupt ID to the device
   ---------------------------------------------------------------------- *)
PROCEDURE ThreadInit(self : T; cpu : CPU.T; teletype: BOOLEAN) : T =
  BEGIN
    self.cpu := cpu;

    self.interruptID := cpu.interrupt.allocate();

    self.readyLock := NEW(MUTEX);
    self.run := NEW(Flags.Signal).init();

    self.teletype := teletype;  (* set flag whether teletype: and hence *)
                                (* whether to set the top bit of any *)
                                (* character read *)
    RETURN self;
  END ThreadInit;



BEGIN
  (* look in the initialisation file to see whether a delay has been set *)
  (* to use between accepting characters for input *)
  Init.Reset();

  TRY
    initLine := Init.Read("CharReader");

    IF Text.Equal(initLine.name, "Delay") THEN
      interCharDelay := Scan.LongReal(initLine.value);
    END;

  EXCEPT
    Init.Empty =>
  | Lex.Error => (*Diag.Message(Diag.T.Error, "invalid number format in " &
                                            "[CharReader]Delay in init file");*)
  END;

END CharReader.








