(* ----------------------------------------------------------------------
   $Id: CharWriter.m3,v 1.4 1995/06/19 14:00:18 sgp13 Exp $

   The object which deals with writing characters. It is more straightforward
   than the reader, as characters are assumed to get written immediately
   ----------------------------------------------------------------------

   $Log: CharWriter.m3,v $
# Revision 1.4  1995/06/19  14:00:18  sgp13
# Optional filter added to remove unwanted characters from output stream
#
# Revision 1.3  1995/03/15  14:45:22  sgp13
# Interrupts now implemented using IntBus.T
#
# Revision 1.2  1995/03/14  12:15:40  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/03/07  23:07:00  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE CharWriter;

IMPORT ASCII, Thread, Wr;
IMPORT CPU, Diag;

REVEAL
  T = PublicT BRANDED OBJECT
    stream : Wr.T := NIL;
    filter : Filter := DefaultFilter;
  OVERRIDES
    connect := Connect;
    disconnect := Disconnect;
    init := Init;
    put := Put;
  END;


(* ----------------------------------------------------------------------
   Procedure : Connect

   Connect file to the writer
   ---------------------------------------------------------------------- *)
PROCEDURE Connect(self : T; file : Wr.T) RAISES { AlreadyConnected } =
  BEGIN
    (* make sure no-one is attempting to write to the writer *)
    LOCK self DO
      IF self.stream # NIL THEN
        RAISE AlreadyConnected;
      ELSE
        self.stream := file;
      END;
    END;
  END Connect;


(* ----------------------------------------------------------------------
   Procedure : DefaultFilter

   Default filter : no action, just return the character supplied
   ---------------------------------------------------------------------- *)
PROCEDURE DefaultFilter(c : CHAR) : CHAR =
  BEGIN
    RETURN c;
  END DefaultFilter;


(* ----------------------------------------------------------------------
   Procedure : Disconnect

   Disconnect file : flushing output to disc
   ---------------------------------------------------------------------- *)
PROCEDURE Disconnect(self : T) : Wr.T RAISES { Thread.Alerted } =
  VAR
    temp : Wr.T;
  BEGIN
    (* make sure no-one is writing to the file *)
    LOCK self DO
      IF self.stream # NIL THEN
        TRY
          Wr.Flush(self.stream);
        EXCEPT
        | Wr.Failure =>
                (*Diag.Message(Diag.T.Error, "failed to flush output to disc");*)
        END;
      END;

      temp := self.stream;
      self.stream := NIL;
    END;

    RETURN temp;
  END Disconnect;

(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise, by setting teletype flag and by allocating an interrupt
   number for the device.
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T; cpu : CPU.T; teletype : BOOLEAN; filter: Filter) : T =
  BEGIN
    self.teletype := teletype;
    self.interruptID := cpu.interrupt.allocate();

    IF filter # NIL THEN
      self.filter := filter;
    END;

    RETURN self;
  END Init;


(* ----------------------------------------------------------------------
   Procedure : Put

   Output character, first having filtered it
   ---------------------------------------------------------------------- *)
PROCEDURE Put(self : T; c : CHAR)
  RAISES { NotConnected, Wr.Failure, Thread.Alerted } =
  BEGIN
    IF self.stream = NIL THEN
      RAISE NotConnected;
    ELSE
      WITH filteredC = self.filter(c) DO
        IF filteredC # ASCII.NUL THEN
          Wr.PutChar(self.stream, self.filter(filteredC));
          Wr.Flush(self.stream);
        END
      END;
    END;
  END Put;


BEGIN
END CharWriter.
