(* ----------------------------------------------------------------------
   $Id: ExtdMem.m3,v 1.5 1995/06/24 21:54:06 sgp13 Exp $

   The module implementing the extended memory instructions 62xx

   The instructions are of the following form:

   62x4 are general housekeeping instructions, and are not field-specific.

       6214 : Read Data Field
       6224 : Read Instruction Field
       6234 : Read Interrupt Buffer
       6244 : Restore Memory Field

   62N1, 62N2, and 62N3 are instructions which specify a field (N):

       62N1 : Change to Data Field N
       62N2 : Change to Instruction Field N
       62N3 : CDF N, CIF N combined

   ----------------------------------------------------------------------

   $Log: ExtdMem.m3,v $
# Revision 1.5  1995/06/24  21:54:06  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.4  1995/06/23  07:45:51  sgp13
# Registers are now accessed directly
#
# Revision 1.3  1995/06/20  10:05:02  sgp13
# Changes due to version 1.2 of IntBus
#
# Revision 1.2  1995/03/15  13:11:53  sgp13
# Interrupts now use IntBus.T
#
# Revision 1.1  1995/02/24  22:17:02  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE ExtdMem;

IMPORT Fmt, Word;
IMPORT Bits, CPU, Diag, Memory, Registers;
IMPORT Device;

<* FATAL Registers.OutOfRange *>

TYPE
  ExtdMemCommand = PROCEDURE(cpu : CPU.T);

(* array indexing the 62x4 housekeeping instructions *)
CONST
  array = ARRAY [1..4] OF ExtdMemCommand { DoRDF, DoRIF, DoRIB, DoRMF };

(* ----------------------------------------------------------------------
   Procedure : Command

   Route the commands invoked to the bus to the appropriate procedures
   ---------------------------------------------------------------------- *)
PROCEDURE Command(<* UNUSED *> self : T; cpu : CPU.T; devselB : CPU.OpCode;
                  command : CPU.OpCode )
    RAISES { Device.CommandUnimplemented } =
  BEGIN
    CASE command OF
    (* the field specific commands *)
    | 1 => DoCDF(cpu, devselB);
    | 2 => DoCIF(cpu, devselB);

    | 3 => DoCDF(cpu, devselB);
           DoCIF(cpu, devselB);

    (* the general housekeeping commands - or unimplemented command *)
    | 4 => IF devselB > 0 AND devselB < 5 THEN
             array[devselB](cpu);
           ELSE
             RAISE Device.CommandUnimplemented;
           END;

    (* anything else is unimplemented *)
    ELSE
      RAISE Device.CommandUnimplemented;
    END;
  END Command;


(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise the extended memory object
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T) : Device.T =
  BEGIN
    RETURN self;
  END Init;


(* ----------------------------------------------------------------------
   Procedure : Reset

   Propagate the INITIALIZE pulse to the extended memory device.
   ---------------------------------------------------------------------- *)
PROCEDURE Reset(<* UNUSED *> self : T) =
  BEGIN
  END Reset;


(* ----------------------------------------------------------------------
   Procedure : DoCDF

   Change data field to that specified
   ---------------------------------------------------------------------- *)
PROCEDURE DoCDF(cpu : CPU.T; field : Memory.FieldNumber) =
  BEGIN
    (*Diag.Message(Diag.T.ExtdMem, "CDF " & Fmt.Int(field));*)
    cpu.EM_DFnum := field;
    cpu.EM_DFptr := cpu.memory.mem[field];
  END DoCDF;


(* ----------------------------------------------------------------------
   Procedure : DoCIF

   Change instruction buffer to the field specified. Inhibit interrupts
   until the next JMS / JMP, at which point, uninhibit interrupts and
   transfer IB to IF.
   ---------------------------------------------------------------------- *)
PROCEDURE DoCIF(cpu : CPU.T; field : Memory.FieldNumber) =
  BEGIN
    (*Diag.Message(Diag.T.ExtdMem, "CIF " & Fmt.Int(field));*)
    cpu.EM_IB := field;
    cpu.interrupt.uninhibited := FALSE;
  END DoCIF;


(* ----------------------------------------------------------------------
   Procedure : DoRDF (6214)

   OR the value of DF into AC 6-8
   ---------------------------------------------------------------------- *)
PROCEDURE DoRDF(cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.ExtdMem, "RDF");*)
    WITH AC = cpu.AC DO
      AC := Word.Or(AC, cpu.EM_DFnum * 8_10);
    END;
  END DoRDF;


(* ----------------------------------------------------------------------
   Procedure : DoRIF (6224)

   OR the value of IF into AC 6-8
   ---------------------------------------------------------------------- *)
PROCEDURE DoRIF(cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.ExtdMem, "RIF");*)
    WITH AC = cpu.AC DO
      AC := Word.Or(AC, cpu.EM_IFnum * 8_10);
    END;
  END DoRIF;


(* ----------------------------------------------------------------------
   Procedure : DoRIB (6234)

   OR the value of the save field register into AC 6-11

   6-8 becomes saved IF
   9-11 becomes saved DF

   bit 5 is cleared (should be set to value of user flip-flop)
   ---------------------------------------------------------------------- *)
PROCEDURE DoRIB(cpu : CPU.T) =
  VAR
    ACtemp : INTEGER;
  BEGIN
    (*Diag.Message(Diag.T.ExtdMem, "RIB");*)
    WITH AC = cpu.AC DO
      AC := Word.And(Word.Or(AC, cpu.EM_SIF * 8_10 + cpu.EM_SDF),
                     Bits.clearBitFive);

    END;
  END DoRIB;


(* ----------------------------------------------------------------------
   Procedure : DoRMF (6244)

   Copy save field registers to IB and DF. Inhibit interrupts so that
   IB is copied to IF at next JMP / JMS

   ---------------------------------------------------------------------- *)
PROCEDURE DoRMF(cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.ExtdMem, "RMF");*)
    cpu.EM_IB := cpu.EM_SIF;
    cpu.EM_DFnum := cpu.EM_SDF;

    cpu.interrupt.uninhibited := FALSE;

  END DoRMF;


BEGIN
END ExtdMem.







