(* ----------------------------------------------------------------------
   $Id: Interrupts.m3,v 1.6 1995/06/24 21:54:58 sgp13 Exp $

   Interrupt services module. Provides the implementation for instructions
   6000-6007
   ----------------------------------------------------------------------

   $Log: Interrupts.m3,v $
# Revision 1.6  1995/06/24  21:54:58  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.5  1995/06/23  07:48:35  sgp13
# Registers are now accessed directly
#
# Revision 1.4  1995/06/20  10:05:53  sgp13
# Changes due to version 1.2 of IntBus
#
# Revision 1.3  1995/03/15  13:10:23  sgp13
# Interrupts now use IntBus.T
#
# Revision 1.2  1995/03/14  12:17:26  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/02/24  22:17:11  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Interrupts;

IMPORT Word;
IMPORT Bits, CPU, Diag, Registers;
IMPORT Device;

<* FATAL Registers.OutOfRange *>

TYPE
  InterruptCommand = PROCEDURE(cpu : CPU.T)
                         RAISES { Device.CommandUnimplemented };

(* the array of command procedures *)
CONST
  array = ARRAY [0..6] OF InterruptCommand {
                 IOT_6000, IOT_6001, IOT_6002, IOT_6003,
                 IOT_6004, IOT_6005, IOT_6006 };


(* ----------------------------------------------------------------------
   Procedure : Command

   Routing procedure which passes control to the appropriate command.
   ---------------------------------------------------------------------- *)
PROCEDURE Command(<* UNUSED *> self : T;
                               cpu : CPU.T;
                  <* UNUSED *> devselB : CPU.OpCode;
                               command : CPU.OpCode)
    RAISES { Device.CommandUnimplemented } =
  BEGIN
    array[command](cpu);
  END Command;


(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise the interrupt object - no action.
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T) : Device.T =
  BEGIN
    RETURN self;
  END Init;


(* ----------------------------------------------------------------------
   Procedure : Reset

   Respond to INITIALIZE pulse - no action, dealt with elsewhere
   ---------------------------------------------------------------------- *)
PROCEDURE Reset(<* UNUSED *> self : T) =
  BEGIN
  END Reset;
                              

(* ----------------------------------------------------------------------
   Procedure : IOT_6000 (SKON)

   Skip next instruction if interrupts are enabled. Side effect of this
   instruction is also to disable interrupts if they are enabled - see IOF
   for details.

   This skip is taken either if interrupts are enabled or if they are
   scheduled to be at the end of this instruction.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6000(cpu : CPU.T) =
  BEGIN
    IF cpu.interrupt.areEnabled() THEN
      (*Diag.Message(Diag.T.Interrupt, "SKON - skip taken");*)
      IF cpu.PC = LAST(CPU.Register_12) THEN
        cpu.PC := 0;
      ELSE
        INC(cpu.PC);
      END;
      cpu.interrupt.disable();
    ELSE
      (*Diag.Message(Diag.T.Interrupt, "SKON - skip not taken");*)
    END;
  END IOT_6000;


(* ----------------------------------------------------------------------
   Procedure : IOT_6001 (ION)

   If interrupts not enabled, enable them after the next instruction.
   This is implemented by
   setting a flag which is read after the interrupt state is tested at the
   start of the next cycle.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6001(cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Interrupt, "ION");*)
    WITH interrupt = cpu.interrupt DO
      IF NOT interrupt.areEnabled() THEN
        interrupt.enable();
      END;
    END;
  END IOT_6001;


(* ----------------------------------------------------------------------
   Procedure : IOT_6002 (IOF)

   Disable interrupts if they are enabled. Do this by clearing all three
   interrupt enabling flags. This means that the sequence ION IOF leaves
   interrupts off.

   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6002(cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Interrupt, "IOF");*)
    cpu.interrupt.disable();
  END IOT_6002;


(* ----------------------------------------------------------------------
   Procedure : IOT_6003 (SRQ)

   Skip next instruction if one or more devices is requesting an interrupt.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6003(cpu : CPU.T) =
  BEGIN
    IF cpu.interrupt.arePending() THEN
      IF cpu.PC = LAST(CPU.Register_12) THEN
        cpu.PC := 0;
      ELSE
        INC(cpu.PC);
      END;
      (*Diag.Message(Diag.T.Interrupt, "SRQ - skip taken");*)
    ELSE
      (*Diag.Message(Diag.T.Interrupt, "SRQ - skip not taken");*)
    END;
  END IOT_6003;


(* ----------------------------------------------------------------------
   Procedure : IOT_6004 (GTF)

   Read flags into the accumulator as follows:

   bits (0 is LSB)  information
   0                link
   1                0 - supposed to be GT flag - not implemented
   2                interrupt bus - ie whether an interrupt is pending
   3                0 - should be inhibit flag - but the original h/w did not
                        implement this
   4                interrupt enable flag
   5                0 - supposed to be time share but this is not implemented
   6-8              SIF - saved value of the instruction field before interrupt
   9-11             SDF - saved value of the data field before interrupt
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6004(cpu : CPU.T) =
  VAR
    bitArray : Bits.Array;
    ACtemp : INTEGER;
  BEGIN
    (*Diag.Message(Diag.T.Interrupt, "GTF");*)

    (* read information into the bit array *)
    bitArray[0] := cpu.L;
    bitArray[1] := FALSE;

    (* lock the interrupts while we read status *)
    bitArray[2] := cpu.interrupt.arePending();
    bitArray[3] := FALSE;
    bitArray[4] := cpu.interrupt.areEnabled();

    bitArray[5] := FALSE;

    (* clear the rest of the bits *)
    FOR i := 6 TO 11 DO
      bitArray[i] := FALSE;
    END;

    cpu.AC := Bits.Implode(bitArray) + cpu.EM_SIF * 8_10 + cpu.EM_SDF;

  END IOT_6004;


(* ----------------------------------------------------------------------
   Procedure : IOT_6005 (RTF)

   The reverse of GTF above, except:

   bit 2 is ignored
   interrupts are enabled regardless of the state of bit 4
   the contents of bits 6-8 and 9-11 are written to IB and DF respectively,
   NOT their SF register counterparts.

   Also inhibit interrupts as usual for anything setting IB. When a JMS or
   JMP is next issued, the contents of IB are transferred to IF and
   interrupts are uninhibited.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6005(cpu : CPU.T) =
  VAR
    bitArray : Bits.Array;
  BEGIN
    (*Diag.Message(Diag.T.Interrupt, "RTF");*)

    WITH AC = cpu.AC DO
      (* distribute the bits into the array *)
      bitArray := Bits.Explode(AC);

      (* write IB and DF *)
      cpu.EM_IB := (AC DIV 8_10) MOD 8_10;
      cpu.EM_DFnum := AC MOD 8_10;
    END;

    (* interpret the bits in the array *)
    cpu.L := bitArray[0];

    (* RTF always performs ION plus inhibiting interrupts *)
    WITH interrupt = cpu.interrupt DO
      interrupt.uninhibited := FALSE;
      interrupt.enable();
    END;
  END IOT_6005;


(* ----------------------------------------------------------------------
   Procedure : IOT_6006 {SGT}

   Not implemented. Supposed to skip if the GT flag in the unimplemented
   EAE is set.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6006(<* UNUSED *> cpu : CPU.T)
    RAISES { Device.CommandUnimplemented } =
  BEGIN
    (*Diag.Message(Diag.T.Interrupt, "SGT - unimplemented");*)
    RAISE Device.CommandUnimplemented;
  END IOT_6006;

BEGIN
END Interrupts.
