(* ----------------------------------------------------------------------
   $Id: Tape.m3,v 1.4 1995/06/23 07:49:41 sgp13 Exp $

   This module implements the tape reader/punch device. The T type is
   derived from CharDevice, with an override of the command method.

   Much of the code for the implementation is supplied by CharDevice to
   allow sharing with Teletype
   ----------------------------------------------------------------------

   $Log: Tape.m3,v $
# Revision 1.4  1995/06/23  07:49:41  sgp13
# Registers are now accessed directly
#
# Revision 1.3  1995/03/15  14:46:06  sgp13
# Interrupts now implemented using IntBus.T
#
# Revision 1.2  1995/03/14  12:18:07  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/03/07  23:08:53  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Tape;

IMPORT Word, Thread;
IMPORT CPU, Diag, Registers;
IMPORT Device, CharDevice;

<* FATAL Registers.OutOfRange *>

(* the array for indexing commands to procedures *)
VAR
  array := ARRAY [1..2] OF CharDevice.CommandArray {
        CharDevice.CommandArray { IOT_6010, CharDevice.ReaderSkipFlag (*6011*),
                                 IOT_6012, CharDevice.Unimplemented (*6013*),
                                 IOT_6014, CharDevice.Unimplemented (*6015*),
                                 IOT_6016, CharDevice.Unimplemented (*6017*) },
        CharDevice.CommandArray { IOT_6020, CharDevice.WriterSkipFlag (*6021*),
                                 CharDevice.WriterClearFlag (*6022*),
                                 CharDevice.Unimplemented (*6023*),
                                 CharDevice.WriterPunchCharacter (*6024*),
                                 CharDevice.Unimplemented (*6025*),
                                 CharDevice.WriterLoadSequence (*6026*),
                                 CharDevice.Unimplemented (*6027*) }};


(* ----------------------------------------------------------------------
   Procedure : Command

   Indexes commands into array
   ---------------------------------------------------------------------- *)
PROCEDURE Command(self : T; cpu : CPU.T; devselB : CPU.OpCode;
                  command : CPU.OpCode) RAISES { Device.CommandUnimplemented,
                                                 Thread.Alerted } =
  BEGIN
    array[devselB, command](self, cpu);
  END Command;


(* ----------------------------------------------------------------------
   Procedure : IOT_6010  (RPE)

   Set the tape device interrupt flag - enabling interrupts. If the devices
   are ready then raise interrupts for them.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6010(tape : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Tape, "RPE");*)
    IF NOT tape.reader.interruptFlag THEN
      tape.reader.interruptFlag := TRUE;
      IF tape.reader.ready THEN
        cpu.interrupt.request(tape.reader.interruptID);
      END;
      IF tape.writer.ready THEN
        cpu.interrupt.request(tape.writer.interruptID);
      END;
    END;
  END IOT_6010;


(* ----------------------------------------------------------------------
   Procedure : IOT_6012  (RRB)

   Read reader buffer into AC 4-11. There's a possible race condition here
   as it is possible to write the reader byte w/o holding the readyLock
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6012(tape : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Tape, "RRB");*)
    WITH AC = cpu.AC DO
      AC := Word.Or(AC, ORD(tape.reader.byte));
    END;
    IF tape.reader.ready THEN
      CharDevice.ClearReaderReadyFlag(tape, cpu);
    END;

  END IOT_6012;


(* ----------------------------------------------------------------------
   Procedure : IOT_6014  (RFC)

   Signal to reader thread to read next character
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6014(tape : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Tape, "RFC");*)
    IF tape.reader.ready THEN
      CharDevice.ClearReaderReadyFlag(tape, cpu);
    END;
    tape.reader.run.signal();
  END IOT_6014;


(* ----------------------------------------------------------------------
   Procedure : IOT_6016  (no mnemonic)

   Combine RRB and RFC
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6016(tape : CharDevice.T; cpu : CPU.T) =
  BEGIN
    IOT_6012(tape, cpu);
    IOT_6014(tape, cpu);
  END IOT_6016;


(* ----------------------------------------------------------------------
   Procedure : IOT_6020  (PCE)

   Clear the tape device interrupt flag - disabling interrupts.
   If it was enabled then cancel interrupts: one for each of the ready flags
   which was set: to a maximum of two.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6020(tape : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Tape, "PCE");*)
    IF tape.reader.interruptFlag THEN
      tape.reader.interruptFlag := FALSE;
      IF tape.reader.ready THEN
        cpu.interrupt.cancel(tape.reader.interruptID);
      END;
      IF tape.writer.ready THEN
        cpu.interrupt.cancel(tape.writer.interruptID);
      END;
    END;
  END IOT_6020;


BEGIN
END Tape.












