(* ----------------------------------------------------------------------
   $Id: Teletype.m3,v 1.4 1995/06/23 07:49:56 sgp13 Exp $

   This module implements the teletype device on the PDP8/e bus. It actually
   consists of two devices, a keyboard and a punch/teleprinter. They are
   implemented as connections to user selected files / streams, with the
   eventual intention that they will be implemented as the terminal keyboard
   and screen.

   The T type is a specialisation of the CharDevice type which consists of
   a generalised character-by-character reader and writer, plus some common
   code.
   ----------------------------------------------------------------------

   $Log: Teletype.m3,v $
# Revision 1.4  1995/06/23  07:49:56  sgp13
# Registers are now accessed directly
#
# Revision 1.3  1995/03/15  14:46:38  sgp13
# Interrupts now implemented using IntBus.T
#
# Revision 1.2  1995/03/14  12:18:50  sgp13
# Interrupts method altered to counting pending interrupts.
#
# Revision 1.1  1995/03/07  23:09:15  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Teletype;

IMPORT Thread, Word;
IMPORT CPU, Diag, Registers;
IMPORT Device, CharDevice;

<* FATAL Registers.OutOfRange *>

(* the array for indexing commands to procedures *)
VAR
  array := ARRAY [3..4] OF CharDevice.CommandArray {
        CharDevice.CommandArray { IOT_6030, CharDevice.ReaderSkipFlag (*6031*),
                                 IOT_6032, CharDevice.Unimplemented (*6033*),
                                 IOT_6034, IOT_6035,
                                 IOT_6036, CharDevice.Unimplemented (*6037*) },
        CharDevice.CommandArray { IOT_6040, CharDevice.WriterSkipFlag (*6041*),
                                 CharDevice.WriterClearFlag (*6042*),
                                 CharDevice.Unimplemented (*6043*),
                                 CharDevice.WriterPunchCharacter (*6044*),
                                 IOT_6045, CharDevice.WriterLoadSequence,
                                 CharDevice.Unimplemented (*6047*) }};

(* ----------------------------------------------------------------------
   Procedure : Command

   Index the command into the array and execute it
   ---------------------------------------------------------------------- *)
PROCEDURE Command(self : T; cpu : CPU.T; devselB : CPU.OpCode;
                  command : CPU.OpCode) RAISES { Device.CommandUnimplemented,
                                                 Thread.Alerted } =
  BEGIN
    array[devselB, command](self, cpu);
  END Command;


(* ----------------------------------------------------------------------
   Procedure : IOT_6030 (* KCF *)

   If the ready flag is set: clear keyboard ready flag. If interrupts from this
                             device are enabled, then decrement the number of
                             pending interrupts.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6030(teletype : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Teletype, "KCF");*)
    IF teletype.reader.ready THEN
      CharDevice.ClearReaderReadyFlag(teletype, cpu);
    END;
  END IOT_6030;


(* ----------------------------------------------------------------------
   Procedure : IOT_6032 (* KCC *)

   Clear keyboard flag and AC, and request character to be read
   If the keyboard flag was set and the device interrupts were enabled then
   decrement the interrupt pending count.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6032(teletype : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Teletype, "KCC");*)
    cpu.AC := 0;
    IF teletype.reader.ready THEN
      CharDevice.ClearReaderReadyFlag(teletype, cpu);
    END;
    teletype.reader.run.signal();
  END IOT_6032;


(* ----------------------------------------------------------------------
   Procedure : IOT_6034 (* KRS *)

   Read (logical OR) next character into AC 4-11
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6034(teletype : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Teletype, "KRS");*)
    WITH AC = cpu.AC DO
      cpu.AC := Word.Or(cpu.AC, ORD(teletype.reader.byte));
    END;
  END IOT_6034;


(* ----------------------------------------------------------------------
   Procedure : IOT_6035 (* KIE *)

   Set or clear interrupt flag which dictates whether setting the ready
   flag raises an interrupt. The flag is set to the value of AC bit 11.

   If interrupts become enabled then one should be raised for each of the
   ready flags which is set. If they become disabled then one should be
   cancelled for each flag.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6035(teletype : CharDevice.T; cpu : CPU.T) =
  VAR
    target : BOOLEAN;
  BEGIN
    IF cpu.AC MOD 2 = 1 THEN
      (*Diag.Message(Diag.T.Teletype, "KIE-flag set");*)
      target := TRUE;
    ELSE
      (*Diag.Message(Diag.T.Teletype, "KIE-flag cleared");*)
      target := FALSE;
    END;

    LOCK teletype.reader.readyLock DO
      teletype.reader.interruptFlag := target;
    END;

    IF target = FALSE THEN
      (* if interrupts cleared then cancel the interrupts *)
      cpu.interrupt.cancel(teletype.reader.interruptID);
      cpu.interrupt.cancel(teletype.writer.interruptID);
    ELSE
      (* if interrupts set then raise the interrupts if flags are set *)
      IF teletype.reader.ready THEN
        cpu.interrupt.request(teletype.reader.interruptID);
      END;

      IF teletype.writer.ready THEN
        cpu.interrupt.request(teletype.writer.interruptID);
      END;
    END;
      
  END IOT_6035;


(* ----------------------------------------------------------------------
   Procedure : IOT_6036 (* KRB *)

   Read buffer and clear ready flag, signalling to load a new character.
   NB this CANNOT be a straight combination of KRS and KCC, as this would
   end up either with an empty AC or a race condition for the value of the
   byte actually read into AC

   NB If the ready flag was set and device interrupts are enabled then
      decrement the interrupt pending count.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6036(teletype : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Teletype, "KRB");*)

    cpu.AC := ORD(teletype.reader.byte);

    IF teletype.reader.ready THEN
      CharDevice.ClearReaderReadyFlag(teletype, cpu);
    END;
    teletype.reader.run.signal();
  END IOT_6036;


(* ----------------------------------------------------------------------
   Procedure : IOT_6040 (* TFL *)

   Set teletype ready flag. Optionally also raise an interrupt.
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6040(teletype : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Teletype, "TFL");*)
    IF NOT teletype.writer.ready THEN
      CharDevice.SetWriterReadyFlag(teletype, cpu);
    END;

  END IOT_6040;


(* ----------------------------------------------------------------------
   Procedure : IOT_6045 (* TSK *)

   Skip next instruction if : teletype interrupts enabled and either
   device is ready
   ---------------------------------------------------------------------- *)
PROCEDURE IOT_6045(teletype : CharDevice.T; cpu : CPU.T) =
  BEGIN
    (*Diag.Message(Diag.T.Teletype, "TSK");*)
    LOCK teletype.writer DO
      LOCK teletype.reader.readyLock DO
        IF teletype.reader.interruptFlag AND (teletype.reader.ready OR
                                              teletype.writer.ready) THEN
          IF cpu.PC = LAST(CPU.Register_12) THEN
            cpu.PC := 0;
          ELSE
            INC(cpu.PC);
          END;
        END;
      END;
    END;
  END IOT_6045;


BEGIN
END Teletype.







