(* ----------------------------------------------------------------------
   $Id: Buffer.m3,v 1.1 1995/06/19 13:48:24 sgp13 Exp $

   This is a buffer for characters. It is dynamically self-sizing. Reading
   a character may block until there is one available.

   Methods:

     init - allows specification of an initial size for the buffer
     get  - return a character from the buffer
     charsReady - returns number of characters ready
     put - add a character to the buffer

   It is implemented using an pointer to an open character array, and indexes
   of the next positions to be read and written.
   The buffer does not wrap around. Instead, if the writing position overflows
   the end of the buffer, an attempt is made to shift characters down the
   buffer. If this fails, then a new, larger buffer is allocated.

   There is also a condition variable which allows get to block, waiting until
   the next character is written to the buffer.

   ----------------------------------------------------------------------

   $Log: Buffer.m3,v $
# Revision 1.1  1995/06/19  13:48:24  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Buffer;

IMPORT Thread;

REVEAL
  PrivateT = MUTEX BRANDED OBJECT END;
  T = PublicT BRANDED OBJECT
    a : REF ARRAY OF CHAR;
    signal : Thread.Condition;
    nextRead : CARDINAL;
    nextWrite : CARDINAL;
    count : CARDINAL;
  OVERRIDES
    charsReady := CountReady;
    get := Get;
    init := Init;
    put := Put;
  END;

(* ----------------------------------------------------------------------
   Procedure : CountReady

   Return the number of characters ready. NOTE this may change between the
   call and when you actually do something about it.
   ---------------------------------------------------------------------- *)
PROCEDURE CountReady(self : T) : CARDINAL =
  BEGIN
    LOCK self DO
      RETURN self.nextWrite - self.nextRead;
    END;
  END CountReady;


(* ----------------------------------------------------------------------
   Procedure : Get

   Read character from buffer
   ---------------------------------------------------------------------- *)
PROCEDURE Get(self : T) : CHAR RAISES { Thread.Alerted } =
  BEGIN
    LOCK self DO
      (* if no characters then block until there are *)
      WHILE self.nextRead = self.nextWrite DO
        Thread.AlertWait(self, self.signal);
      END;

      INC(self.nextRead);

      RETURN self.a^[self.nextRead - 1];
    END;
  END Get;


(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise buffer. Allocate array and condition variable. Reset pointers.
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T; size : CARDINAL) : T =
  BEGIN
    self.a := NEW(REF ARRAY OF CHAR, size);
    self.signal := NEW(Thread.Condition);
    self.nextRead := 0;
    self.nextWrite := 0;

    RETURN self;
  END Init;

(* ----------------------------------------------------------------------
   Procedure : Put

   Write a character to the buffer.

   If the buffer is currently full, then attempt to shift down the array.
   If this is not possible then allocate a new, larger array.

   ---------------------------------------------------------------------- *)
PROCEDURE Put(self : T; c : CHAR) =
  BEGIN
    WITH oldSize = NUMBER(self.a^) DO
      LOCK self DO
        (* if trying to write beyond buffer *)
        IF self.nextWrite >= oldSize THEN
          (* shift down the buffer if possible *)
          IF self.nextRead > 0 THEN
            FOR i := self.nextRead TO self.nextWrite - 1 DO
              self.a^[i-self.nextRead] := self.a^[i];
            END;
            DEC(self.nextWrite, self.nextRead);
            self.nextRead := 0;
          ELSE
            WITH newBuffer = NEW(REF ARRAY OF CHAR, oldSize * 2) DO
              SUBARRAY(newBuffer^, 0, oldSize) := self.a^;
              self.a := newBuffer;
            END;
          END;
        END;

        (* add character *)
        self.a^[self.nextWrite] := c;
        INC(self.nextWrite);

        (* signal character ready *)
        Thread.Signal(self.signal);

      END;
    END;

  END Put;

BEGIN
END Buffer.
