(* ----------------------------------------------------------------------
   $Id: Diag.m3,v 1.12 1995/06/21 13:55:20 sgp13 Exp $

   Diagnostics information handler for the PDP8/e emulator.

   For usage information see the interface file.

   The initialisation code of this module is used to read the options for
   where to write the information, and for which classes of message to enable.
   It then opens the file if one is specified and writes an initial message
   to it.

   ----------------------------------------------------------------------

   $Log: Diag.m3,v $
# Revision 1.12  1995/06/21  13:55:20  sgp13
# Superfluous call to Init.Reset() removed
#
# Revision 1.11  1995/03/22  12:49:38  sgp13
# Update forced
#
# Revision 1.10  1995/03/07  21:07:44  sgp13
# Update forced
#
# Revision 1.9  1995/02/21  09:36:10  sgp13
# Bug fixed: NEW MUTEX now declared rather than just the MUTEX type
#
# Revision 1.8  1995/02/10  12:04:44  sgp13
# Monitor mutex added on Message, to prevent double entry into this procedure.
#
# Revision 1.7  1995/02/01  13:25:12  sgp13
# Update forced to keep up with Diag.i3
#
# Revision 1.6  1995/01/19  09:18:37  sgp13
# Changed to flush output after every message - to prevent loss of messages.
#
# Revision 1.5  1995/01/12  15:11:45  sgp13
# Update forced to keep version numbers of module and interface the same
#
# Revision 1.3  1994/12/14  11:22:10  sgp13
# Initialisation changed to take account of new Init.Scan method.
#
# Revision 1.2  1994/12/09  11:47:01  sgp13
# Changed to ignore Thread.Alerted exception
#
# Revision 1.1  1994/12/06  12:05:24  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Diag;

IMPORT FileWr, Wr, Stdio, IO, Scan, Lex, Text, OSError, Thread;
IMPORT Init;

TYPE
  EnabledSet = SET OF T;

VAR
  enabled := EnabledSet {T.Error}; (* the set of enabled diagnostics classes *)
  file : Wr.T;                     (* the file for output *)
  tofile : BOOLEAN := FALSE;       (* flag whether to output to file *)
  tostdout : BOOLEAN := FALSE;     (* flag whether to output to stdout *)
  namefound : BOOLEAN := FALSE;    (* flag whether a match is found for *)
                                   (* the parameter *)
  filename : TEXT := NIL;          (* name of file to write to *)
  param : Init.T;                  (* parameter returned by Init *)
  initactions := ARRAY [0..1] OF Init.Action { Init.Action {"File", InitFile },
                                         Init.Action {"Stdout", InitStdOut }};
  monitor := NEW(MUTEX);           (* mutex to ensure that only one thread *)
                                   (* at a time can generate a Diag message *)

(* ----------------------------------------------------------------------
   Procedure : Message

   Produces a message of the class given, containing the text given, if
   messages of this class are enabled.
   Lock the monitor mutex while doing this to ensure that only one thread
   can output a message at a time.

   If writing to diagnostic file fails then bomb out
   ---------------------------------------------------------------------- *)
<* FATAL Wr.Failure *>
PROCEDURE Message(class : T; message : TEXT) =
  BEGIN
    LOCK monitor DO
      IF class IN enabled THEN
        IF tofile THEN
          TRY
            Wr.PutText(file, varname[class] & ": " & message & "\n");
            Wr.Flush(file);
          EXCEPT
          | Thread.Alerted =>
          END;
        END;
        IF tostdout THEN
          TRY
            Wr.PutText(Stdio.stdout, varname[class] & ": " & message & "\n");
            Wr.Flush(Stdio.stdout);
          EXCEPT
          | Thread.Alerted =>
          END;
        END;
      END;
    END;
  END Message;

(* ----------------------------------------------------------------------
   Procedure : Close

   Close the diagnostics output file and flush standard output if
   appropriate.
   ---------------------------------------------------------------------- *)
<* FATAL Wr.Failure *>
PROCEDURE Close() =
  BEGIN
    IF tofile THEN
      TRY
        Wr.Flush(file);
      EXCEPT
      | Thread.Alerted =>
      END;
      TRY
        Wr.Close(file);
      EXCEPT
      | Thread.Alerted =>
      END;
      tofile := FALSE;
    END;

    IF tostdout THEN
      TRY
        Wr.Flush(Stdio.stdout);
      EXCEPT
      | Thread.Alerted =>
      END;
    END;
  END Close;


(* ----------------------------------------------------------------------
   The following procedures are called by the initialisation routine Init.Scan
   if the corresponding lines are found in the initialisation file.
   ---------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------
   Procedure : InitFile

   Initialise the diagnostics output file
   ---------------------------------------------------------------------- *)
PROCEDURE InitFile(<* UNUSED *> name : TEXT; value : TEXT) =
  BEGIN
    tofile := TRUE;
    filename := value;

    TRY
      file := FileWr.Open(filename);
      Wr.PutText(file, "Diagnostics file for PDP8/e emulator\n");
      Wr.PutText(file, "------------------------------------\n\n");
      Wr.Flush(file);
    EXCEPT
    | OSError.E, Wr.Failure => IO.Put("Error opening diagnostics file: " &
                                      filename & "\n");
                               tofile := FALSE;
    END;
  END InitFile;



(* ----------------------------------------------------------------------
   Procedure : InitStdOut

   Initialise standard output for diagnostics
   ---------------------------------------------------------------------- *)
PROCEDURE InitStdOut(name : TEXT; value : TEXT) =
  BEGIN
    TRY
      tostdout := Scan.Bool(value);
    EXCEPT
    | Lex.Error => IO.Put("Diag ignoring invalid boolean value for "
                          & name & "\n");
    END;
  END InitStdOut;



(* ----------------------------------------------------------------------
   Initialisation. This is in two stages, for the two sections of the
   init file, [DiagSetup] and [Diag]

   If thread is alerted during initialisation then bomb out
   ---------------------------------------------------------------------- *)
<* FATAL Thread.Alerted *>
BEGIN
  
  (* first read the setup information *)
  Init.Scan("DiagSetup", initactions);

  (* now read the message class information *)
  (* reset initialisation file pointer *)
  Init.Reset();

  (* read through initialisation file processing [Diag] lines *)
  TRY
    LOOP
      (* read parameter *)
      param := Init.Read("Diag");

      (* flag name not yet found *)
      namefound := FALSE;

      (* search in array and place the appropriate item in the set *)
      (* or take it out *)
      FOR i := FIRST(T) TO LAST(T) DO
        (* name found so process *)
        IF Text.Equal(varname[i], param.name) THEN
          namefound := TRUE;
          TRY
            IF Scan.Bool(param.value) THEN
              enabled := enabled + EnabledSet{i};
            ELSE
              enabled := enabled - EnabledSet{i};
            END;
          EXCEPT
          | Lex.Error => IO.Put("Diag ignoring invalid boolean value for "
                                & param.name & "\n");
          END;
          EXIT;
        END;
      END;

      IF NOT namefound THEN
        IO.Put("Diag ignoring unknnown message class (" & param.name &
               ") in init file.\n");
      END;
    END;
  EXCEPT
  | Init.Empty =>
  END;

END Diag.












