(* ----------------------------------------------------------------------
   $Id: Flags.i3,v 1.7 1995/03/15 14:39:25 sgp13 Exp $

   The interface providing flags for the emulator.

   The types provided are:

   Single - one bit flag
   Signal - a guarded condition variable with which to signal between threads.
            For the protocols on using this see below.
   Run    - the flags to do with running the CPU
   Request- the flags for requesting a service - specialised into
            InterruptRequest and DMARequest

   ----------------------------------------------------------------------

   $Log: Flags.i3,v $
# Revision 1.7  1995/03/15  14:39:25  sgp13
# Interrupt type removed (replaced by IntBus.T)
#
# Revision 1.6  1995/03/14  11:53:35  sgp13
# Interrupt type altered.
#
# Revision 1.5  1995/01/23  12:27:36  sgp13
# Single.invert() method added.
#
# Revision 1.4  1995/01/19  09:49:13  sgp13
# Added clear() method to signal type.
#
# Revision 1.3  1995/01/11  11:38:37  sgp13
# Extra flags added to interrupt request type (requestEnable and inhibited)
#
# Revision 1.2  1994/12/12  16:38:44  sgp13
# Changed to implement the Signal type in a way which ensures correct usage and avoids deadlock
#
# Revision 1.1  1994/12/09  09:41:15  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
INTERFACE Flags;

IMPORT Thread;

TYPE
  Single <: PublicSingle;
  PublicSingle = Thread.Mutex OBJECT
  METHODS
    invert();
    reset();
    set();
    test() : BOOLEAN;
  END;
  (* a single bit flag *)


  Signal <: PublicSignal;
  PublicSignal = PrivateSignal OBJECT
  METHODS
    clear();
    init() : Signal;
    signal();
    wait() RAISES { Thread.Alerted };
  END;
  PrivateSignal <: ROOT;

  (* an inter-thread signal: to wait on this signal execute the following:

     signal.wait()

     This is equivalent to the following :

     LOCK signal DO
       WHILE NOT signal.flag DO
         Thread.Wait(signal, signal.signal);
       END;
       signal.flag := FALSE;
     END


     to signal on the signal execute signal.signal(), which is equivalent to:

     LOCK signal DO
       signal.flag := TRUE;
       Thread.Signal(signal.signal);
     END;

     This protocol is to ensure that if more than one thread is awakened by
     a signal, only one proceeds with valid data.
     The protocol is also necessary to avoid a signal being lost due to
     the target thread not having yet executed the Wait instruction when
     the signal occurs.

     This data type should not be used if more than one thread can act on
     the issue of the signal - a different structure should be used instead.

     Because a signal which is not detected will hold until the next time you
     try to look at it - it may be necessary to clear a signal, so the clear
     method does this.
  *)



  Run <: PublicRun;
  PublicRun = Thread.Mutex OBJECT
    continuous : BOOLEAN := FALSE;
    running : BOOLEAN := FALSE;
    request : Signal;
  METHODS
    init() : Run;
  END;
  (* the flags signalling the run mode of the processor:

     continuous : running continuously if at all
     running : actually executing an instruction
     request : a request to the processor to execute at least one instruction

  *)


  DMARequest = Thread.Mutex OBJECT
    request : INTEGER := 0;
    running : BOOLEAN := FALSE;
    line : ARRAY DMAlines OF BOOLEAN;
  END;
  DMAlines = [0..11];
  (* request for a DMA service:

     request : set to the number of pending requests for this service
     running : a DMA request is already running
     lines : flag whether specific line is being requested on
  *)


END Flags.




