(* ----------------------------------------------------------------------
   $Id: Flags.m3,v 1.7 1995/03/15 14:39:49 sgp13 Exp $

   The initialisations and methods for the flag types.
   Atomic signal and wait methods which work independent of the order in
   which a cv is signalled/waited are implemented.
   ----------------------------------------------------------------------

   $Log: Flags.m3,v $
# Revision 1.7  1995/03/15  14:39:49  sgp13
# Update forced
#
# Revision 1.6  1995/03/14  11:53:58  sgp13
# Update forced
#
# Revision 1.5  1995/01/23  12:27:02  sgp13
# Single.invert() method added.
#
# Revision 1.4  1995/01/19  09:49:36  sgp13
# Added clear() method to signal type.
#
# Revision 1.3  1995/01/11  11:37:40  sgp13
# Change forced to keep in step with Flags.i3
#
# Revision 1.2  1994/12/12  16:37:32  sgp13
# Changed to enhance the Signal type so that it provides the correct signalling
  services.
#
# Revision 1.1  1994/12/09  09:41:20  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Flags;

IMPORT Thread;

REVEAL
  Single = PublicSingle BRANDED OBJECT
    value : BOOLEAN := FALSE;
  OVERRIDES
    invert := SingleInvert;
    reset := SingleReset;
    set := SingleSet;
    test := SingleTest;
  END;

  Signal = PublicSignal BRANDED OBJECT
    cv : Thread.Condition;
    flag : BOOLEAN := FALSE;
  OVERRIDES
    clear := SignalClear;
    init := SignalInit;
    signal := SignalSignal;
    wait := SignalWait;
  END;
  PrivateSignal = Thread.Mutex BRANDED OBJECT END;

  Run = PublicRun BRANDED OBJECT
  OVERRIDES
    init := RunInit;
  END;

(* ----------------------------------------------------------------------
   Procedure : RunInit

   Initialise an object of the Run type
   ---------------------------------------------------------------------- *)
PROCEDURE RunInit(self : Run) : Run =
  BEGIN
    self.request := NEW(Signal).init();
    RETURN self;
  END RunInit;


(* ----------------------------------------------------------------------
   Procedure : SignalClear

   Clear the flag of an object of the signal type - to allow new signals
   to be detected.
   To be used with GREAT CARE.
   ---------------------------------------------------------------------- *)
PROCEDURE SignalClear(self : Signal) =
  BEGIN
    LOCK self DO
      self.flag := FALSE
    END;
  END SignalClear;


(* ----------------------------------------------------------------------
   Procedure : SignalInit

   Initialise an object of the Signal type
   ---------------------------------------------------------------------- *)
PROCEDURE SignalInit(self : Signal) : Signal =
  BEGIN
    self.cv := NEW(Thread.Condition);
    self.flag := FALSE;
    RETURN self;
  END SignalInit;


(* ----------------------------------------------------------------------
   Procedure : SignalSignal

   Signal on a signal object
   ---------------------------------------------------------------------- *)
PROCEDURE SignalSignal(self : Signal) =
  BEGIN
    LOCK self DO
      self.flag := TRUE;
      Thread.Signal(self.cv);
    END;
  END SignalSignal;


(* ----------------------------------------------------------------------
   Procedure : SignalWait

   Wait on a signal object
   ---------------------------------------------------------------------- *)
PROCEDURE SignalWait(self : Signal) RAISES { Thread.Alerted } =
  BEGIN
    LOCK self DO
      WHILE NOT self.flag DO
        Thread.AlertWait(self, self.cv);
      END;
      self.flag := FALSE;
    END;
  END SignalWait;


(* ----------------------------------------------------------------------
   Procedure : SingleInvert

   Invert the flag
   ---------------------------------------------------------------------- *)
PROCEDURE SingleInvert(self : Single) =
  BEGIN
    LOCK self DO
      IF self.value THEN
        self.value := FALSE;
      ELSE
        self.value := TRUE;
      END;
    END;
  END SingleInvert;


(* ----------------------------------------------------------------------
   Procedure : SingleReset

   Reset a single flag
   ---------------------------------------------------------------------- *)
PROCEDURE SingleReset(self : Single) =
  BEGIN
    LOCK self DO
      self.value := FALSE;
    END;
  END SingleReset;


(* ----------------------------------------------------------------------
   Procedure : SingleSet

   Set a single flag
   ---------------------------------------------------------------------- *)
PROCEDURE SingleSet(self : Single) =
  BEGIN
    LOCK self DO
      self.value := TRUE;
    END;
  END SingleSet;


(* ----------------------------------------------------------------------
   Procedure : SingleTest

   Test a single flag
   ---------------------------------------------------------------------- *)
PROCEDURE SingleTest(self : Single) : BOOLEAN =
  BEGIN
    LOCK self DO
      RETURN self.value;
    END;
  END SingleTest;

BEGIN
END Flags.







