(* --------------------------------------------------------------------------
   $Id: Init.i3,v 1.3 1995/01/10 11:32:04 sgp13 Exp $

   Initialisation interface.

   This interface looks for a '-i' switch in the command line and treats the
   parameter following it as a filename specifying an initialisation file.
   The file (if found) is read and interpreted as follows.

   Each line is a potential
   set up command for a module within the program. The file is read into a
   table which is then available for each module to read. A table entry
   consists of the name of the module for which the information is intended,
   the token, and the value.

   To access the information a module calls Reset to set the internal pointer
   to the beginning of the table, followed by Read with its own identifier as
   a parameter. The parameters are returned in the order in which they appear
   in the file.
   Another way for a module to read the information is to pass its name and
   an array of Action types (parameter names and procedures) to the Scan
   procedure, which will scan all the entries in the table which correspond
   to the module and execute the appropriate procedures.

   The structure of the file is as follows:

   [module]        specifies the module to which the next section refers
   name=value      specifies the parameter to be passed to the current module

   Any line commencing with a # is a comment. The splitting into name and
   value is performed strictly around the first '=' sign in the line. No other
   checks are made.

   NB Because of a quirk in the Modula3 runtime, the last line must end with
      a carriage return if it is to be recognised.

   NB If no '-i' switch is specified or the file is not found the
      initialisation is skipped. This allows utilities using modules which
      would normally be initialised from this library to be run standalone and
      to have command line parameters.
   ------------------------------------------------------------------------

   $Log: Init.i3,v $
# Revision 1.3  1995/01/10  11:32:04  sgp13
# Modified so that the initialisation file is specified as the parameter after
# the -i switch on the command line
#
# Revision 1.2  1994/12/14  11:24:17  sgp13
# Added Scan procedure
#
# Revision 1.1  1994/12/06  09:08:22  sgp13
# Initial revision
#

   ------------------------------------------------------------------------ *)
   
INTERFACE Init;

EXCEPTION Empty;    (* returned by read if list is empty *)

TYPE
  T = RECORD        (* a parameter and its value *)
    name  : TEXT := NIL;
    value : TEXT := NIL;
  END;
  Action = RECORD
    name : TEXT;
    proc : Initialisor;
  END;
  Initialisor = PROCEDURE (name : TEXT; value : TEXT) RAISES ANY;


PROCEDURE Reset();
  (* resets the internal pointer back to the beginning of the
     list of parameters
  *)

PROCEDURE Read(module : TEXT) : T RAISES { Empty };
  (* reads the next parameter for the specified module
     empty is raised if there are no more parameters for that module
  *)

(* ----------------------------------------------------------------------
   Procedure : Scan

   Scan all the parameters for the named module and execute its procedures
   accordingly
   ---------------------------------------------------------------------- *)
PROCEDURE Scan(module : TEXT; actions : ARRAY OF Action) RAISES ANY;

END Init.


