(* --------------------------------------------------------------------------
   $Id: Init.m3,v 1.3 1995/01/10 11:31:43 sgp13 Exp $

   Initialisation module.
   For specification details see Init.i3

   The table is implemented as a singly linked list for simplicity. Speed is
   not an issue as the table is only accesssed on initialisation.
   Ideally information should be deleted once it has been read. This is not
   part of the current implementation.

   The module is constructed in this way so that the normal Modula-3
   inferences about initialisation can be used, ie the file is read before any
   of the client modules need it, and so that initialisation is transparent
   to the main program. This also means that it is the responsibility of
   each module to initialise itself.

   -------------------------------------------------------------------------

   $Log: Init.m3,v $
# Revision 1.3  1995/01/10  11:31:43  sgp13
# Modified so that the initialisation file is specified as the parameter after
# the -i switch on the command line
#
# Revision 1.2  1994/12/14  11:21:19  sgp13
# Added Scan procedure.
#
# Revision 1.1  1994/12/06  09:08:34  sgp13
# Initial revision
#

   ----------------------------------------------------------------------- *)
   
MODULE Init;

IMPORT IO, Fmt, Rd, FileRd, OSError, Params, ASCII, Text, Thread;

EXCEPTION Syntax;

TYPE
  ListElement = RECORD
    module : TEXT := NIL;
    param : T;
    next : REF ListElement := NIL;
  END;

VAR
  listHead : REF ListElement := NIL;
  listTail : REF ListElement := NIL;
  line : TEXT := NIL;
  lineno : CARDINAL := 0;
  module : TEXT := NIL;
  param : T;
  currentElement : REF ListElement := NIL;
  index  := 1;        (* the parameter number of the '-i' switch on the *)
                      (* command line *)
  switchFound : BOOLEAN := FALSE;  (* flag whether '-i' has been found *)

(* ----------------------------------------------------------------------- 
   Reset - reset list pointer
   ----------------------------------------------------------------------- *)
PROCEDURE Reset() =
  BEGIN
    currentElement := listHead;
  END Reset;


(* ------------------------------------------------------------------------
   Read - find and return entry for specified module
   ------------------------------------------------------------------------ *)
PROCEDURE Read(module : TEXT) : T RAISES { Empty } =
  VAR
    param : T;
    found : BOOLEAN := FALSE;

  BEGIN
    WHILE currentElement # NIL AND NOT found DO
      IF Text.Equal(module, currentElement^.module) THEN
        param := currentElement^.param;
        found := TRUE;
      END;
      currentElement := currentElement^.next;
    END;

    IF found THEN
      RETURN param;
    ELSE
      RAISE Empty;
    END;
  END Read;


(* ----------------------------------------------------------------------
   Procedure : Scan

   Scan all the parameters for the named module and execute its procedures
   accordingly
   ---------------------------------------------------------------------- *)
PROCEDURE Scan(module : TEXT; actions : ARRAY OF Action) RAISES ANY =
  VAR
    param : T;
    found : BOOLEAN := FALSE;
  BEGIN
    Reset();

    TRY
      LOOP
        param := Read(module);
        found := FALSE;
        FOR i:= FIRST(actions) TO LAST(actions) DO
          IF Text.Equal(param.name, actions[i].name) THEN
            found := TRUE;
            actions[i].proc(param.name, param.value);
            EXIT;
          END;
        END;
        IF NOT found THEN
          IO.Put("Initialising '" & module & "' : ignoring unknown parameter '"
                 & param.name & "'\n");
        END;
      END;
    EXCEPT
    | Empty =>
    END;
  END Scan;


(* ------------------------------------------------------------------
   The following procedures deal with the initialisation code for the
   module

   ------------------------------------------------------------------ *)

(* Tokenise - split the line into parameter name and value *)
PROCEDURE Tokenise(line : TEXT) : T RAISES { Syntax } =
  VAR
    param : T;
  BEGIN
    WITH pos = Text.FindChar(line, '=') DO
      (* check for an '=' sign *)
      IF pos = -1 THEN
        RAISE Syntax;
      END;

      param.name := Text.Sub(line, 0, pos);
      param.value := Text.Sub(line, pos + 1);
      RETURN param;
    END;
  END Tokenise;


(* NextLine - reads the next non-comment, non-blank line from the file *)
PROCEDURE NextLine(file : Rd.T) : TEXT
    RAISES { Rd.EndOfFile, Rd.Failure, Thread.Alerted} =
  VAR
    line : TEXT;
  BEGIN
    REPEAT
      INC(lineno);
      line := Rd.GetLine(file);
    UNTIL NOT(IsBlank(line)) AND NOT (IsComment(line));
    RETURN line;
  END NextLine;

(* IsBlank - determines whether a line contains only whitespace *)
PROCEDURE IsBlank(line : TEXT) : BOOLEAN =
  BEGIN
    IF line = NIL THEN
      RETURN TRUE;
    END;
    WITH len = Text.Length(line) DO
      IF len = 0 THEN
        RETURN TRUE;
      END;

      FOR i := 0 TO len - 1 DO
        IF NOT Text.GetChar(line, i) IN ASCII.Spaces THEN
          RETURN FALSE;
        END;
      END;
    END;
    RETURN TRUE;
  END IsBlank;
      
(* IsComment - determines whether the line is a comment *)
PROCEDURE IsComment(line : TEXT) : BOOLEAN =
  BEGIN
    IF Text.GetChar(line, 0) = '#' THEN
      RETURN TRUE;
    ELSE
      RETURN FALSE;
    END
  END IsComment;

(* the initialisation code identifies and reads the initialisation file *)

(* NB because the diagnostics module can rely upon this module we must not
   rely on that module for error handling.
*)

(* the best way to deal with the below exceptions at this stage is
   to allow the Modula-3 runtime to bomb out
*)
<* FATAL Thread.Alerted, Rd.Failure *>

BEGIN
  (* search for an '-i' switch. don't bother checking the last parameter *)
  (* as there cannot be a filename specified after it! *)
  index := 1;
  WHILE NOT switchFound AND index < Params.Count-1 DO
    IF Text.Equal(Params.Get(index), "-i") THEN
      switchFound := TRUE;
    ELSE
      INC(index);
    END;
  END;

  (* if successful, use the next parameter as the name of the *)
  (*  initialisation file *)
  IF switchFound THEN
    WITH fileName = Params.Get(index+1) DO
      TRY
        WITH file = FileRd.Open(fileName) DO
          (* scan through the file, reading lines, tokenising them *)
          (* and storing in the linked list *)
          WHILE NOT Rd.EOF(file) DO
            TRY
              line := NextLine(file);
            
              (* if the line is a module specifier change the module name *)
              IF Text.GetChar(line, 0) = '[' THEN
                (* strip the '[' and ']' *)
                line := Text.Sub(line, 1);
                WITH pos = Text.FindChar(line, ']') DO
                  IF pos = -1 THEN
                    IO.Put("Line " & Fmt.Int(lineno) & " in parameter file is"
                           & " incomplete: ignoring it\n");
                  ELSE
                    module := Text.Sub(line, 0, pos);
                  END;
                END;
              ELSE
                (* otherwise add the parameter to the list *)
                TRY
                  param := Tokenise(line);
                  IF module = NIL THEN
                    IO.Put("Parameter specified before module name. Line " &
                      Fmt.Int(lineno) & " ignored\n");
                  ELSE
                    currentElement := NEW(REF ListElement);
                    currentElement^.module := module;
                    currentElement^.param := param;
                    currentElement^.next := NIL;
                    
                    (* if list is non-empty then add to tail of list *)
                    (* otherwise start list *)
                    IF listHead # NIL THEN
                      listTail^.next := currentElement;
                      listTail := currentElement;
                    ELSE
                      listHead := currentElement;
                      listTail := currentElement;
                    END;
                  END;
                EXCEPT
                | Syntax => IO.Put("Syntax error in line " & Fmt.Int(lineno) &
                                   " of parameter file : ignoring line\n");
                END;
              END;
            EXCEPT
            | Rd.EndOfFile =>
            END;
          END;
          Rd.Close(file);
        END;
      EXCEPT
      | OSError.E => IO.Put("Error opening init file - using defaults\n");
      END;
    END;
  END;

  (* this pointer is used to point to the current entry in the list *)
  (* start with pointing nowhere to enforce the use of Reset *)
  currentElement := NIL;
END Init.




