(* ----------------------------------------------------------------------
   $Id: IntBus.i3,v 1.5 1995/06/23 14:10:35 sgp13 Exp $

   This module handles the interrupt signals part of the system bus.
   It is based on the following assumptions:

   a) Any device may raise an interrupt: as part of the initialisation process
   the device may request the allocation of a unique interrupt number which
   identifies it in all subsequent dealings with the interrupt system.

   b) The processor increments the cycle count once on each cycle. This allows
      the interrupt bus to keep track of whether interrupts are active.

   c) All interaction with the interrupt flags passes through this module.

   In the following, it is understood that 'specified device' means the piece
   of code offering a given interrupt identifier.

   The following services are provided:

   allocate: allocates a unique interrupt number
   areActive: test whether an interrupt request should be granted:
              interrupts are granted only if:
                  they are enabled             ) implemented using a counter
                  there is no outstanding delay) specifying the cycle when
                                               ) interrupts come active
                  they are uninhibited
   areEnabled: test whether an enable was issued after the last disable: ie
               whether interrupts are or are about to become enabled.
   arePending: test whether any requests are pending
   cancel: cancels any interrupt raised by the specified device
   disable: disable all interrupts - also prevent further calls to 'cycle'
            re-enabling until 'enable' is called.
   enable: request enabling of interrupts on the second call to 'cycle' after
           this call
   request: the specified device is requesting an interrupt
   reset: resets the interrupt system: called by the CAF command

   The 'interrupts inhibited' flag is provided as a BOOLEAN for speed of access
   ----------------------------------------------------------------------

   $Log: IntBus.i3,v $
# Revision 1.5  1995/06/23  14:10:35  sgp13
# delayCount and totalPending made directly available
#
# Revision 1.4  1995/06/23  07:43:18  sgp13
# Update forced
#
# Revision 1.3  1995/06/22  08:34:20  sgp13
# Bug fix: AreEnabled returned the wrong result
#
# Revision 1.2  1995/06/20  10:01:10  sgp13
# Altered to implement using a cycleCount which is directly incremented by
# the processor module. This is intended to increase the speed of the critical
# parts of the interrupt system, ie cycle() and areActive().
#
# Revision 1.1  1995/03/15  09:56:35  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
INTERFACE IntBus;

IMPORT Flags;

EXCEPTION TooMany;
EXCEPTION BadIndex;

CONST
  maxIntTypes = 12;

TYPE
  ID = [1..maxIntTypes];  (* the type of the unique identifier *)

  (* the object type providing the above services *)
  T <: PublicT;
  PublicT = OBJECT
    cycleCount : CARDINAL := 0;
    delayCount : CARDINAL := LAST(CARDINAL);
    totalPending : [0..maxIntTypes] := 0;
    uninhibited : BOOLEAN := TRUE;
  METHODS
    allocate() : ID RAISES { TooMany };
    areActive() : BOOLEAN;
    areEnabled() : BOOLEAN;
    arePending() : BOOLEAN;
    cancel(id : ID) RAISES { BadIndex };
    disable();
    enable();
    init() : T;
    request(id : ID) RAISES  { BadIndex };
    reset();
  END;
  
END IntBus.

