(* ----------------------------------------------------------------------
   $Id: IntBus.m3,v 1.5 1995/06/23 14:11:04 sgp13 Exp $

   This module handles the interrupt signals part of the system bus.
   For details of the interface see the interface file.

   To implement the services the object holds the following data fields:

   delayCount: the number of the cycle on which interrupts become enabled
   uninhibited: flag whether uninhibited
   totalPending: a count of the total interrupts pending: this allows a quick
                 check of whether any are pending to minimise processing time
                 during main processor loop
   pending[]: array of flags, one for each device
   lastAllocated: the last unique number currently allocated

   ----------------------------------------------------------------------

   $Log: IntBus.m3,v $
# Revision 1.5  1995/06/23  14:11:04  sgp13
# delayCount and totalPending made directly available
#
# Revision 1.4  1995/06/23  07:42:50  sgp13
# Some methods optimised
#
# Revision 1.3  1995/06/22  08:33:55  sgp13
# Bug fix: AreEnabled returned the wrong result
#
# Revision 1.2  1995/06/20  10:02:42  sgp13
# Altered to implement using a cycleCount which is directly incremented by
# the processor module. This is intended to increase the speed of the critical
# parts of the interrupt system, ie cycle() and areActive().
#
# Revision 1.1  1995/03/15  09:56:25  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE IntBus;

IMPORT Diag, Flags;

REVEAL
  T = PublicT BRANDED OBJECT
    lastAllocated : [0..maxIntTypes] := 0;
    pending : ARRAY [1..maxIntTypes] OF BOOLEAN;
  OVERRIDES
    allocate := Allocate;
    areActive := AreActive;
    areEnabled := AreEnabled;
    arePending := ArePending;
    cancel := Cancel;
    disable := Disable;
    enable := Enable;
    init := Init;
    request := Request;
    reset := Reset;
  END;

(* ----------------------------------------------------------------------
   Procedure : Allocate

   Allocate a unique interrupt index number
   ---------------------------------------------------------------------- *)
PROCEDURE Allocate(self : T) : ID RAISES { TooMany } =
  BEGIN
    (* if there are none left to allocate then raise fatal exception *)
    IF self.lastAllocated = maxIntTypes THEN
      RAISE TooMany;
    ELSE
      INC(self.lastAllocated);
      RETURN self.lastAllocated;
    END;
  END Allocate;


(* ----------------------------------------------------------------------
   Procedure : AreActive

   Interrupts are active only if:
                  they are enabled
                  there is no outstanding delay
                  they are uninhibited

   ---------------------------------------------------------------------- *)
PROCEDURE AreActive(self : T) : BOOLEAN =
  BEGIN
    RETURN self.cycleCount >= self.delayCount AND self.uninhibited;
  END AreActive;


(* ----------------------------------------------------------------------
   Procedure : AreEnabled

   Check whether interrupts are enabled. If this returns TRUE, it may still
   be the case that an interrupt request may not be granted: see above.
   ---------------------------------------------------------------------- *)
PROCEDURE AreEnabled(self : T) : BOOLEAN =
  BEGIN
    RETURN NOT(self.delayCount = LAST(CARDINAL));
  END AreEnabled;


(* ----------------------------------------------------------------------
   Procedure : ArePending

   There are interrupts pending if the number pending is larger than zero
   ---------------------------------------------------------------------- *)
PROCEDURE ArePending(self : T) : BOOLEAN =
  BEGIN
    RETURN self.totalPending > 0;
  END ArePending;


(* ----------------------------------------------------------------------
   Procedure : Cancel

   Cancel the interrupt associated with the specified device.
   ---------------------------------------------------------------------- *)
PROCEDURE Cancel(self : T; id : ID) RAISES { BadIndex } =
  BEGIN
    IF id > self.lastAllocated THEN
      (*Diag.Message(Diag.T.Error, "illegal interrupt number in cancel()");*)
      RAISE BadIndex;
    END;

    IF self.pending[id] THEN
      self.pending[id] := FALSE;
      DEC(self.totalPending);
    END;

  END Cancel;


(* ----------------------------------------------------------------------
   Procedure : Cycle

   Another processor cycle has finished: if there is a delay in operation
   then decrease it.

   No longer used ... directly increment counter instead
   ---------------------------------------------------------------------- *)
(* PROCEDURE Cycle(self : T) =
  BEGIN
    LOCK self DO
      IF self.delay > 0 THEN
        DEC(self.delay);
      END;
    END;
  END Cycle;
*)

(* ----------------------------------------------------------------------
   Procedure : Disable

   Disable interrupts: reset both the enable flag and the delay counter
   ---------------------------------------------------------------------- *)
PROCEDURE Disable(self : T) =
  BEGIN
    self.delayCount := LAST(CARDINAL);
  END Disable;


(* ----------------------------------------------------------------------
   Procedure : Enable

   Enable interrupts: set the enable flag and institute a two-cycle delay
   before they become active

   ---------------------------------------------------------------------- *)
PROCEDURE Enable(self : T) =
  BEGIN
    IF self.delayCount = LAST(CARDINAL) THEN
      self.delayCount := self.cycleCount + 2;
    END;
  END Enable;


(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise object: instantiate inhibited flag, clear array to zero
   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T) : T =
  BEGIN
    FOR i := 1 TO maxIntTypes DO
      self.pending[i] := FALSE;
    END;

    self.uninhibited := TRUE;

    RETURN self;
  END Init;


(* ----------------------------------------------------------------------
   Procedure : Request

   Request an interrupt
   ---------------------------------------------------------------------- *)
PROCEDURE Request(self : T; id : ID) RAISES { BadIndex } =
  BEGIN
    IF id > self.lastAllocated THEN
      (*Diag.Message(Diag.T.Error, "illegal interrupt number in request()");*)
      RAISE BadIndex;
    END;

    IF NOT self.pending[id] THEN
      self.pending[id] := TRUE;
      INC(self.totalPending);
    END;

  END Request;


(* ----------------------------------------------------------------------
   Procedure : Reset

   Reset the interrupt system: clear all pending interrupts, disable and
   uninhibit interrupts. Do not reset the counter as this is also used as
   an instruction counter.
   ---------------------------------------------------------------------- *)
PROCEDURE Reset(self : T) =
  BEGIN
    FOR i := 1 TO self.lastAllocated DO
      self.pending[i] := FALSE;
    END;

    self.totalPending := 0;
    self.delayCount := LAST(CARDINAL);
    self.uninhibited := TRUE;
  END Reset;


BEGIN
END IntBus.


