(* ----------------------------------------------------------------------
   $Id: Memory.m3,v 1.12 1995/06/24 21:53:20 sgp13 Exp $

   This module implements the memory object.
   Memory is implemented as 8 fields of 4096 12-bit values.
   ----------------------------------------------------------------------

   $Log: Memory.m3,v $
# Revision 1.12  1995/06/24  21:53:20  sgp13
# Memory split into fields so that current field pointers may be used to speed
# access
#
# Revision 1.11  1995/06/22  13:58:46  sgp13
# Memory made direct access (rather than via object methods)
#
# Revision 1.10  1995/06/21  16:06:36  sgp13
# Bug in image reading corrected: did create a new object from the result of
# Pickle.Read() which meant that if was called with EVAL memory.init(file),
# the 'memory' object was unchanged. VAR parameters not allowed for first
# parameter of object method.
#
# Revision 1.9  1995/06/21  13:57:52  sgp13
# Unnecessary RAISES { Thread.Alerted } clauses removed.
#
# Revision 1.8  1995/06/20  13:24:05  sgp13
# Memory is no longer a MUTEX.
# Also make the memory block a direct member of the object, rather than a
# dynamically allocated reference - to save time.
# This means Pickle must pickle the object rather than the memory reference, but
# this should be OK and is not time critical.
#
# Revision 1.7  1995/02/01  13:25:54  sgp13
# Update forced to keep up with Memory.i3
#
# Revision 1.6  1995/01/18  13:17:41  sgp13
# Diagnostics values now padded to four figures where appropriate
#
# Revision 1.5  1995/01/18  12:54:49  sgp13
# dump method now flushes output to file
#
# Revision 1.4  1995/01/11  14:21:26  sgp13
# Added inc() method to increment location and then read
#
# Revision 1.3  1995/01/11  10:22:28  sgp13
# Changed for new memory protocol, where this module is ONLY responsible for
# reads and writes of specified addresses.
#
# Revision 1.2  1994/12/12  16:41:29  sgp13
# Bug in Memory.T.init() fixed - now returns itself
#
# Revision 1.1  1994/12/07  12:38:12  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Memory;

IMPORT Thread, Wr, Rd, Fmt;
IMPORT Pickle;
IMPORT Diag;

REVEAL
  T = Public BRANDED OBJECT
  OVERRIDES
    init := Init;
    dump := Dump;
(*    inc := Inc;
    read := Read;
    write := Write; *)
  END;

(* ----------------------------------------------------------------------
   Procedure : Dump

   Dump whole of core memory to given disc file
   ---------------------------------------------------------------------- *)
PROCEDURE Dump(self : T; sink : Wr.T) RAISES { Failure, Thread.Alerted } =
  BEGIN
    TRY
      FOR i := FIRST(FieldNumber) TO LAST(FieldNumber) DO
        Pickle.Write(sink, self.mem[i]);
      END;
      Wr.Flush(sink);
      (*Diag.Message(Diag.T.Memory, "Dump successful");*)
    EXCEPT
    | Pickle.Error, Wr.Failure =>
        (*Diag.Message(Diag.T.Error, "Memory.Dump failed");*)
        RAISE Failure;
    END;
  END Dump;


(* ----------------------------------------------------------------------
   Procedure : Inc

   Implement autoincrement on specified location. Identical to Read except
   that the value of the location is incremented before returning it.
   
   The value cycles if the top of the range is reached.

   NB The value of 'address' is not checked before incrementing - this
      is the responsibility of the calling code
   ---------------------------------------------------------------------- *)
PROCEDURE Inc(self : T; address : Address; field : FieldNumber) : Value =
  BEGIN
    WITH value = self.mem[field, address] DO
      IF value = LAST(Value) THEN
        value := 0;
      ELSE
        INC(value);
      END;
      (*Diag.Message(Diag.T.Memory, "Read(autoinc): " & Fmt.Int(field) & ":"
                   & Fmt.Pad(Fmt.Int(address, 8), 4, '0') & "=" &
                   Fmt.Pad(Fmt.Int(value, 8), 4, '0'));*)
      RETURN value;
    END;
  END Inc;


(* ----------------------------------------------------------------------
   Procedure : Init

   Initialise the memory object - either from file or as a new blank
   memory. If the memory has already been allocated and initialising a
   blank one then just overwrite the current contents

   ---------------------------------------------------------------------- *)
PROCEDURE Init(self : T; source : Rd.T := NIL) : T RAISES { Thread.Alerted } =
  VAR
    done := FALSE;
  BEGIN
    TRY
      IF source # NIL THEN
        FOR i := FIRST(FieldNumber) TO LAST(FieldNumber) DO
          self.mem[i] := Pickle.Read(source);
        END;
        done := TRUE;
        (*Diag.Message(Diag.T.Memory, "File Init successful");*)
      END;
    EXCEPT
    | Pickle.Error, Rd.Failure, Rd.EndOfFile =>
          (*Diag.Message(Diag.T.Error, "Bad memory file format");*)
    END;

    IF done = FALSE THEN
      FOR i := FIRST(FieldNumber) TO LAST(FieldNumber) DO
        self.mem[i] := NEW(REF Field);
        FOR j := FIRST(Address) TO LAST(Address) DO
          self.mem[i]^[j] := 0;
        END;
      END;
      (*Diag.Message(Diag.T.Memory, "Blank Init successful");*)
    END;

    RETURN self;

  END Init;



(* ----------------------------------------------------------------------
   Procedure : Read

   Read from the specified memory address. Indirection and autoincrement
   are performed by the calling code.

   ---------------------------------------------------------------------- *)
PROCEDURE Read(self : T; address : Address; field : FieldNumber) : Value =
  BEGIN
    WITH value = self.mem[field, address] DO
      (*Diag.Message(Diag.T.Memory, "Read: " & Fmt.Int(field) & ":"
                   & Fmt.Pad(Fmt.Int(address, 8), 4, '0') & "=" &
                   Fmt.Pad(Fmt.Int(value, 8), 4, '0'));*)
      RETURN value;
    END;
  END Read;



(* ----------------------------------------------------------------------
   Procedure : Write

   Write to the specified memory address. Indirection and autoincrement
   are performed by the calling code.

   ---------------------------------------------------------------------- *)
PROCEDURE Write(self : T; address : Address; field : FieldNumber;
                value : Value) =
  BEGIN
    self.mem[field, address] := value;
    (*Diag.Message(Diag.T.Memory, "Write: " & Fmt.Int(field) & ":"
                 & Fmt.Pad(Fmt.Int(address, 8), 4, '0') & "=" &
                 Fmt.Pad(Fmt.Int(value, 8), 4, '0'));*)
  END Write;

BEGIN
END Memory.




