(* ----------------------------------------------------------------------
   Filename : Registers.i3

   This interface supplies the objects to be used for the register file for
   the emulator.

   Registers are not mutexes as the only critical accesses all occur in the
   same thread.


   T is the standard register type

   There are also groups of registers defined : 

   Main : the core processor register file
   ExtdMem : the extended memory register file
   ExtdArith : the EAE register file

   ----------------------------------------------------------------------

   $Log: Registers.i3,v $
# Revision 1.4  1995/06/20  10:46:07  sgp13
# Registers are no longer mutexes as this is unnecessary.
#
# Revision 1.3  1995/01/11  11:40:17  sgp13
# Save field split into SIF and SDF to save IF and DF separately
#
# Revision 1.2  1995/01/09  11:40:52  sgp13
# Added SW line and switch register to main registers file.
# Commented out CPMA from main registers file.
#
# Revision 1.1  1994/12/09  12:23:30  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
INTERFACE Registers;

IMPORT Flags;

EXCEPTION OutOfRange;    (* raised when attempting to write an out of *)
                         (* range value to a register *)
EXCEPTION Reinitialisation;  (* attempted to reinitialise a register *)


TYPE
  (* the possible return types for the 'size' method: number of bits, or
     octal digits, or as the maximum value the register can hold *)
  Size = { Bits, Octal, Max };

  T <: Public;             (* a standard register *)

  Public = OBJECT  (* the allowed actions for a standard register *)
  METHODS
    size(type : Size := Size.Octal) : CARDINAL;
    inc();                 (* NB inc does NOT generate error on overflow *)
    init(size (* in bits *): CARDINAL) : T RAISES { Reinitialisation };
    read() : CARDINAL;
    write(value : CARDINAL) RAISES { OutOfRange };
  END;

  (* the main core processor register file *)
  Main <: PublicMain;
  PublicMain = OBJECT
    AC : T;              (* accumulator *)
    L : Flags.Single;    (* link *)
    IR : T;              (* instruction register *)
    MQ : T;              (* multiplier quotient *)
    PC : T;              (* program counter *)
(*  ----------------- CPMA just confuses and slows us down ? ------------- *)
(*  CPMA : T;            (* central processor memory address *) *)
(* ----------------------------------------------------------------------- *)
    MB : T;              (* memory buffer *)
    SR : T;              (* switch register *)
    SW : Flags.Single;   (* flag for SW line *)
  METHODS
    init() : Main RAISES { Reinitialisation };
  END;


  (* extended arithmetic element register file *)
  ExtdArith <: PublicExtdArith;
  PublicExtdArith = OBJECT
    SC : T;              (* step counter *)
    IR : T;              (* instruction register *)
  METHODS
    init() : ExtdArith RAISES { Reinitialisation };
  END;


  (* extended memory element register file *)
  ExtdMem <: PublicExtdMem;
  PublicExtdMem = OBJECT
    I_F : T;             (* instruction field *)
                         (* NB naming problem - cannot be called IF *)
    DF : T;              (* data field *)
    IB : T;              (* instruction field buffer *)
    SIF : T;             (* save instruction field *)
    SDF : T;             (* save data field  *)
  METHODS
    init() : ExtdMem RAISES { Reinitialisation };
  END;


END Registers.












