(* ----------------------------------------------------------------------
   $Id: Registers.m3,v 1.4 1995/06/20 10:46:17 sgp13 Exp $


   ----------------------------------------------------------------------

   $Log: Registers.m3,v $
# Revision 1.4  1995/06/20  10:46:17  sgp13
# Registers are no longer mutexes as this is unnecessary.
#
# Revision 1.3  1995/01/11  11:39:57  sgp13
# Save field split into SIF and SDF to save IF and DF separately
#
# Revision 1.2  1995/01/09  11:41:49  sgp13
# Added SW line and switch register to main registers file.
# Commented out CPMA from main registers file.
#
# Revision 1.1  1994/12/09  12:23:03  sgp13
# Initial revision
#

   ---------------------------------------------------------------------- *)
MODULE Registers;

IMPORT Diag, Flags;

REVEAL
  (* the standard register *)
  T = Public BRANDED OBJECT
    sizeStore : CARDINAL := 0;  (* stored as maximum of the register value+1 *)
    value : CARDINAL := 0;
  OVERRIDES
    size := GetSize;
    inc := Increment;
    init := InitReg;
    read := Read;
    write := Write;
  END;


  (* the core register file *)
  Main = PublicMain BRANDED OBJECT
  OVERRIDES
    init := InitMain;
  END;


  (* the EAE register file *)
  ExtdArith = PublicExtdArith BRANDED OBJECT
  OVERRIDES
    init := InitExtdArith;
  END;


  (* the extended memory register file *)
  ExtdMem = PublicExtdMem BRANDED OBJECT
  OVERRIDES
    init := InitExtdMem;
  END;



(* ----------------------------------------------------------------------
   Procedure : GetSize

   Returns the size of a register : in bits, octal digits, or as a maximum
   value storable. The default is in octal digits
   ---------------------------------------------------------------------- *)
PROCEDURE GetSize(self : T; type : Size := Size.Octal) : CARDINAL =
  VAR
    bitsize : CARDINAL := 0;
    temp : CARDINAL;
  BEGIN
    IF type = Size.Max THEN
      RETURN self.sizeStore;
    ELSE
      (* calculate size in bits *)
      temp := self.sizeStore;
      WHILE temp > 1 DO
        INC(bitsize);
        temp := temp DIV 2;
      END;

      IF type = Size.Bits THEN
        RETURN bitsize;
      ELSE
        IF bitsize MOD 3 = 0 THEN
          RETURN bitsize DIV 3;
        ELSE
          RETURN bitsize DIV 3 + 1;
        END;
      END;
    END;
  END GetSize;



(* ----------------------------------------------------------------------
   Procedure : Increment

   Increment a register. Does not generate an error on overflow
   ---------------------------------------------------------------------- *)
PROCEDURE Increment(self : T) =
  BEGIN
    INC(self.value);
    IF self.value >= self.sizeStore THEN
      self.value := self.value MOD self.sizeStore;
      (*Diag.Message(Diag.T.Registers, "Increment (overflow)");*)
    ELSE
      (*Diag.Message(Diag.T.Registers, "Increment");*)
    END
  END Increment;



(* ----------------------------------------------------------------------
   Procedure : InitExtdArith

   Initialise the extended arithmetic register file
   ---------------------------------------------------------------------- *)
PROCEDURE InitExtdArith(self : ExtdArith) : ExtdArith
    RAISES { Reinitialisation } =
  BEGIN
    (* fails if object already initialised *)
    IF self.SC # NIL THEN
      (*Diag.Message(Diag.T.Registers, "Reinitialisation of ExtdArith object" &
                                     " attempted");*)
      RAISE Reinitialisation;
    END;

    self.SC := NEW(T).init(5);
    self.IR := NEW(T).init(4);
    (*Diag.Message(Diag.T.Registers, "Initialised EAE object");*)
    RETURN self;
  END InitExtdArith;



(* ----------------------------------------------------------------------
   Procedure : InitExtdMem

   Initialise the extended memory register file
   ---------------------------------------------------------------------- *)
PROCEDURE InitExtdMem(self : ExtdMem) : ExtdMem RAISES { Reinitialisation } =
  BEGIN
    (* fails if object already initialised *)
    IF self.I_F # NIL THEN
      (*Diag.Message(Diag.T.Registers, "Reinitialisation of ExtdMem object" &
                                     " attempted");*)
      RAISE Reinitialisation;
    END;

    self.I_F := NEW(T).init(3);
    self.DF := NEW(T).init(3);
    self.IB := NEW(T).init(3);
    self.SIF := NEW(T).init(3);
    self.SDF := NEW(T).init(3);
    (*Diag.Message(Diag.T.Registers, "Initialised ExtdMem object");*)
    RETURN self;
  END InitExtdMem;



(* ----------------------------------------------------------------------
   Procedure : InitMain

   Initialise the main register file
   ---------------------------------------------------------------------- *)
PROCEDURE InitMain(self : Main) : Main RAISES { Reinitialisation } =
  BEGIN
    (* fails if object already initialised *)
    IF self.AC # NIL THEN
      (*Diag.Message(Diag.T.Registers, "Reinitialisation of Main object" &
                                     " attempted");*)
      RAISE Reinitialisation;
    END;

    self.AC := NEW(T).init(12);
    self.L := NEW(Flags.Single);
    self.IR := NEW(T).init(3);
    self.MQ := NEW(T).init(12);
    self.PC := NEW(T).init(12);
(*  self.CPMA := NEW(T).init(12);        (* UNUSED ? *) *)
    self.MB := NEW(T).init(12);
    self.SR := NEW(T).init(12);
    self.SW := NEW(Flags.Single);
    (*Diag.Message(Diag.T.Registers, "Initialised Main object");*)
    RETURN self;
  END InitMain; 



(* ----------------------------------------------------------------------
   Procedure : InitReg

   Create a new register of the specified size : the size is specified
   as a width in bits.
   ---------------------------------------------------------------------- *)
PROCEDURE InitReg(self : T; size : CARDINAL) : T RAISES {Reinitialisation} =
  VAR
    maxValue : CARDINAL := 1;
  BEGIN
    (* calculate max value from size in bits *)
    WHILE size > 0 DO
      maxValue := maxValue * 2;
      DEC(size);
    END;

    (* if not already initialised *)
    IF self.sizeStore = 0 THEN
      self.sizeStore := maxValue;
      (*Diag.Message(Diag.T.Registers, "Initialised T object");*)
      RETURN self;
    ELSE
      (* otherwise *)
      (*Diag.Message(Diag.T.Registers, "Reinitialisation of T object" &
                                     " attempted");*)
      RAISE Reinitialisation;
    END;
  END InitReg;



(* ----------------------------------------------------------------------
   Procedure : Read

   Read the value of a register
   ---------------------------------------------------------------------- *)
PROCEDURE Read(self : T) : CARDINAL =
  BEGIN
    RETURN self.value;
  END Read;



(* ----------------------------------------------------------------------
   Procedure : Write

   Write the value of a register - if valid
   ---------------------------------------------------------------------- *)
PROCEDURE Write(self : T; value : CARDINAL ) RAISES { OutOfRange } =
  BEGIN
    IF value >= self.sizeStore THEN
      RAISE OutOfRange;
    ELSE
      self.value := value;
    END;
  END Write;


BEGIN
END Registers.






